
local scale_notes, scale_chrd_notes
local chrd_with_scale = false

local prev_note, prev_src_note


function findExactScale(notes)
  local best_scale_val = 0
  local best_scale, best_key

  for k=2,#Scales.forms do 
    local scale_key_tb = {0,0,0,0,0,0,0,0,0,0,0,0}
    for j=0,11 do     
      local foundTb = {0,0,0,0,0,0,0,0,0,0,0,0}
      local scale = scaleKey (Scales.forms[k], j)  
      
      for _,note in ipairs(notes) do
        if member (note, scale) then
          foundTb[note+1] = 1
         end 
      end
      scale_key_tb[j+1] =  sumTb(foundTb)
    end  
    
    local ix = tbMaxIx(scale_key_tb)  
    
    if ix == nil then 
      return 1,1,1
    end
    
    if scale_key_tb[ix] > best_scale_val then
      best_scale_val = scale_key_tb[ix]
      best_scale = k
      best_key = ix
    end
  end
  return best_scale, best_key, best_scale_val
end


function findScaleKeyProf(typ, pitch_duration_dst)
  local sca, key
  local notes = {}
  
  local every_n = 64
  
  local best_major, best_minor = -9999,-9999
  local major_key, minor_key 
  
  local pat    
  local line = 0
  local pat_con, pat_len, line2 = getPatFun(MW.block_editing)  
    
  if MW.block_editing == false then line2 = line2 - 1 end
  line = line2     

  local major = copy(Scales.profiles[typ].major)
  local minor = copy(Scales.profiles[typ].minor)
  
  for shift=0,11 do    
    if shift > 0 then
      wrap(major,1)  
      wrap(minor,1)    
    end
             
    local c = pearson2tbs(major, pitch_duration_dst)        
    if c > best_major then
      best_major = c  
      major_key = shift
    end
        
    c = pearson2tbs(minor, pitch_duration_dst)
    if c > best_minor then
      best_minor = c
      minor_key = shift
    end     
  end
    
  sca = 3
  key = minor_key
  if best_major > best_minor then
    sca = 2 
    key = major_key
  end  
    
  return sca,key+1
end


function findScaleKey()
  if Scales.profiles.no > 1 then
    local profile = Scales.profiles.pu[Scales.profiles.no] 
    return findScaleKeyProf(profile, Stats.pitch_duration_dst)  
  else
    return findExactScale(Stats.unique_notes) 
  end
end

 
function findScaleKeyArg(unique_notes, pitch_duration_dst)
  if Scales.profiles.no > 1 then
    local profile = Scales.profiles.pu[Scales.profiles.no] 
    return findScaleKeyProf(profile, pitch_duration_dst)  
  else
    return findExactScale(unique_notes) 
  end
end 
 

local function getChordsScale()

  local scale,key 
  local chrd_scale,chrd_key
  local all_chords = {}
  
  for _,v in pairs(Chords.pos) do
    tableConcat(all_chords,v.chord_notes) -- combine all chords 
  end 
   
  all_chords = copy(map(all_chords,function(v) return v % 12 end)) 
  chrd_scale, chrd_key = findScaleKey(all_chords)  
  scale_chrd_notes = scaleKey (Scales.forms[chrd_scale], chrd_key - 1)       

  if Chords.scale and Chords.key and Chords.scale ~= 1 and Chords.key ~= 1 then 
    chrd_with_scale = true
    
    scale = Chords.scale
    key = Chords.key
    scale_notes = scaleKey (Scales.forms[scale], key - 1)  
  end 
end


local function clearChordSpace(pat_con)   

  if MW.phr.editor_chords and MW.harmonize_typ == 1 then
    for k,v in ipairs(Chords.pos) do    
      local st = v.st
      local en = v.en      
      
       if st > 0 then
         for i=st,en do
           pat = pat_con(i, j)
           pat.note_value = 121
           pat.instrument_string  = ""
           pat.volume_string  = ""
         end
       end 
    end
  end   
end    


local function getNotesFromTb()  

  local pat_con, pat_len, line2 = getPatFun(MW.block_editing)
  local line = 0   
  
  if MW.block_editing == false then line2 = line2 - 1 end
  line = line2   
  
  for i=1,#Nptr.len do         
    line = line +  Nptr.len[i]
           
    if line > pat_len then break end
    
    if Nptr.len[i] > 0 then         
      local len = 1     
      
      if (Nptr.off[i] > 0) then
        len = Nptr.off[i]
      elseif i<#Nptr.len and Nptr.len[i+1] > 0 then   
        len = Nptr.len[i+1]  
      else
        len = pat_len - line + 1
      end
    
      for _,v in ipairs(Chords.pos) do
         
        if line >=  v.st and line <=  v.en then

          if v.notes_r[ Nptr.note[i] ] then  -- count note importance
            v.notes_r[ Nptr.note[i] ]  = v.notes_r[Nptr.note[i]] + len
          else
            v.notes_r[ Nptr.note[i] ] = {}
            v.notes_r[ Nptr.note[i] ] = len         
          end  

          if member(Nptr.note[i], v.notes) == false then   
            table.insert(v.notes, Nptr.note[i])    
            table.insert(v.oct, Nptr.oct[i])   
            table.insert(v.notes_full, Nptr.note[i] + Nptr.oct[i] * 12)                
          end                               
        end
      end
    end         
  end
end 


function getChordPos(len_tb) 

  local line = 0
  local last_note_ix = 0  
  local sums = {}
  
  local pat_con, pat_len, st2 = getPatFun(MW.block_editing)
  
  for i=1,#len_tb do
    if len_tb[i] > 0 then 
      last_note_ix = i
    end   
  end  
  
  if last_note_ix == 0 then return false end
  
  for i=1,last_note_ix do  
    line = line + len_tb[i]
    sums[i] = line - 1
  end
  
  local len_tb2 = copy(len_tb)
  
  if sums[last_note_ix] < pat_len or last_note_ix == #len_tb then  
    len_tb2[last_note_ix+1] = pat_len - sums[last_note_ix] + 1
  end
   
  Chords.pos = {}
   
  for i=1,last_note_ix do
    local st =  sums[i] + st2-1
    local en = sums[i] + len_tb2[i+1] + st2-1      
    
    if st == en then 
      en = pat_len  
    end    
    
      Chords.pos[i] = {}
      Chords.pos[i].chord_notes = {}
      Chords.pos[i].chord_notes_full = {}
      Chords.pos[i].notes = {}
      Chords.pos[i].notes_r = {}
      Chords.pos[i].notes_full = {}      
      Chords.pos[i].oct = {}    
      Chords.pos[i].st = st + 1
      Chords.pos[i].en = en  
      
      if en == 1 then 
        Chords.pos[i].en = Chords.pos[i].en + 1
      end
  end 
  
  return true
end   


function getChordNotes()
  for j=1,#Chords.pos do    
    if CHptr.notes[j] then
      Chords.pos[j].chord_notes = map(CHptr.notes[j], function(v) return v%12 end)
      Chords.pos[j].chord_notes_full = tbOct0(CHptr.notes[j])
    end
  end
end


local function findHarmony(notes)
  
  if #notes < 3 then return notes end
  
  local base_N = notes[1]        
  local ok = {base_N}
       
  for i=2,#notes do
    local note = notes[i]
        
    if dissonanceNCheckOK(base_N, note) then
      table.insert(ok,note)
    end
  end
      
  if #ok < 3 then return ok end 
  
  local ok2  
  local len = 3
  
  while len <= #ok do    
    ok2 = {unpack(ok,1,len-1)} 

    for i=len,#ok do 
      if dissonanceCheckOK(ok2, ok[i]) then
        table.insert(ok2, ok[i]) 
      end 
    end
    
    ok = {}; ok = ok2  -- grow harmonic notes    
    len = len + 1  
 end
 return ok
end    

  
local function harmonizeNotesChords()  
  local last_chrd_with_notes
  
  for k,v in ipairs(Chords.pos) do 
        
    if #v.notes > 1 then -- combine chords with no notes
      v.notes_r_key = sortKeys(v.notes_r)  
      
      last_chrd_with_notes = k
      
      v.harmony = findHarmony(v.notes_r_key)  -- remove dissonant notes
      table.sort(v.harmony)    
    elseif last_chrd_with_notes then     
      local dst = Chords.pos[last_chrd_with_notes] 

      tableConcat(dst.chord_notes, v.chord_notes)
      dst.chord_notes = findHarmony(dst.chord_notes)  
    end    
  end  
end
 
 
local function harmonizeNote(k,note,oct)  
  local v = Chords.pos[k]   
  local src_note = note + oct*12
              
  if v.harmony and #v.harmony > 0 then 
    local ix = retIx (v.harmony,note)          
    local ix2, val = nearestValue(v.harmony, note)
            
    if ix > #v.chord_notes then ix = #v.chord_notes end  
    if ix2 > #v.chord_notes then ix2 = #v.chord_notes end  
                 
    if ix > 0 then                     
      note = v.chord_notes[ix]
    elseif val then 
      note = v.chord_notes[ix2]       
    end
  else  
    note = noteInScale(note, v.chord_notes) 
  end

  local full_note =  note + oct * 12    
                  
  if prev_src_note then  
    if src_note == prev_src_note and full_note ~= prev_note then    

      note, oct = oct0(prev_note)
      note =  noteInScale(note, v.chord_notes) 
    elseif src_note > prev_src_note and full_note <= prev_note then 
              
      local pos, val = nearestValue(v.chord_notes, prev_note%12)    
      if pos > 0 then   
              
        if pos == #v.chord_notes then
          oct = oct + 1  
        end              
         pos = clampAround(pos+1,1,#v.chord_notes)              
         note = v.chord_notes[pos]    
      end
    elseif src_note < prev_src_note and full_note >= prev_note then          

      local pos, val = nearestValue(v.chord_notes,prev_note%12)        
       if pos > 0 then               
         if pos == 1 then
           oct = oct - 1 
         end
         pos = clampAround(pos-1,1,#v.chord_notes)          
         note = v.chord_notes[pos]  
      end
    end  
  end

  return clamp(note + oct*12,0,121)
end


function harmonizeP2Ch () 
 
  if #Chords.pos == 0 then
    return
  else
    sprint('Harmonize')  
  end  

  if noteLengthEmpty (CHptr.len) or noteLengthEmpty (Nptr.len) then sprint ('Load chord to harmonize first') return end  

  local pat    
  local line           = 0
  local last_note_ix   = 0 
  local last_note_line = 0 
  local pat_con, pat_len, line2 = getPatFun(MW.block_editing)  
    
  if MW.block_editing == false then line2 = line2 - 1 end
  line = line2   

  getChordNotes()  
  getChordsScale() 
  getNotesFromTb()
  harmonizeNotesChords()  
  
  clearChordSpace(pat_con) 
  
  local sums = {}   
  local note 

  for i=1,#Nptr.len do     
    line = line +  Nptr.len[i]  
    sums[i] = line
     
    if line > pat_len then break end
    
    if Nptr.len[i] > 0 then     
    
      last_note_line = line
     
      local chord_notes = {}
      local chord_octs = {}   
      local oct = Nptr.oct[i]
     
      local note = Nptr.note[i] 
  
      for k,v in pairs(Chords.pos) do                  
        if #v.notes > 0 and line >=  v.st and line <=  v.en then  
                  
          local new_note = harmonizeNote(k,note,oct)  
          
          note, oct    = oct0(new_note)          
          Nptr.oct[i]  = oct
          Nptr.note[i] = note       
        end
      end           
    end         
  end
  
  MW.timer_pos = {} 
  for i in pairs(Chords.pos) do 
    for j=Chords.pos[i].st,Chords.pos[i].en do
      MW.timer_pos[j-1] = i
    end
  end 
end


function harmonize(track) 

  if #Chords.pos == 0 then
    return
  else
    sprint('Harmonize track')  
  end    
  
  getChordNotes()   
  getChordsScale()    
        
  local pat_con = getPatTrFun(track)      
  
  local cur_pat = renoise.song().selected_pattern_index
  
  local n_tb = {}  
  for n_col=1,12 do
    n_tb[n_col] = grabColBasicFun(n_col,getPatFun,{MW.block_editing,cur_pat,track})    
  end 
      
  for i,v in ipairs(Chords.pos) do 
    
    local notes = {}
    local notes_full = {}
       
    for note_col=1,12 do      
      for _,pat_ln in ipairs(n_tb[note_col]) do 
        if pat_ln.line >= v.st and pat_ln.line < v.en then
      
          local note = pat_ln.note             
          local norm_note, oct = oct0(pat_ln.note) 
          local len = pat_ln.alen
            
          v.notes_full = note
          v.notes = norm_note 
              
          if member(norm_note+1, notes) == false then           
            notes[#notes+1] = norm_note   
            notes_full[#notes_full+1] = note    
          end 
            
          if v.notes_r then  -- count repetition of each note
            if v.notes_r[norm_note] then
              v.notes_r[norm_note]  = v.notes_r[norm_note] + len
            else
              v.notes_r[norm_note] = {}
              v.notes_r[norm_note] = len              
            end  
          else
            v.notes_r = {}
          end               
        end
      end
    end                         
    
    table.sort(notes)      
    Chords.pos[i].notes = notes  
    
    table.sort(notes_full)  
    Chords.pos[i].notes_full = notes_full
  end        

  harmonizeNotesChords()  

  for k,v in ipairs(Chords.pos) do   
    if #v.notes > 0 and v.st > 0 then           
      for ln=v.st, v.en do 
        for note_col=1,12 do
     
          local pat = pat_con(ln, note_col)
          local note = pat.note_value
          if note < 120 then    

            local norm_note, oct = oct0(note)          
            local new_note = harmonizeNote(k,norm_note,oct)  
          
            prev_note = new_note
            prev_src_note = note
        
            fillNote2(track,new_note, note_col, ln)  
          end              
        end
      end
    end
  end    
end


local function chrdBasic(score)

  local sca = UI.scales_pu.value
  local key = UI.keys_pu.value
  
  if UI.scale_sw.value == 2 or sca == 1 then        
    sca, key =  findScaleKey()  -- findScaleKey('exact',n_tb , n_off_tb, l_tb)  
  end
  
  local res, res2 = {},{}   
      
  for form_ix,form in ipairs(Chords.forms) do    
    for deg=0,11 do    
      local chrd = getChrd (deg, sca, key, form)  
        chrd = copy(chrd)   
            
        local harmonic_score = consonanceRating2Tb(chrd,chrd)  -- dissonanceScore2Tb(chrd,chrd)      
      if #chrd >= CH2P.chord_min and #chrd <= CH2P.chord_max then
        local nltb2 = {0,0,0,0,0,0,0,0,0,0,0,0}    

        if harmonic_score <= score then -- score then  
          local not_found = true
          for k,v in ipairs(res2) do
            if tableComp(v,chrd) then
              not_found = false
              break
            end
          end
            
          if not_found then
            for _,note in ipairs(chrd) do 
              nltb2[note+1] = 1      
            end  
                
            local tb = {}
            tb.notes = nltb2
            tb.scale = sca
            tb.key = key
            tb.deg = deg
            tb.form = form_ix
                
            table.insert(res, tb)
            table.insert(res2, chrd)
          end
        end
      end          
    end
  end
  
  return res
end    

  
function harmonizePhraseAuto(o, ...) 
  
  local chords = chrdBasic(Stats.consonance_score) 
  
  local best_cor, last_best_cor, best_ix = -9998, -9999
    
  local cond, pat
      
  local line, last_line = 0,1
  local pat_con, pat_len, line2 = getPatFun(MW.block_editing) 
    
  local nltb = {0,0,0,0,0,0,0,0,0,0,0,0}
  
  local durs, octs, notes, dif, last_chord = {},{},{},{},{}
      
  local count   = 0     
  local first_n = true     
      
  local output   
  if arg[1] then 
    output = function(chrd)   
      table.insert(arg[1], {notes = notes, dur = durs, octs = octs, line = last_line, chord = chrd, dif = dif})  
    end
  else    
    emptyTrN(MW.harmonize_track)
      
    output = function(chrd) 
      fillNotes({120,120,120,120,120,120}, MW.harmonize_track, last_line, 255) 
      fillNotes(chrd, MW.harmonize_track, last_line, MW.harmonize_inst)
    end
  end  
    
  for i=1,#o do 
        
    if o[i].line >= line2 then         

      funOverElOrTb(function(note, dur, tb) tb[note+1] = tb[note+1] + dur end, o[i].note, o[i].dur, nltb)
    
      table.insert(dif, o[i].line - last_line + 1)
       
      if not member(o[i].note,notes) then              
        table.insert(durs,  o[i].dur)
        table.insert(octs,  o[i].oct)
        table.insert(notes, o[i].note)        
      end
          
      for k,v in ipairs(chords) do
        local c = pearson2tbs(nltb, v.notes)   
  
        if c > best_cor then
          best_cor = c
          best_ix = k
        end
      end
        
      if best_cor == last_best_cor then
        count = count+1
      end            
  
      cond = i>1 and #notes > 1 and best_cor <= last_best_cor and count >= CH2P.best_count and #notes >= CH2P.notes_min
  
      if not tableComp(chords[best_ix], last_chord) and (cond or i == #o) then    
        if first_n then 
          last_line = 1
          first_n = false 
        end
          
        local chrd = chords[best_ix]      
        local form = Chords.forms[chrd.form]
        
        local chrd_tmp = getChrd(chrd.deg, chrd.scale, chrd.key, form)         
        chrd = map(chrd_tmp, function(v) return v+CH2P.root_oct*12 end)            
          
        if CH2P.allow_dups then
          output(chrd)
          last_line = o[i].line
          
        else
          local chrd_str = table.concat(chrd,'_')
            
          if chrd_str ~= last_chord then
            output(chrd) 
            last_line = o[i].line
            
          elseif arg[1] and #arg[1]>0 then
            tableConcat(arg[1][#arg[1]].dif, dif) 
            tableConcat(arg[1][#arg[1]].dur, durs)     
            tableConcat(arg[1][#arg[1]].octs, octs)
            tableConcat(arg[1][#arg[1]].notes, notes) 
          end
            
          last_chord = chrd_str
        end
          
        if arg[2] then return end   
          
        nltb = {0,0,0,0,0,0,0,0,0,0,0,0}        
          
        best_cor      = -9998
        last_best_cor = -9999  
          
        octs, durs, notes, dif = {},{},{},{}                     
        count = 0
      end      
        
      if o[i].line > pat_len then return end           

      last_best_cor = best_cor
    end
  end  
end


local function harmonizePhraseTb(rp, o, ...) 
  
  local chords = chrdBasic(Stats.consonance_score)
  
  local best_cor, last_best_cor, best_ix = -9998, -9999
  
  local pat 
  local line = 0
  local pat_con, pat_len, line2 = getPatFun(MW.block_editing)
  
  local notes, durs, octs, dif = {},{},{},{}
    
  local count = 2
  local line_chrd = rp[1]
  
  --if MW.block_editing == false then line = line2 - 1 end
  
  local nltb = {0,0,0,0,0,0,0,0,0,0,0,0}
  
  local output   
  if arg[1] then 
    output = function(chrd)   
      table.insert(arg[1], {notes = notes, dur = durs, octs = octs, line = line_chrd, chord = chrd, dif = dif})
    end
  else    
    emptyTrN(MW.harmonize_track)
    
    output = function(chrd)  
      fillNotes({120,120,120,120,120,120}, MW.harmonize_track, line_chrd, 255) 
      fillNotes(chrd, MW.harmonize_track, line_chrd, MW.harmonize_inst)
    end
  end  
  
  if not arg[1] then 
    emptyTrN(MW.harmonize_track)
  end
  
  for i=1,#o do
            
    if o[i].line >= line2 then 
      
      funOverElOrTb(function(note, dur, tb) tb[note+1] = tb[note+1] + dur end, o[i].note, o[i].dur, nltb)
      
      table.insert(dif, o[i].line - line_chrd + 1)
      
      table.insert(durs,  o[i].dur)
      table.insert(octs,  o[i].oct) 
      table.insert(notes, o[i].note)
      
      
      if o[i].line >= rp[count] or i == #o  then
            
        for k,v in ipairs(chords) do
          local c = pearson2tbs(nltb, v.notes)   
        
          if c > best_cor then
            best_cor = c
            best_ix = k
          end
        end
           
        local chrd = chords[best_ix] 
        local form = Chords.forms[chrd.form]
      
        chrd = getChrd (chrd.deg, chrd.scale, chrd.key, form) 
        chrd = map(chrd, function(v) return v+CH2P.root_oct*12 end)
        
        output(chrd)

        line_chrd = rp[count]
        
        nltb = {0,0,0,0,0,0,0,0,0,0,0,0}        
        best_cor = -9999
        last_best_cor = -9999
        
        dif = {}; octs = {}; durs  = {}; notes = {}  
        
        if count < #rp then
          count=count+1
        end
      end 
    
      if line > pat_len then return end            
    end
  end 
end


local function harmonizePhraseFX(o,...) 
  local tb = o.fx -- getTrackFxH(false)
  
  if #tb == 0 then 
    return 
  else
    harmonizePhraseTb(tb,o,...)  
  end
end


local function harmonizePhraseLn(o,...)   
  local pat_len = renoise.song().selected_pattern.number_of_lines  
  local tb = map(enumFromTo(0,math.floor(pat_len/CH2P.chord_dur)), function(v) return v*CH2P.chord_dur+1 end)

  harmonizePhraseTb(tb,o,...)  
end


function harmonizeCh2P(o,...) 
  if #o == 0 then return end
    
  if CH2P.type == 1 then  -- auto  
    harmonizePhraseAuto(o,...)
  elseif CH2P.type == 2 then -- interval
    harmonizePhraseLn(o,...) 
  elseif CH2P.type == 3 then -- fx
    harmonizePhraseFX(o,...) 
  end
end


function harmonizePhraseApply()   
  if MW.phr.event_type <= 2 then
    harmonizeCh2P(ScannerH())
  else 
    harmonizeCh2P(ScannerChordH())
  end
end

