

function getSlopeR2(ranks, counts)    

  if #ranks ~= #counts then return 0,0 end
 
  local sumX, sumY, sumXY, sumX2, sumY2 = 0,0,0,0,0
  local sumX, sumY, sumXY, sumX2, sumY2 = 0,0,0,0,0
  local numberOfRanks = #ranks
   
  local r,r2,slope
  if numberOfRanks == 1 then
    r2    = 0
    slope = 5
  else
    local i = 0
    local allCountsEqual = true  
    
    while allCountsEqual and i < numberOfRanks-1 do
      allCountsEqual = counts[i] == counts[i + 1]  
      i = i + 1
    end
        
    if allCountsEqual then 
      r2    = 1
      slope = 5      
    else
      for index = 1,numberOfRanks do
        sumX  = sumX  + math.log(ranks[index])
        sumY  = sumY  + math.log(counts[index])
        sumXY = sumXY + math.log(ranks[index]) * math.log(counts[index])
        sumX2 = sumX2 + math.log(ranks[index])^2
        sumY2 = sumY2 + math.log(counts[index])^2                
      end
              
      if numberOfRanks * sumX2 - sumX * sumX == 0 then
        slope = 0
      else
        slope = ((numberOfRanks * sumXY - sumX * sumY) / (numberOfRanks * sumX2 - sumX * sumX))
      end
               
      if math.sqrt((numberOfRanks * sumX2 - sumX * sumX) * (numberOfRanks * sumY2 - sumY * sumY)) == 0 then
        r2 = 0
      else
        r  = (numberOfRanks * sumXY - sumX * sumY) / math.sqrt((numberOfRanks * sumX2 - sumX * sumX) * (numberOfRanks * sumY2 - sumY * sumY))
        r2 = r * r
      end 
        
    end 
  end     
  local yint = sumY - slope * sumX / #ranks
        
  return slope, r2, yint
end

function exponentialPDF(x,a)  
  return 1/a * math.exp(-x/a)
end

function expInterpolate(size)
  local res = {}
  for i=1,5,(5/size) do  
    local val = round(exponentialPDF(i,1)*100)  
    for j = 1,val do
      table.insert(res,round(i*(size/5)))
    end
  end
  return res
end

function sigmoid(x) 
  return 1 / (1 + math.exp(-x))
end

function Reverse (arr)
  local i, j = 1, #arr

  while i < j do
    arr[i], arr[j] = arr[j], arr[i]

    i = i + 1
    j = j - 1
  end
end

function blockFunApply(fun) 

  local patCon, _,  _, sel = getPatFun(true)  
  
  if sel == nil then return end
  
  local tr_end,tr_start
  if MW.edit_RN_phrase then
    tr_end,tr_start = 1,1
  else
    tr_end   = sel.end_track
    tr_start = sel.start_track 
  end
  
  for tr = tr_start, tr_end do
    for ln = sel.start_line, sel.end_line do
      local n_col_tb = {}
      
      for n_col=1,12 do      
        local pat  = patCon(ln,n_col,tr) 
        if pat.is_selected then       
          fun(pat, {ln = ln, col = n_col, tr = tr})
        end
      end
    end    
  end
end

function tableMakeEqualLen(tb1, tb2) 
  
  if #tb1 > #tb2 then 
    for i=#tb1, #tb2+1, -1 do
      table.remove(tb1, i)
    end
  elseif #tb2 > #tb1 then 
    for i=#tb2, #tb1+1, -1 do
      table.remove(tb2, i)
    end
  end
end

function mathInterpolate(val, min, max) 
  return val*(max-min)+min
end

function Fibonacci(n)
  local function inner(m, a, b)
    if m == 0 then
      return a
    end
    return inner(m-1, b, a+b)
  end
  return inner(n, 0, 1)
end

function randomFib(div)
   local r = math.random(2)
   
   if r == 1 then
     r = -1
   else
     r = 1
   end
   
  return round((Fibonacci(math.random(99)) % 12) / div * r)
end

function readOnly (t)
  t.tb_len = #t
  local proxy = {}
  local mt = {
    __index = t,
    __newindex = function (t, k, v)
      MW.table_write_test = 'ro'
    end 
  }
  setmetatable(proxy, mt)
  return proxy
end

function reverseTb (arr)
  local i, j = 1, #arr

  while i < j do
    arr[i], arr[j] = arr[j], arr[i]

    i = i + 1
    j = j - 1
  end
end

function wrap( t, l ) 
  for i = 1, l do
    table.insert( t, 1, table.remove( t, #t ) )
  end
end

function pearson2tbs(tb1,tb2)
  
  if #tb1 ~= #tb2 then return end

  local x1, y1 = 0, 0

  for i=1,#tb1 do
    x1 = x1 + tb1[i]
    y1 = y1 + tb2[i]
  end 
   
  x1 = x1 / #tb1
  y1 = y1 / #tb2
  
  local x2, y2, xy = 0, 0, 0
  
  for i=1,#tb1 do
    local tx = tb1[i] - x1
    local ty = tb2[i] - y1
    
    xy = xy + tx * ty  
    x2 = x2 + tx * tx
    y2 = y2 + ty * ty
  end

  return xy / math.sqrt(x2) / math.sqrt(y2)
end

function round(x)
  return x>=0 and math.floor(x+0.5) or math.ceil(x-0.5)
end

function sortKeys2(tb,res,max,key)
  if next(tb) ~= nil then
    for k,v in pairs(tb) do
      if v > max then
        max = v
        key = k
      end
    end    
    
    if key then 
      table.insert(res,key)
    end
    
    tb[key] = nil
    return sortKeys2(tb,res,0,nil)
  end
  
  return res 
end

function sortKeys(tb)
  local tb2 = copy(tb)
  local res = {}
  return sortKeys2(tb2,res,0,nil)
end

function permutation(a, n, cb)
  if n == 0 then
    cb(a)
  else
    for i = 1, n do
      a[i], a[n] = a[n], a[i]
      permutation(a, n - 1, cb)
      a[i], a[n] = a[n], a[i]
    end
  end
end
 
function mapAll (fcn, tab, idx, ...)
  if idx < 1 then
    fcn(...)
  else
    local t = tab[idx]
    for i = 1, #t do mapAll(fcn, tab, idx-1, t[i], ...) end
  end
end

function createFillTb(len,el)
  local tb = {}
  
  for i=1,len do
    tb[#tb+1] = el
  end
  return tb
end

function tableConcat(t1,t2)
  for i=1,#t2 do
    t1[#t1+1] = t2[i]
  end
end

function tableConcatRetNew(t1,t2)
  local tb = copy(t1)
  
  for i=1,#t2 do
    tb[#tb+1] = t2[i]
  end
  return tb
end

function tbMaxIx(tb)
  local max = 0
  local ix
  
  for k,v in ipairs(tb) do
    if v > max then
      max = v
      ix = k
    end
  end
  return ix
end

function copy(obj, seen)
  if type(obj) ~= 'table' then return obj end
  if seen and seen[obj] then return seen[obj] end
  local s = seen or {}
  local res = setmetatable({}, getmetatable(obj))
  s[obj] = res
  for k, v in pairs(obj) do res[copy(k, s)] = copy(v, s) end
  return res
end

function copy2(tb)
  local hash = {}
  local res = {}

  for _,v in ipairs(tb) do
     if (not hash[v]) then
       res[#res+1] = v 
       hash[v] = true
     end
  end
  return res
end

function _deepcopy(o, tables)
 
  if type(o) ~= 'table' then
    return o
  end
 
  if tables[o] ~= nil then
    return tables[o]
  end
 
  local new_o = {}
  tables[o] = new_o
 
  for k, v in next, o, nil do
    local new_k = _deepcopy(k, tables)
    local new_v = _deepcopy(v, tables)
    new_o[new_k] = new_v
  end
 
  return new_o
end
 
function deepcopy(o)
  return _deepcopy(o, {})
end

function _deepFlatten(o, tables)
 
  if type(o) ~= 'table' then
    return o
  end
 
  if tables[o] ~= nil then
    return tables[o]
  end
 
  local new_o = {}
 
  for k, v in next, o, nil do
    local new_k = _deepFlatten(k, tables)
    local new_v = _deepFlatten(v, tables)
    
    if type(new_v) == 'table' then
      tableConcat(new_o,new_v)
    else
      table.insert(new_o,new_v)
    end
  end
 
  return new_o
end
 
function deepFlatten(o)
  return _deepFlatten(o, {})
end

function unique(tb)
  local res = {}
  local unique = {}
  
  for _,v in pairs(tb) do
    if not unique[v] then
      unique[v] = {}
      table.insert(res,v) 
    end
  end
  return res
end

function uniqueKey(tb,key) 
  local res = {}
  
  local check   
  check = function(v2) 
    for k,v in pairs(res) do 
      if v[key] == v2[key] then
       return false
      end
    end
    return true      
  end
    
  for _,v in pairs(tb) do 
    if check(v) then 
      table.insert(res,v) 
    end
  end
  return res
end

function mean(t)
  local sum = 0
  local count= 0

  for k,v in pairs(t) do
    if type(v) == 'number' then
      sum = sum + v
      count = count + 1
    end
  end

  return (sum / count)
end

function median(t)
  local temp={}

  for k,v in pairs(t) do
    if type(v) == 'number' then
      table.insert( temp, v )
    end
  end

  table.sort( temp )

  if math.fmod(#temp,2) == 0 then
    return ( temp[#temp/2] + temp[(#temp/2)+1] ) / 2
  else
    return temp[math.ceil(#temp/2)]
  end
end

function randomKey(tb)

  local keys = {}
  for v in pairs(tb) do
    table.insert(keys,v)
  end  
  return keys[ math.random(#keys) ]
end

function serializeHelper(val) 
  local num = tonumber(val)
    
  if num then   
    return val   
  elseif  val == 'true' then  return 'true'
 
  elseif  val == 'false' then return 'false'
    
  elseif string.sub(val,1,1) ~= '{' then
    return '"' .. val ..  '"'
  else
    return val
  end
end

function serialize(o)  
   if type(o) == 'table' then
      local s = '{ ' 
      for k,v in pairs(o) do

         if type(k) ~= 'number' then 
           k = '"'..k..'"'            
         end
           
         s = s .. '['..k..'] = ' .. serializeHelper(serialize(v)) .. ','
      end
      return s .. '} '
   else
      return tostring(o)
   end
end

function nearestValue(tb, number) 
  local smallestSoFar, smallestIndex
 
  for i, y in ipairs(tb) do
    if not smallestSoFar or (math.abs(number-y) < smallestSoFar) then
      smallestSoFar = math.abs(number-y)
      smallestIndex = i
    end
  end
  return smallestIndex, tb[smallestIndex]
end

function splitDl (str, dl)  

  local pattern = '([^' .. dl .. ']+)'
  local tb = {}
  
  for w in str:gmatch(pattern) do 
    table.insert(tb,w) 
  end
   
  return tb
end

function splitPunRetNum(str)
  local tb = {}
  for i in string.gmatch(str, "%P+") do
    table.insert(tb,tonumber(i))
  end
  return tb
end

function sprint(msg)
  renoise.app():show_status(msg)
end

function lenTb (tb)
  
  if tb == nil then return 0 end
  
  local i = 0
  local tb_ret = {}
  for v in pairs(tb) do
    i=i+1
    table.insert(tb_ret, v)
  end
  return i, tb_ret
end 
   
function nextLen(ix,tb)
  for i=ix,#tb do
     if tb[i] > 0 then 
       return tb[i] 
     end
  end
  
  return 0
end
   
function nearestValue(table, number)
    local smallestSoFar, smallestIndex
    for i, y in ipairs(table) do
        if not smallestSoFar or (math.abs(number-y) < smallestSoFar) then
            smallestSoFar = math.abs(number-y)
            smallestIndex = i
        end
    end
    return smallestIndex, table[smallestIndex]
end

function fillNote(note, note_col, ln)

  if MW.edit_RN_phrase then   
  
    local inst_no = renoise.song().selected_instrument_index   
    local phrase = renoise.song().instruments[inst_no].phrases[MW.phr.no]
  
    if (phrase == nil) then
      renoise.song().instruments[inst_no]:insert_phrase_at(MW.phr.no)
      phrase = renoise.song().instruments[inst_no].phrases[MW.phr.no]
    end
    
    if (note == 121) then
      phrase:line(ln):note_column(note_col).volume_string  = ""
    end
    phrase:line(ln):note_column(note_col).note_value = note  
    
  else  
    local pattern =  curPattern ()
    local tr = renoise.song().selected_track_index 
    
    pattern:track(tr):line(ln):note_column(note_col).note_value = note
    
    if (note == 121) then
      pattern:track(tr):line(ln):note_column(note_col).instrument_string  = ""
      pattern:track(tr):line(ln):note_column(note_col).volume_string  = ""
    else
      pattern:track(tr):line(ln):note_column(note_col).instrument_value = renoise.song().selected_instrument_index - 1
    end
  end
end

function num2note(num)
  local oct = math.floor(num / 12)
  local note = num % 12
  note = note + 1
 
  return Scales.keys[note] .. Scales.keys_dash[note] .. oct
end

function num2noteOnly(num)
  local oct = math.floor(num / 12)
  local note = num % 12
  note = note + 1
 
  return Scales.keys[note]
end

function getNote(note_col, ln)
  local note 
  
  if MW.edit_RN_phrase then   
  
    local inst_no = renoise.song().selected_instrument_index   
    local phrase = renoise.song().instruments[inst_no].phrases[MW.phr.no]
  
    if (phrase == nil) then
      renoise.song().instruments[inst_no]:insert_phrase_at(MW.phr.no)
      phrase = renoise.song().instruments[inst_no].phrases[MW.phr.no]
    end
    
    note = phrase:line(ln):note_column(note_col).note_value  
    
  else  
    local pattern =  curPattern ()
    local tr = renoise.song().selected_track_index 
    
    note = pattern:track(tr):line(ln):note_column(note_col).note_value
  end
  
  return note
end

function pinknoise (noise,amplitude) 

  local a = {} 
  
  local max_s = 0x1f
  
  local b0 = 0.0
  local b1 = 0.0
  local b2 = 0.0
  local b3 = 0.0
  local b4 = 0.0
  local b5 = 0.0
  local b6 = 0.0
  
  local s = 0
  
  return function ()
    s=s+1
    
    if (s > max_s) then
      s = 0
    end
    
    if noise == 0 then
    
      a[s] = amplitude * 2 * (math.random() - 0.5)
      
    elseif noise == 1 then
    
      local white = (amplitude * 0.25) * 2 * (math.random() - 0.5)
      
      b0 = 0.99886 * b0 + white * 0.0555179;
      b1 = 0.99332 * b1 + white * 0.0750759;
      b2 = 0.96900 * b2 + white * 0.1538520;
      b3 = 0.86650 * b3 + white * 0.3104856;
      b4 = 0.55000 * b4 + white * 0.5329522;
      b5 = -0.7616 * b5 - white * 0.0168980;
      b6 = white * 0.115926;
      a[s] = b0 + b1 + b2 + b3 + b4 + b5 + b6 + white * 0.5362;
    end
    
    return a[s]
  end
end

function setLPB()
  local len
  local tb = {}
  
  if MW.edit_RN_phrase then
    local inst_no = renoise.song().selected_instrument_index
    
    if (renoise.song().instruments[inst_no].phrases[1] ~= nil) then
      len = renoise.song().instruments[inst_no].phrases[1].number_of_lines
      MW.phr.lpb = renoise.song().instruments[inst_no].phrases[1].lpb
    end
  else
    len = renoise.song().selected_pattern.number_of_lines
    MW.phr.lpb = renoise.song().transport.lpb
  end
   
  if (len ~= nil and MW.phr.lpb ~= nil) then  
    for i=1,len,MW.phr.lpb do
      tb[#tb+1] = i
    end
  end
  
  return tb
end

function getPatLn(line)
  if MW.edit_RN_phrase then    
    local phrase = renoise.song().instruments[inst_no].phrases[1]
    
    return phrase:line(line)
  else
    local tr      = renoise.song().selected_track_index  
    local pattern = curPattern()
    
    return pattern:track(tr):line(line)
  end
end

function getPatFun(block, ...) -- arg[1] patter ix, [2] track

  local st, sel, pat_len 
  local inst_no = renoise.song().selected_instrument_index
    
  if MW.edit_RN_phrase then       
       
    if (renoise.song().instruments[inst_no].phrases[1] == nil) then
      return false
     end   
     
    local phrase = renoise.song().instruments[inst_no].phrases[1]
    
    local fun = 
      function (line, note_col)
        return phrase:line(line):note_column(note_col)
      end
   
    if block then 
      sel = renoise.song().selection_in_phrase   
        
        if (sel ~= nil) then    

          local len = sel.end_line - sel.start_line
    
          st = sel.start_line  - 1  
          pat_len = sel.end_line
          
        else  --------- empty block
          st = 1
          pat_len = renoise.song().instruments[inst_no].phrases[1].number_of_lines 
        end
         
      return fun, pat_len, st, sel   
      
    else ------ block off
      local pat_len = renoise.song().instruments[inst_no].phrases[1].number_of_lines    
      return fun, pat_len, 1    
    end
  
  else -------------- pattern

    local pattern    
    if arg[1] then 
      pattern = getPattern (arg[1]) 
    else
      pattern =  curPattern()
    end     
   
    local tr     
    if arg[2] then
      tr = arg[2]
    else
      tr = renoise.song().selected_track_index  
    end 
     
    local fun = 
      function (line, note_col, ...)
        if arg[1] then
          tr = arg[1]
        end
        return  pattern:track(tr):line(line):note_column(note_col)
      end 

    if block then 
        
        sel = renoise.song().selection_in_pattern   
        
        if (sel ~= nil) then    

          local len = sel.end_line - sel.start_line
    
          st = sel.start_line  - 1  
          pat_len = sel.end_line
          
        else --------- empty block
           st = 1
           if arg[1] then 
             pat_len = pattern.number_of_lines
           else
             pat_len = renoise.song().selected_pattern.number_of_lines           
          end
        end

      return fun, pat_len, st, sel 
      
    else ------ block off
    
      if arg[1] then 
        pat_len = pattern.number_of_lines
      else
        pat_len = renoise.song().selected_pattern.number_of_lines           
      end
      return fun, pat_len, 1
      
    end
  end 
   
  return false
end

function growChrdTbs()          
  table.insert(UI.chord_scales,2)
  table.insert(UI.chord_keys ,1)
  table.insert(UI.chord_forms,22)
end

function growSeqTbs(len_tb, note_tb, note_off_tb, octave_tb, velo_tb)
  table.insert(len_tb,0)
  table.insert(note_tb,0)
  table.insert(note_off_tb,0)
  table.insert(octave_tb,4)
  table.insert(velo_tb,127)
end

function getPhaseLPB()
  local inst_no = renoise.song().selected_instrument_index
  local phrase = renoise.song().instruments[inst_no].phrases[MW.phr.no]
    
  if (phrase == nil) then 
    return 4
  end
    
  return renoise.song().instruments[inst_no].phrases[MW.phr.no].lpb
end

function noteLengthEmpty (tb) 
  if tb == nil then 
    return true
  end
  
  for i,v in ipairs(tb) do
    if (v > 0) then return false end 
  end
  return true
end

function emptyPoly()

  for i=1,12 do
    assignNoteTbs(i)   
    
    if lenTb(Nptr) == 0 then
      Nptr = Notes.new(0) 
    elseif not noteLengthEmpty(Nptr.len) then        
      return false  
    end
  end
  return true
end          

function emptyPolyTb(ptr)

  for i=1,12 do      
    if not noteLengthEmpty(ptr[i].len) then        
      return false  
    end
  end
  return true
end 

function gaussian (mean, variance)
  return  math.sqrt(-2 * variance * math.log(math.random())) *
            math.cos(2 * math.pi * math.random()) + mean
end
 
function fillTb (len,n)
  local tb = {}
  
  for i=1,len do
    tb[i] = n
  end
  return tb
end

function remExt2End(file)
  for i=#file,1,-1 do
    if (string.sub(file, i, i) == ".") then
      return string.sub(file, 1, i-1)
    end
  end
end

function remExtEnd(files)  
  local tb = {}
  for i=1,#files do
    tb[i] = remExt2End(files[i])
  end
  return tb
end 

function remExt2(file)
  for i=1,#file do
    if (string.sub(file, i, i) == ".") then
      return string.sub(file, 1, i-1)
    end
  end
end

function remExt(files)  
  local tb = {}
  for i=1,#files do
    tb[i] = remExt2(files[i])
  end
  return tb
end  

function probTb (tb)
  local tbret = {}
  
  for i=1,#tb do    
    if (tb[i][2] > 0) then
      for j=1,tb[i][2] do
        tbret[#tbret+1] = tb[i][1]  
      end
    end    
  end

  return tbret
end

function copy (tb)
  local tbret = {}
  for k,v in pairs(tb) do
    tbret[k] = v
  end
  return tbret
end   

function copyDstSrc(dst,src)  
  for k,v in pairs(src) do
    dst[k] = v
  end
end   

function copyDeep(t)
    if type(t) ~= "table" then return t end
    local meta = getmetatable(t)
    local target = {}
    for k, v in pairs(t) do
        if type(v) == "table" then
            target[k] = clone(v)
        else
            target[k] = v
        end
    end
    setmetatable(target, meta)
    return target
end

function formStr (tb)
  local tb_str = {}
  
  for k,v in ipairs(tb) do
    tb_str[k] = table.concat(v,'-')  
  end  
  return tb_str
end

function tableComp(o1, o2)
  if #o1 ~= #o2 or #o1 == 0 or #o2 == 0 then
    return false
  else 
    for i=1,#o1 do  
      if tostring(o1[i]) ~= tostring(o2[i]) then
        return false
      end
    end    
  end
  return true
end

function tablesCompRet (tbs,tb) 
  
  if type(tbs) == 'string' then
    tbs = {tbs}
  end 
  
  local len = lenTb(tbs) 

  for i=1,#tb do
    local tb2 = {unpack(tb,i,i+len-1)} 
    local tb3 = {unpack(tb,i+len,#tb)} 

    if tableComp(tbs, tb2) then
      return tb3            
    end
  end
  return {}
end

function remEl(elm, tb)
  local tb2 = {}
  
  for i=1,#tb do
    if (tb[i] ~= elm) then
      tb2[#tb2+1] = tb[i]     
    end
  end
  
  return tb2
end

function member(elm, tb)
  for k , v in pairs(tb) do
    if v == elm then
      return true, k
    end
  end  
  return false, nil
end  

function compareTbs(tb1,tb2)
  local c = 0
  
  for _,v1 in ipairs(tb1) do
    for _,v2 in ipairs(tb2) do  
      if v1 == v2 then
        c = c + 1
      end
    end
  end 
  return c
end

function removeElmFromTb(elm,tb)

  local exist, key = member(elm,tb)
  if exist then
    tb[key] = nil
  end   
end

function clamp (val, l,h)
  if (val < l) then
    return l
  elseif (val > h) then
    return h
  end  
  return val  
end

function clamp2 (val, l,h)
  if (val < l) then
    return val + 12
  elseif (val > h) then
    return val - 12
  end  
  return val  
end

function clampAround (val, l,h)
  if val < l then
    return h - math.abs(val)
  elseif val > h then
    return l + val-h-1
  end  
  return val  
end

function clampAround2 (val, l,h)
  if val < l then
    return h - math.abs(l - val)
  elseif val > h then
    return l + val-h-1
  end  
  return val  
end

function clampOct (val, root_oct, max_oct)
  
  if max_oct == 0 then return val end

  local l = root_oct * 12  
  local h = (root_oct+max_oct) * 12
  
  if val < l or val >= h then  
    return val%(max_oct*12) + root_oct*12   
  end 
  return val  
end

function tbAvg (tb)
  local sum = 0
  for i=1,#tb do
    sum = sum + tb[i]    
  end
  return round(sum/#tb)
end

function tbNotZero (tb)
  for i=1,#tb do
    if (tb[i] ~= 0) then return true end
  end
  
  return false
end

function randNeg()
  local neg     
  
  if math.random(2) == 1 then 
    neg = 1
  else
    neg = -1
  end
  return neg 
end

function elemensTb(tb)
  local sum = 0
  for i=1,#tb do
    if (tb[i] ~= nil and tb[i] > 0) then
      sum = sum + 1
    end
  end
  return sum
end 

function sumTb(tb)  
  if type(tb) == 'number' then return tb end

  local sum = 0
  
  for k,v in pairs(tb) do 
    sum = sum + v    
  end
  return sum  
end

function cumSum(tb)
  local cumsum = {}
  local sum = 0
  
  for k,v in ipairs(tb) do 
    sum = sum + v 
    cumsum[k] = sum    
  end
  return cumsum  
end

function normalize(tb)
  local res = {}
  local mul = 1 / sumTb(tb)
  
  for k,v in ipairs (tb) do
    res[k] = v * mul      
  end
  return res
end

function normalize2(tb) 
  local max = math.max(unpack(tb))
  
  return map(tb, function(v) if v > 0 then return v / max end return v end)
end

function oct0 (note)
  local x = math.floor(note / 12) 
  return note - (x * 12) + 0 , x
end

function tbOct0(tb)
  local tb2 = {}
  local note, oct = oct0(tb[1]) 
  for i=1,#tb do
    tb2[i] = tb[i] - oct*12  
  end
  return tb2, oct
end

function tbOct0all(tb)
  local res = {}

  for _,v in ipairs(tb) do
    res[#res+1] =  oct0(v) 
  end  
  
  table.sort(res)
  
  return copy(res)
end

function addToTb (tb,n)
  local tb2 = {}
  
  for i=1,#tb do
    if  (type(tb[i]) == "table") then
      tb2[i] =  addToTb (tb[i],n)
    else
      tb2[i] = tb[i] + n
    end
  end 
  return tb2
end

function retIx (tb,val)
  local n = 0
  for i=1,#tb do
    if (tb[i] == val) then return i end  
  end
  return n
end

function retKeyIx(tb,val)
  if tb then 
    for k,v in pairs(tb) do
      if tostring(val) == tostring(v) then   
        return k,v
      end
    end
  end  
  return nil,nil
end

function retNonZeroIx (tb)
  local res ={}
  
  for k,v in pairs(tb) do
    if v > 0 then 
      table.insert(res,k)
    end  
  end
  return res
end

function fileExists(name)
   local f=io.open(name,"r")
   if f~=nil then 
     io.close(f) return true 
   else
     return false    
   end
end

function nameCounter (str, n)

  local is_number = false
  local numLen = 0
  local str_num = nil
  local char = nil
  local dialog
  
  local len = string.len(str)

  for i=1,len do  
    char = string.sub(str, -i, -i)
    if (tonumber(char) ~= nil) then
      numLen = i
    elseif (numLen > 0) then
      break   
    elseif (i == 1) then
      break
    end      
  end
   
  if (numLen > 0) then
    local name = string.sub(str, 1, len - numLen)
    local number = string.sub(str, len - numLen + 1, len )

    local number = tonumber(number) 
    
    if (number == 0 and n == -1) then return str end
    
    local numberOut = number + n .. ""

    local numberOutLen = string.len(numberOut)
      
    if (numberOutLen < numLen) then
      for i=numberOutLen,numLen-1 do
        numberOut = "0" .. numberOut
      end   
    end  
    str = name .. numberOut 
  else  
    str = str .. "000"        
  end 
  return str
end

function stretchTb (tb, len)
  local tb_ret = {}
  
  local n = #tb / len

  local sum = n
  local ix = 1
  
  while sum <= #tb do
    local ix2 = math.floor(sum-n) + 1 
    if (ix2 > #tb) then ix2 = #tb end 
    
    tb_ret[ix] = tb[ix2]

    ix = ix + 1
    sum = sum + n    
  end    
  return tb_ret
end

function blockStEn()
  local sel = renoise.song().selection_in_pattern   
        
  if sel and MW.block_editing then    

    local len = sel.end_line - sel.start_line
     
    return sel.start_line, sel.end_line
  end
  return 0, renoise.song().selected_pattern.number_of_lines  
end

function buildProbTb(tb,n)
  local res = {}
  
  for k,v in pairs(tb) do
    if v > 0 then
      for i=1,v do
        table.insert(res,k+n)
      end
    end
  end
  return res
end

function randomExceptHelper (exept_tb, rand)
  local res = false
  
  for _,v in ipairs(exept_tb) do
    if v == rand then 
      return true 
    end
  end
  return res  
end

function randomExcept(n,except_tb)
  local rand = math.random(n)
  
  while randomExceptHelper (except_tb, rand) do
    rand = math.random(n)
  end
   
  return rand 
end

function entropyCross(pk,qk) 
  local s = {}
  for i=1,#pk do 
    if pk[i] == 0 or qk[i] == 0 then
      s[i] = 0
    else
      s[i] = pk[i] * math.log(qk[i])  
    end   
  end
  return -sumTb(s)
end

function entropy(pk)
  local s = {}
  for i,v in ipairs(pk) do
    if v > 0 then
      s[i] = v * math.log(v)  
    else
      s[i] = 0
    end
  end
  return -sumTb(s)
end

function getTrackFxH(block,...) 
  local tb = {}
  local pattern    
  local all_cols = {1,2,3,4,5,6,7,8}
  
  if arg[1] then
    pattern = getPattern(arg[1]) 
  else
    pattern = curPattern()
  end
  
  for pat,ln,col,tr,sel in patternMW(all_cols,'pat_fx',block,pattern) do
    if pat.number_value == 17 then      
      table.insert(tb,ln)
    end
  end
  return unique(tb)
end
