
local ph  = MW.phr.no
local col = MW.phr.col 


Mt.notes.__index = function (table, key, fun)
  table[key] = Notes.new()
  return table[key]
end


Mt.chords.__index = function (table, key, fun)
  table = Chords.new()
  return table
end


Mt.phr.__index = function (table, key, fun) 
  table[key]        = {}
  table[key].name   = 'Phrase ' .. key
  table[key].notes  = {}
  table[key].chords = {}
  
  setmetatable(table[key].notes, Mt.notes)  
  setmetatable(table[key].chords, Mt.chords)  
  
  return table[key]
end

setmetatable(MW.phr, Mt.phr)    


function initTonnetzFormsMtb()
  UI.form_tonnetz = {}
  
  Mt.tonnetz_form.__index = function (table, key)
    table[key] = 12
    return table[key]
  end

  setmetatable(UI.form_tonnetz, Mt.tonnetz_form) 
end


function Notes.grow(o)      
  table.insert(o.fn,  0)
  table.insert(o.off, 0)
  table.insert(o.oct, 4)
  table.insert(o.len, 0)  
  table.insert(o.vel, 127)
  table.insert(o.note,0)
  table.insert(o.alen,0)
  table.insert(o.line,0)
end  


function Notes.concat(tb1,tb2)
  for k,len in ipairs(tb2.len) do
    if len > 0 then
      table.insert(tb1.len,  len)
      table.insert(tb1.oct,  tb2.oct[k])
      table.insert(tb1.off,  tb2.off[k])
      table.insert(tb1.note, tb2.note[k])      
    end
  end
end


function Notes.new(...) 

  local o = {  
    fx    = {},
    fn    = {},  
    off   = {},       
    oct   = {},    
    len   = {},    
    vel   = {},
    alen  = {},    
    note  = {},
    line  = {},  
    oct_lowest  = 5,
    oct_highest = 5,
  } 
  
  --if MW.started then  
  --  o.phr_len = renoise.song().selected_pattern.number_of_lines   
  --else
  --  o.phr_len = 64
  --end 
  
  o.phr_len = 0
  
  if arg[1] then
    o.phr_len = arg[1]
  end   
  
  for i=1,o.phr_len do
    o.fn  [i] = 0
    o.vel [i] = 127
    o.off [i] = 0 
    o.oct [i] = 4
    o.len [i] = 0
    o.alen[i] = 0
    o.note[i] = 0
    o.line[i] = 0
  end  
  return o
end


function Notes.newPtr(o)
  o = {  
    fn    = {},  
    off   = {},       
    oct   = {},    
    len   = {},    
    vel   = {},
    alen  = {},    
    note  = {},
    line  = {},
    oct_lowest  = 5,
    oct_highest = 5,  
  } 
end


function Notes.newPoly(...) 
  local o = {}
  
  for i=1,12 do
    if arg[1] then
      o[i] = Notes.new(arg[1])
    else
      o[i] = Notes.new()
    end
  end  
  return o
end


function Chords.grow(o)     
  table.insert(o.arp.oct,  1)    
  table.insert(o.arp.off,  1)
  table.insert(o.arp.form, 1)    
  table.insert(o.arp.rand, true)

  table.insert(o.off,  0)
  table.insert(o.vel,  127)
  table.insert(o.len,  0)
  table.insert(o.alen,  0)
  table.insert(o.line, 0)  
  table.insert(o.notes,{60})
  table.insert(o.rand, true) 
end  


function Chords.new(...) 

  local o = {  
    arp = {
      oct  = {},  
      off  = {},
      form = {},      
      rand = {},
      },
    
    fx    = {},
    len   = {},
    vel   = {},
    off   = {}, 
    alen  = {},
    rand  = {},  
    line  = {}, 
    notes = {},
    
    oct_lowest  = 5,
    oct_highest = 5,
  } 
    
  --if MW.started then  
  --  o.phr_len = renoise.song().selected_pattern.number_of_lines   
  --else
  --  o.phr_len = 64
  --end 
  
  o.phr_len = 0
  
  if arg[1] then
    o.phr_len = arg[1]
  end  

  for i=1,o.phr_len do
    o.arp.oct [i] = 1    
    o.arp.off [i] = 1
    o.arp.form[i] = 1    
    o.arp.rand[i] = true
    
    o.len   [i] = 0 
    o.vel   [i] = 127
    o.off   [i] = 0 
    o.rand  [i] = true
    o.line  [i] = 0
    o.notes [i] = {60}
  end  
  return o
end


function Chords.newPtr(o, phr_len) 

  o.arp.oct  = {}  
  o.arp.off  = {}
  o.arp.form = {}      
  o.arp.rand = {}
      
    
  o.len   = {}
  o.vel   = {}
  o.off   = {} 
  o.alen  = {}
  o.rand  = {}  
  o.line  = {} 
  o.notes = {}
  
  o.oct_lowest  = 5
  o.oct_highest = 5
  
  for i=1,phr_len do
    o.arp.oct [i] = 1    
    o.arp.off [i] = 1
    o.arp.form[i] = 1    
    o.arp.rand[i] = true
    
    o.len   [i] = 0 
    o.vel   [i] = 127
    o.off   [i] = 0 
    o.rand  [i] = true
    o.line  [i] = 0
    o.notes [i] = {60}
   -- o.form_tonnetz[i] = 12
  end   
end

function assignNoteTbs(col)
  local ph = MW.phr.no    
  
  Nptr = MW.phr[ph].notes[col]  
end


function assignPolyTbs()
  local ph = MW.phr.no    
    
  POLYptr = MW.phr[ph].notes  
end


function assignChordTbs()
  local ph = MW.phr.no    
  
  CHptr = MW.phr[ph].chords
end


function duplicateChordNotes() 
  local col = MW.phr.col
  local ph  = MW.phr.no
  
  initDupChords()  
  
  DCHptr.off   = copy(CHptr.off) 
  DCHptr.len   = copy(CHptr.len) 
  DCHptr.notes = {}
  
  for k,notes in ipairs(CHptr.notes) do 
    DCHptr.notes[k] = notes 
  end
end


function duplicateNotes()
  local col = MW.phr.col
  local ph  = MW.phr.no

  initDupNotes()
    
  local d = MW.dup[ph].notes[col]
  local s = MW.phr[ph].notes[col]
        
  Dptr.len  = copy(s.len)
  Dptr.oct  = copy(s.oct)
  Dptr.off  = copy(s.off)
  Dptr.note = copy(s.note)
end


function duplicatePoly()
  local col = MW.phr.col
  local ph  = MW.phr.no  

  initDupPoly()
    
  for i=1, 12 do
    MW.dup[ph].notes[i] = {}
      
    local d = MW.dup[ph].notes[i]
    local s = MW.phr[ph].notes[i]
      
    d.len  = copy(s.len)
    d.oct  = copy(s.oct) 
    d.off  = copy(s.oct)        
    d.note = copy(s.note)
  end
end


function duplicateNotesRand() 
  local col = MW.phr.col
  local ph  = MW.phr.no

  MW.dupR = {}
  
  MW.dupR.oct = {}
  MW.dupR.off = {}
  MW.dupR.len = {}
  MW.dupR.note = {}  
    
  if MW.phr.editor_poly then ------------------------------------
           
    for i=1,12 do     
      Notes.concat(MW.dupR,POLYptr[i])             
    end    
  
  MW.dupR.pat_len = POLYptr[1].pat_len
  
  elseif MW.phr.editor_chords then ------------------------- 
    for k,notes in ipairs(CHptr.notes) do    
          
      for _,note in ipairs(notes) do
        table.insert(MW.dupR.off, CHptr.off[k])
        table.insert(MW.dupR.len, CHptr.len[k])
        
        local nn,oct = oct0(note)        
        table.insert(MW.dupR.oct, oct)
        table.insert(MW.dupR.note, nn)
      end
    end
    
    MW.dupR.pat_len = CHptr.pat_len
  else
    MW.dupR.oct  = copy(Nptr.oct)
    MW.dupR.off  = copy(Nptr.off)
    MW.dupR.len  = copy(Nptr.len)
    MW.dupR.note = copy(Nptr.note)
     
    MW.dupR.pat_len = Nptr.pat_len
  end
  
  DRptr = MW.dupR
end


function initNotes(...) 

  local ph = MW.phr.no
  local col = MW.phr.col
  
  local pat_len = 0 
  if arg[1] then
    pat_len = arg[1] 
  end
  
  MW.phr[ph].notes[col] = Notes.new(pat_len)   
  
  assignNoteTbs(col)
end


function initPoly(...)
  local ph  = MW.phr.no
  local col = MW.phr.col
  
  local pat_len = 0 
  if arg[1] then  
    pat_len = arg[1] 
  end
  
  for i=1,12 do   
    MW.phr[ph].notes[i] = Notes.new(pat_len)       
    
    MW.timer_poly_pos[i] = {}    
  end
  
  assignPolyTbs()
end


function initChords(...)  
  local ph = MW.phr.no
  
  local pat_len = 0 
  if arg[1] then
    pat_len = arg[1] 
  end
  
  MW.phr[ph].chords = Chords.new(pat_len)     
  assignChordTbs()
end


function initDupNotes()    

  local ph = MW.phr.no
  local col = MW.phr.col
  
  if MW.dup[ph] == nil then
    MW.dup[ph] = {}
    MW.dup[ph].notes = {}
  elseif MW.dup[ph].notes == nil then
    MW.dup[ph].notes = {}
  end
  
  MW.dup[ph].notes[col] = {} 
  
  Dptr = MW.dup[ph].notes[col]
end


function initDupPoly()
  local ph  = MW.phr.no
  local col = MW.phr.col
  
  if MW.dup[ph] == nil then
    MW.dup[ph] = {}
    MW.dup[ph].notes = {} 
  elseif MW.dup[ph].notes == nil then
    MW.dup[ph].notes = {} 
  end
  
  for i=1,12 do   
    MW.dup[ph].notes[i] = {}  
  end
  
  Dptr = MW.dup[ph].notes
end


function initDupChords()  
  local ph = MW.phr.no
  
  if MW.dup[ph] == nil then 
    MW.dup[ph] = {}
    MW.dup[ph].chords = {}
  elseif MW.dup[ph].chords == nil then  
    MW.dup[ph].chords = {}
  end
  
  MW.dup[ph].chords = {}  

  DCHptr = MW.dup[ph].chords
end


function init(...)
  if MW.phr.editor_chords then 
    initChords(...) 
  elseif MW.phr.editor_poly then
    initPoly(...)
  else
    initNotes(...)
  end
end

function initFillPhr(fil) 
  local col = MW.phr.col  
  
  if not fil then
    local editors = {}
    editors['Step']        = 1
    editors['Step N']      = 2
    editors['Step Poly']   = 3
    editors['Step Poly N'] = 4
    editors['VB']          = 5 
    editors['VB Poly']     = 6
    editors['Chords']      = 7
    editors['Arpeggiator'] = 8 
    editors['Tonnetz']     = 9 
    
    UI.editor_pu.value = editors[MW.phr.editor_name]       
  else 
    sprint('')
  end   

  if MW.phr.editor_chords then  
    setFillChords(fil)
  elseif MW.phr.editor_poly then
    setFillPoly(fil)
  else   
    setFillMono(fil)
  end
  
  updateSlots()
end


function getEventStr()
  local tb = {'mono', 'poly', 'chrd'}
  return tb[MW.phr.event_type]
end

