Load = {}
Save = {}

Load.max_pat_len     = 128 
Load.max_note_len    = 0 
Load.pattern_len_adj = true

local function getScanTranspose(phr_type) 
    
  if Scan.scale_key_on and Scan.typ == 1 then  
    local key   = UI.keys_pu.value
    local scale = UI.scales_pu.value
    
    return Scan.db_scale_key[phr_type][scale][key].transpose[UI.lib_pu.value]     
    
  elseif Scan.correlation_entropy_on and Scan.typ >= 2 then -- and #Lib.transpose > 0 then 
  
    return Lib.transpose[UI.lib_pu.value]
  end 
end


local function getPatLen(ptr) 
  if ptr.pat_len then
    return ptr.pat_len
  else
    return renoise.song().selected_pattern.number_of_lines   
  end 
end

Load.tagsTb = function(fn)  
  fn = MW.path .. fn
  if fileExists(fn) then
    local content = readFile(fn)
    local fun     = loadstring(content)
    
    local tb = fun()
    return tb
  end
end


Load.tags = function(fn)
  fn = fn .. '.tag'
  
  if fileExists(fn) then
    local content = readFile(fn) 
    local fun     = loadstring(content)
    
    Tag = fun()
    UI.tegs.text   = Tag.text
    UI.score.value = Tag.score
  else
    UI.tegs.text   = ''
    UI.score.value = 1
  end
end


Load.maxLen = function(ptr)
  if Load.max_note_len == 0 or #ptr <= 2 then return end
  
  local first = false
  
  for i=2,#ptr do  
    if first == false and ptr[i] > 0 then
      first = true
    end    
    
    if first and ptr[i] > Load.max_note_len then
      ptr[i] = Load.max_note_len
    end
  end
end 


Load.maxLenPoly = function(ptr, dst)
  if Load.max_note_len == 0 then return end

  local tb = {}
  
  for n,col in ipairs(ptr) do
    for i=1,#col.len do
      if col.len[i] > 0 then 
        table.insert(tb, {i    = i, 
                          col  = n,
                          off  = col.off[i], 
                          len  = col.line[i],
                          line = col.line[i]})        
      end
    end
  end  
  
  table.sort(tb, function(a,b) return a.line < b.line end)   

  local last_ln = tb[#tb].line
  
  for i=#tb,2,-1 do
    local ln = tb[i].line - tb[i-1].line
    
    if ln > 0 then
      if ln > Load.max_note_len then 
        last_ln = ln
        local sub = ln - Load.max_note_len  
        
        for j=i,#tb do
          tb[j].len  = tb[j].len - sub
          --tb[j].off  = tb[j].off - sub
          tb[j].line = tb[j].line - sub
        end
      end
    end
  end 
  
  
  local ln_cols = {}
  for i,v in ipairs(tb) do      
    if ln_cols[v.col] == nil then
      ln_cols[v.col] = {}
    end    
    table.insert(ln_cols[v.col], {line = v.line, i = v.i, off  = v.off})
  end  
      
  
  local lens = {}
  for key,col in ipairs(ln_cols) do
      
    local ii = col[1].i
    dst[key].off[ii] = col[1].off   
    dst[key].len[ii] = col[1].line    

    last_ln = col[1].line
    
    if #col > 1 then
      for i=2,#col do
        local ln = col[i].line - col[i-1].line
    
        if ln > 0 then
          last_ln = ln
        end
        
        ii = col[i].i  
        dst[key].off[ii] = col[i].off 
        dst[key].len[ii] = last_ln  
      end                         
    end   
  end           
  
  for col=1,12 do
    for i=1,#dst[col].len-1 do
      if dst[col].len[i] > 0 then
        local dur = dst[col].len[i+1] - dst[col].len[i]
        
        if dur > Load.max_note_len then
          dur = Load.max_note_len
        end
        
        dst[col].off[i] = dur 
      end
    end
    
    dst[col].off[#dst[col].off] = Load.max_note_len
  end                    
end 


Load.notes = function(fn, ...)
  
  Load.tags(fn)
  
  local str = ''
  
  local msg = {}
  msg.tr   = true   
  msg.pos  = true
  msg.load = true
  
  local load = {}
  load.phr  = Lib.editor == 2
  load.step = Lib.editor == 1 and MW.phr.event_type == 1
  load.poly = Lib.editor == 1 and MW.phr.event_type == 2
  load.chrd = Lib.editor == 1 and MW.phr.event_type == 3
  
  if arg[1] then
    msg = arg[1]
  end
  
  if arg[2] then
    load = arg[2]
  end
  
  if fileExists(fn) then
    if msg.load then str = 'Loading ' .. fn end

    local content = readFile(fn)
    local fun     = loadstring(content)             
    
      
    if load.phr  then
      MW.phr = fun()   
    elseif load.step then  --------------------- step mono
      local tb = fun()  
      
      Load.maxLen(tb.off)
      Load.maxLen(tb.len)
      
      Nptr.fx   = tb.fx             
      Nptr.fn   = tb.fn 
      Nptr.off  = tb.off       
      Nptr.oct  = tb.oct   
      Nptr.len  = tb.len 
      Nptr.vel  = tb.vel
      Nptr.alen = tb.alen    
      Nptr.note = tb.note 
      Nptr.line = tb.line 
      
      Nptr.pat_len = tb.pat_len 
      
      Nptr.oct_lowest  = tb.oct_lowest
      Nptr.oct_highest = tb.oct_highest
      
      if Nptr.pat_len == nil then  -------------------------------------- delete later 
        Nptr.pat_len = getPatLen(Nptr) 
      end
      
      if Nptr.pat_len > Load.max_pat_len then
        Nptr.pat_len = Load.max_pat_len
      end
      
      if Load.pattern_len_adj and Nptr.pat_len > renoise.song().selected_pattern.number_of_lines then
        renoise.song().selected_pattern.number_of_lines = Nptr.pat_len
      end
      
      if msg.pos then
        local line = 0
        for _,len in ipairs(Nptr.len) do
          line = line + len
        end  
        str = str .. '  |  ' .. 'Last note pattern position ' .. line .. '  |  ' .. 'pattern length ' .. Nptr.pat_len
      end
      
      local transpose = getScanTranspose('mono')
      
      if transpose then        
        if msg.tr then str = str .. '  |  ' .. 'Transpose mono phrase by ' .. transpose end
        
        for k,note in ipairs(Nptr.note) do 
                  
          note = clamp(note +  Nptr.oct[k] * 12 + transpose, 0, 119)
          
          local nn, oct = oct0(note)
          
          Nptr.oct[k]  = oct
          Nptr.note[k] = nn                      
        end
      end 
      
    elseif load.chrd then  ---------------------------------------------- chords
      local tb = fun()  
      
      Load.maxLen(tb.off)
      Load.maxLen(tb.len)
      
      tableMakeEqualLen(tb.len, tb.notes)
      
      CHptr.arp.oct  = tb.arp.oct     
      CHptr.arp.off  = tb.arp.off
      CHptr.arp.form = tb.arp.form    
      CHptr.arp.rand = tb.arp.rand
    
      CHptr.fx    = tb.fx
      CHptr.len   = tb.len 
      CHptr.vel   = tb.vel
      CHptr.off   = tb.off 
      CHptr.rand  = tb.rand
      CHptr.line  = tb.line
      CHptr.notes = tb.notes  
      
      CHptr.pat_len = tb.pat_len             
      
      CHptr.oct_lowest  = tb.oct_lowest
      CHptr.oct_highest = tb.oct_highest

      if CHptr.pat_len == nil then  -------------------------------------- delete later 
        CHptr.pat_len = getPatLen(CHptr) 
      end   
      
      if CHptr.pat_len > Load.max_pat_len then
        CHptr.pat_len = Load.max_pat_len
      end
      
      if Load.pattern_len_adj and CHptr.pat_len > renoise.song().selected_pattern.number_of_lines then
        renoise.song().selected_pattern.number_of_lines = CHptr.pat_len
      end   
      
      if msg.pos then
        local line = 0
        for _,len in ipairs(CHptr.len) do
          line = line + len
        end  
        str = str .. '  |  ' .. 'Last note pattern position ' .. line .. '  |  ' .. 'pattern length ' .. CHptr.pat_len
      end
      
      local transpose = getScanTranspose('chrd')
      
      if transpose then        
        if msg.tr then str = str .. '  |  ' .. 'Transpose chord notes by ' .. transpose end
        
        for k,notes in ipairs(CHptr.notes) do
          CHptr.notes[k] = map(notes, function(v) local note, oct = oct0(v) return clamp((v+transpose)%12 + oct*12,0,119) end)
        end
      end
      
    elseif load.poly then ------------------------------------------------ poly
    
      assignPolyTbs()    
      local tb  = fun()
            
      for i=1,12 do    
        POLYptr[i].fn   = tb[i].fn 
        POLYptr[i].off  = tb[i].off       
        POLYptr[i].oct  = tb[i].oct   
        POLYptr[i].len  = tb[i].len 
        POLYptr[i].vel  = tb[i].vel
        POLYptr[i].alen = tb[i].alen    
        POLYptr[i].note = tb[i].note 
        POLYptr[i].line = tb[i].line
         
        POLYptr[i].pat_len = tb[i].pat_len 
        
        POLYptr[i].oct_lowest  = tb[i].oct_lowest
        POLYptr[i].oct_highest = tb[i].oct_highest
      end
      
      Load.maxLenPoly(tb, POLYptr)   
       
      POLYptr[1].fx   = tb[1].fx
      
      for _,note_col in ipairs(POLYptr) do      ------------------------------ delete later
        local pat_len = getPatLen(note_col) 
        
        if note_col.pat_len == nil then  
          note_col.pat_len = pat_len 
        end
      end
      
      if POLYptr[1].pat_len > Load.max_pat_len then
        POLYptr[1].pat_len = Load.max_pat_len
      end
      
      if Load.pattern_len_adj and POLYptr[1].pat_len > renoise.song().selected_pattern.number_of_lines then
        renoise.song().selected_pattern.number_of_lines = POLYptr[1].pat_len
      end

      if msg.pos then
        local lines = {0,0,0,0,0,0,0,0,0,0,0,0}
        
          for i=1,12 do
            for _,len in ipairs(POLYptr[i].len) do
              lines[i] = lines[i] + len
            end  
          end
          
        local max_line = math.max(unpack(lines))  
        str = str .. '  |  ' .. 'Last note pattern position ' .. max_line .. '  |  ' .. 'pattern length ' ..  POLYptr[1].pat_len
      end
      
      local transpose = getScanTranspose('poly')
      
      if transpose then               
        if msg.tr then str = str .. '  |  ' ..  'Transpose poly phrase by ' .. transpose end
        
        for _,note_col in ipairs(POLYptr) do
        
          for k,note in ipairs(note_col.note) do 
                    
            note = clamp(note +  note_col.oct[k] * 12 + transpose, 0, 119)
          
            local nn, oct = oct0(note)
          
            note_col.oct[k]  = oct
            note_col.note[k] = nn                      
          end 
        end
      end     
    end
    
    sprint(str)   
  else
    sprint(fn .. ' Not found !')  
  end 
end


Load.notesPtr = function (fn, event_type)  
  
  fn = MW.path .. fn  
  
  if fileExists(fn) then
    sprint('Processing ' .. fn)
       
    local content = readFile(fn)
    local fun     = loadstring(content)    
    local ptr     = fun() 
       
    if event_type == 'mono' then   
      Load.maxLen(ptr.len )
      Load.maxLen(ptr.off)
        
      return {ptr.note}, {ptr.oct}, {ptr.off}, {ptr.len}, getPatLen(ptr), ptr.fx        
   
    elseif event_type == 'poly' then ---------------------------------------------   
      local notes, octs, offs, lens = {},{},{},{}
       
      for _,v in ipairs(ptr) do      
        table.insert(octs,  v.oct)
        table.insert(offs,  v.off) 
        table.insert(lens,  v.len)         
        table.insert(notes, v.note) 
      end 
      
      Load.maxLenPoly(ptr, ptr) 
         
      return notes, octs, offs, lens, getPatLen(ptr[1]), ptr[1].fx         
      
    elseif event_type == 'chrd' then   ------------------------------------------
      Load.maxLen(ptr.len )
      Load.maxLen(ptr.off)
      
      return ptr.notes, ptr.oct, ptr.off, ptr.len, getPatLen(ptr), ptr.fx 
    end

  else
    return nil,nil,nil  
  end 
end


Save.tag = function(fn)
  fn = fn .. '.tag'
  
  if fileExists(fn) then
    os.remove (fn)   
  end 
  
  local file = io.open(fn, "wb")   
  file:write('return ' .. serialize(Tag))
  file:close()   
end


Save.notes = function(fn)
  
  Save.tag(fn)
  
  if fileExists(fn) then
    os.remove (fn)   
  end 
  
  sprint('Saving ' .. fn)      
  local file = io.open(fn, "wb") 
  
  if Lib.editor == 2 then
    file:write('return ' .. serialize(MW.phr))
  elseif Lib.editor == 1 then     
    if MW.phr.editor_pattern and MW.phr.editor_poly == false then
      file:write('return ' .. serialize(Nptr))  
    elseif MW.phr.editor_chords then  
      file:write('return ' .. serialize(CHptr))
    elseif MW.phr.editor_poly then  
      file:write('return ' .. serialize(MW.phr[MW.phr.no].notes))
    end    
  end
       
  file:close()   
end


local function lenTb (tb)
  
  if tb == nil then return 0 end
  
  local i = 0
  local tb_ret = {}
  for v in pairs(tb) do
    i=i+1
    table.insert(tb_ret, v)
  end
  return i, tb_ret
end 
   

function getFilesRemExt()
  if Lib.editor == 2 then
    return remExt(os.filenames(MW.path, {"*" .. '.phr'}))
  elseif Lib.editor == 1 then
    return remExt(os.filenames(MW.path, {"*" .. Lib.file_ext}))
  end
end


function getFiles(ext)
  return os.filenames(MW.path, {"*" .. ext})
end


function getArpFiles()
  Arp.files_full = os.filenames(MW.path, {"*" .. ".arp"})
  Arp.files = remExtEnd(Arp.files_full)
end


function readFile(fn)
  if fileExists(fn) then
     
    local file = io.open(fn, "rb") 
    local content = file:read "*a"
    file:close()
    
    return content
  end 
end


function writeArpForms (name)  

  local num  
  local file = io.open(name, "wb")   
  
  for i=1,#Arp.forms do
  
    for j=1,#Arp.forms[i] do  
    
      num = Arp.forms[i][j]
      
      if (num < 0) then 
        num = num + 256
      end  
      
      file:write(string.char(num))      
    end
        
    file:write(",")
        
    for j=1,#Arp.skip_note_tb[i] do  
      local char = "f"
      if Arp.skip_note_tb[i][j] then
        char = "t"
      end
      
      file:write(char)      
    end
    
    file:write(",")    
    
    num = Arp.directions[i]
    file:write(string.char(num))        
       
    
    if (i < #Arp.forms) then
      file:write(",")
    end
  end    

  file:close()
end


function readArpForms (name)  

  if (fileExists(name)) then

    local file = io.open(name, "rb") 
    local content = file:read "*a"
    file:close()
    
    Arp.forms = {}
    Arp.skip_note_tb = {}
    Arp.directions = {}
         
    local j=1
    local tb = {}
    tb[1] = {}
    for i=1,#content,1 do
      local char = string.sub(content,i,i)   
      
      if (char == ",") then
        j = j + 1
        tb [j] = {}
      else
        tb [j] [#tb[j]+1] = char
      end   
    end      
    
    local k=1
    for i=1,#tb,3 do
      Arp.forms[k]  = {}
      Arp.skip_note_tb[k]  = {}
      Arp.directions[k]  = {}
      
      for j=1,#tb[i] do
      
        local num = string.byte(tb[i][j])
        local char = tb[i+1][j]
        
        if (num > 128) then 
          num = num - 256
        end  
                    
        Arp.forms[k] [#Arp.forms[k]+1] = num
        
        if (char == "t") then
          Arp.skip_note_tb[k] [#Arp.skip_note_tb[k]+1] = true
        else
          Arp.skip_note_tb[k] [#Arp.skip_note_tb[k]+1] = false
        end 
        
        if (j==1) then
          Arp.directions[k] = string.byte(tb[i+2][j])          
        end                
      end
      k=k+1
    end 
    UI.arp_pres_vb.max = #Arp.forms    
    
    UI.arp_form = 1
    UI.arp_pres_vb.value = 1
  end  
end


