

function scaleKey(scale, key, ...)
  local no_of_keys
  
  if key >  0 then
    if #arg > 0 then
     no_of_keys = arg[1]
     else
     no_of_keys = 12
   end
    
    local scale_ret = {}
    for i, v in ipairs(scale) do
      scale_ret[i] = v + key
      if scale_ret[i] >= no_of_keys then
        scale_ret[i] = scale_ret[i] - 12
        end        
      end  
    table.sort (scale_ret , function (a,b) return a < b end)  
    return scale_ret
  end
  return scale  
end        


function nearest(note , scale)
  local smallest , val = 99
  for _, v in ipairs(scale) do
    if math.abs(v - note) < smallest then
      smallest = math.abs(v - note)
      val = v
    end    
  end 
  return val
end


function noteInScale(note,...) 

  local scale_no = UI.scales_pu.value
  local key_no = UI.keys_pu.value    
  
  if (Scales.on == 2 and arg[1] == nil) or (scale_no == 1 and Scales.on == 1) or (Scales.on == 2 and arg[1] == nil) then  
    return note          
  end
 
  local oct
  
  if note < 0 then
    oct =  - 1
  else
    oct = math.floor(note / 12)  
  end 

  local scale
  if arg[1] then 
    scale = arg[1] 
  else  
    scale = scaleKey (Scales.forms[scale_no], key_no - 1) 
  end  

  return nearest(note%12,scale) + oct * 12
end


function noteInScaleForce(note) 

  local scale_no = UI.scales_pu.value
  local key_no = UI.keys_pu.value    
 
  local oct
  
  if note < 0 then
    oct =  - 1
  else
    oct = math.floor(note / 12)  
  end 

  note = note % 12      
  
  local scale = scaleKey (Scales.forms[scale_no], key_no - 1) 
         
  return nearest(note,scale) + oct * 12
end

local function BuildN_tb (scale, key) 
  local n_tb = {}

  for i=-1,10 do
    for _,n in ipairs(scale) do
      local new_n = n + i * 12 + key 
    
      if new_n >= 0 and new_n < 120 then
        table.insert(n_tb,new_n)
      end   
    end
  end
  
  return n_tb
end


function NearestNfromN_Tb (note, tr, scale,key)  

  local n_tb = BuildN_tb (scale, key) 
  local len = #n_tb
  
  local ix = nearestValue(n_tb, note) 
  
  ix = ix + tr
  if ix < 1 then
    ix = 1
  elseif ix > len then
    ix = len
  end
  
  return n_tb[ix]  
end


function transposeTbInScale(tb,tr, ...)
  
  local res = {}
  
  local key_no   = UI.keys_pu.value  
  local scale_no = UI.scales_pu.value
  
  if (Scales.on == 2 and arg[1] == nil) or (scale_no == 1 and Scales.on == 1) or (Scales.on == 2 and arg[1] == nil) then  
    return tb          
  end
  
  local scale
  if arg[1] then 
    scale = arg[1] 
  else  
    scale = scaleKey (Scales.forms[scale_no], key_no - 1) 
  end 
  
  local n_tb = BuildN_tb (scale, 0) 
  local len = #n_tb
  
  for _,v in ipairs(tb) do
    local ix = nearestValue(n_tb, v) 
  
    ix = ix + tr
    if ix < 1 then
      ix = 1
    elseif ix > len then
      ix = len
    end
    table.insert(res, n_tb[ix])
  end
  
  return res
end 
 
 
function transpose2(tr, norm_note ,oct, sc, key)
  
  local scale
    
  if (Scales.on == 1) then
    scale = Scales.forms[sc]
    key = key - 1
  else
    scale = scaleKey (Scales.forms[1],0) 
  end
  
  local note = NearestNfromN_Tb (norm_note + oct*12, tr, scale,key)
  return oct0 (note)
end


function transpose2FN(tr, note, sc, key)
  
  local scale
    
  if (Scales.on == 1) then
    scale = Scales.forms[sc]
    key = key - 1
  else
    scale = scaleKey (Scales.forms[1],0)    
  end
  
  local res = NearestNfromN_Tb (note, tr, scale,key)
  return res
end


function transposeCh(tr, norm_note ,oct, sc, key) 
  
  key = key-1
  
  local scale = Scales.forms[sc] 
  
  local note = NearestNfromN_Tb (norm_note + oct*12, tr, scale,key)
  return oct0(note)
end
      
   
function transpose(amnt, norm_note ,oct)
  
  local scale_no = UI.scales_pu.value 
  local key_no = UI.keys_pu.value
  
  return transpose2(amnt, norm_note ,oct, scale_no, key_no)
end


function transposeFN(amnt, note) 
  
  local scale_no = UI.scales_pu.value 
  local key_no = UI.keys_pu.value
  
  return transpose2FN(amnt, note, scale_no, key_no)
end


function transposeFN2(tr, note)   
  
  local scale_no = 1
  local key_no = 1
  
  if UI.scales_pu and UI.keys_pu then
   scale_no = UI.scales_pu.value 
   key_no = UI.keys_pu.value  
  end     

  local scale = Scales.forms[scale_no]
  local key = key_no - 1
  
  return  NearestNfromN_Tb (note, tr, scale,key)
end


function transposeBlock(amnt) 

  local fun = function(pat)      
    local note = pat.note_value    
     
    if (note < 120) then       
      pat.note_value = transposeFN2(amnt, note)               
    end
  end

  blockFunApply(fun)
end


function getChrd (note,sca,key,form)
  local tb = {}
  
  for i=1,#form do  
    local noteChrd, octChrd =  transposeCh(form[i],  note ,0, sca, key)
 
    tb[#tb+1] = noteChrd    
  end  
  table.sort(tb)
  return tb
end
            

function getChrd2oct (note,sca,key,form)
  local tb = {}
  
  for i=1,#form do  
    local noteChrd, octChrd =  transposeCh(form[i],  note ,0, sca, key)
    tb[#tb+1] = noteChrd + octChrd * 12
  end  
  table.sort(tb)
  return tb
end


function getChrd2octAll (note,sca,key,form,i)
  if UI.chord_scales[i] > 1 then
    return getChrd2oct (note,sca,key,form)
  else
    return CHptr.notes[i]
  end
end


function getChrd2octAllMarkov (note,sca,key,form,i)
  if UI.chord_scales[i] > 1 then
    return getChrd2oct(note,sca,key,form)
  else
    return map(CHptr.notes[i], function(v) return clampOct (v, Scan.Markov_oct_root,Scan.Markov_oct_range) end)   
  end
end  

  
function getChrdFN2oct(ptr,i)
  local form = UI.chord_forms[i]    
  local chrd = {}
    
  if ptr.len[i] > 0 then
    if UI.chord_scales[i] > 1 then
      chrd = getChrd2oct (ptr.note[i], UI.chord_scales[i] , UI.chord_keys[i], Chords.forms[form])
      chrd = map(chrd, function(v) return ptr.oct[i]*12 + v end) 
    else 
      chrd = CHptr.notes[i]
    end
  end 
  return chrd    
end


function captureFirstChordFromBlock() 

  Chords.captured_slot  = nil
  Chords.captured_notes = nil
  Chords.captured_chord = nil
  
  local patCon, _,  _, sel, pat = getPatFun(true)  
  
  local tr_end,tr_start
  if MW.edit_RN_phrase then
    tr_end,tr_start = 1,1
  else
    tr_end   = sel.end_track
    tr_start = sel.start_track 
  end
  
  local notes = {}
  
  for tr = tr_start, tr_end do
    for ln = sel.start_line, sel.end_line do
      local n_col_tb = {}      
      
      for n_col = sel.start_column, sel.end_column do
        if n_col <= 12 then      
          local patl = patCon(ln, n_col, tr) -- pat:track(tr):line(ln):note_column(n_col)
          local note = patl.note_value
         
          if (note < 120) then             
            table.insert(notes, patl.note_value)
          end
        end
      end 
      
      Chords.captured_notes = copy(notes)
      table.sort(notes) 
            
      if #notes > 0 then 
        
        if MW.phr.editor_name == "Tonnetz" then   
          local line = 0
          for i,len in ipairs(CHptr.len) do
            line = line + len 
            if line == ln then
              Chords.captured_slot = i              
           
              MW.seq_start = i 
              MW.seq_length  = i        
              
              updateSlots()    
            end
          end
        end
   
       -- local notes0_unique = unique(map(notes, function(v) return v%12 end))  
       -- table.sort(notes0_unique)     
        
       -- local chrd = getChFromNodes(Scan.chords_tb,notes0_unique,1) 
        
        local scale,key   
        
        if MW.started then
          key   = UI.keys_pu.value
          scale = UI.scales_pu.value 
        end
        
        local chrd
        
        if MW.started and scale > 1 then 
          local sk = scale .. '_' .. key
          local scan = Scan.chords_tb_oct0_sk[sk]  
        
          chrd = getChFromNodes(scan, unique(tbOct0all(notes)), 1)
          
          if chrd == nil then
            chrd = getChFromNodes(Scan.chords_tb_oct0, unique(tbOct0all(notes)), 1)
          end          
        else 
          chrd = getChFromNodes(Scan.chords_tb_oct0, unique(tbOct0all(notes)), 1)  
        end
        
        if chrd then
          sprint('Form ' .. chrd.form .. '  ' .. Scales.keys[chrd.key] .. ' ' ..  Scales.names[chrd.scale])
          Chords.current_form       = chrd.form         
          Chords.captured_form      = chrd.form 
          Chords.captured_chord     = chrd          
          Chords.captured_pattern   = getPatLn(ln)   --pat:track(tr):line(ln)
          Chords.captured_chord.oct = math.floor(notes[1]/12)
          return
        end          
      end
    end    
  end
end


function blockLineSetForm(n) 
  local scale, key = 1,1 
   
  local captured = Chords.captured_chord    
  
  if not captured then 
    sprint('Capture notes first')
    return 
  end 
   
  if MW.started and scale == 1 then
    scale = captured.scale
    key = captured.key
  end         
    
    local old_form = Chords.forms[Chords.captured_form]  
    
    local nf = clamp(Chords.current_form + n,1,#Chords.forms-1)
    local new_form = Chords.forms[nf] 
    
    local chord_tmp = getChrd (captured.note,scale,key,new_form) 
    
    while consonanceRating2Tb(chord_tmp, chord_tmp) > Stats.forms_cs do -- dissonanceScore2Tb(chord_tmp, chord_tmp) > 0.5 do --   
      nf        = clampAround (nf + n, 1, #Chords.forms-1)  
      new_form  = Chords.forms[nf]       
      chord_tmp = getChrd (captured.note,scale,key,new_form) 
    end   
    
    Chords.current_form = nf      
    
    local note = captured.note  
    local captued_notes0 = map(Chords.captured_notes, function(v) return v%12 end)
    
    local form_ix, oct_ix, oct_d = {},{},{}
    for k,v in ipairs(captued_notes0) do
      local exists, key = member(v, captured.chrd_tb)
      if exists then
        local oct = math.floor(Chords.captured_notes[k]/12)
        table.insert(oct_ix, oct)
        table.insert(form_ix, key)        
        
        if oct_d[oct] == nil then
          oct_d[oct] = {}
          oct_d[oct] = 1
        else
          oct_d[oct] = oct_d[oct] + 1
        end
      end
    end   
 
    local highest, dominant_oct = 0      
    for oct,count in pairs(oct_d) do
      if count > highest then
        highest = count
        dominant_oct = oct
      end
    end
        
    local new_notes  = {}     
    local new_notes0 = {}     
    for _,f in ipairs(new_form) do  
      local noteChrd, octChrd = transposeCh(f, note ,0 , scale, key)
      table.insert(new_notes0, noteChrd)
       
      noteChrd, octChrd = transposeCh(f, note ,dominant_oct , scale, key)
      table.insert(new_notes, noteChrd + octChrd*12)
    end  
     
    local new_chord = {}
    for k,v in ipairs(form_ix) do
      if v <= #new_notes0 then
        table.insert(new_chord, new_notes0[v] + oct_ix[k]*12)
      end
    end    
     
    if #new_form > #old_form then
      for i=#old_form+1, #new_form do
        table.insert(new_chord, new_notes[i])
      end 
    end
                    
    local patl = Chords.captured_pattern
     
    for n_col=1,12 do
      patl:note_column(n_col).note_value = 120
      patl:note_column(n_col).instrument_value = 255
    end   
                     
    for n_col,note in ipairs(new_chord) do
      patl:note_column(n_col).note_value = note
      patl:note_column(n_col).instrument_value = renoise.song().selected_instrument_index-1   
    end         
     
    local str = 'Chord form ' .. Chords.current_form .. ',  ' .. table.concat(new_form, '-') .. ',  scale ' .. Scales.keys[key] .. ' ' ..  Scales.names[scale]
    sprint(str)
     
    return 
end
 
 
function restoreCapturedBlockNotes()
  if Chords.captured_notes then
    local patl = Chords.captured_pattern
                     
    for n_col,note in ipairs(Chords.captured_notes) do
      patl:note_column(n_col).note_value = note
      patl:note_column(n_col).instrument_value = renoise.song().selected_instrument_index-1   
    end         
        
    sprint('Restore notes')    
  end
end
 

function nToChrdNS (len, i)
  
  local str  = ''
  local chrd = CHptr.notes[i]
  
  for k,v in ipairs(chrd) do 
     str = str .. v
     
     if k ~= #chrd then
       str = str .. '_'
     else
       str = str .. ','  .. len
     end   
  end
  
  return str, chrd  
end

  
function nToChrdNS (note, oct, len, i)

  local scale = UI.chord_scales[i]
  local key = UI.chord_keys [i]
  local form = Chords.forms[UI.chord_forms[i]]  
        
  local str = ''

  local chrd = getChrd2octAll(note, scale,key,form,i)  --  getChrd2oct (note, scale,key,form)
  
  for k,v in ipairs(chrd) do 
     str = str .. v
     
     if k ~= #chrd then
       str = str .. '_'
     else
       str = str .. ',' .. oct .. '_' .. len
     end   
  end
  
  return str, chrd
end  
  
  
function bestScales(notes)
  local res = {}
  
  for k=2,#Scales.forms do 
    for j=0,11 do 
      local scale = scaleKey (Scales.forms[k], j) 
      
      local count = 0
      for i,note in pairs(notes) do        
        if member (note, scale) then   
          count = count+1                   
        end     
      end  
      
      if count == #notes then
        table.insert(res,{scale = k, key = j+1}) 
      end    
    end
  end
  return res
end  
  

function dissonanceNCheckOK(note, note2)
  
  local distance = math.abs(note - note2) % 12
    
  -- if distance > 6 then
   --  distance = 12 - distance
  -- end
     
   if distance == 1 or distance == 2 or distance == 6 then
     return false
   end             
  return true
end


function dissonanceCheckOK(tb, note)
  for _,v in ipairs(tb) do
    local distance = math.abs(note - v) % 12
    
--     if distance > 6 then
  --     distance = 12 - distance
    -- end
     
     if distance == 1 or distance == 2 or distance == 6 then
       return false
     end         
  end  
  return true
end


function dissonanceScoreTbN(tb, note)
  local score = 0 
  
  for _,v in ipairs(tb) do
    local distance = math.abs(note - v) % 12
    
   --  if distance > 6 then
    --   distance = 12 - distance
    -- end
     
     score = score + Stats.chromatic_penalty[distance]      
  end  
  return 1/score
end   


function dissonanceScore2Tb(tb,tb2)

  local score = 0 
  
  for _,v in ipairs(tb) do
    for _,v2 in ipairs(tb2) do    
      local distance = math.abs(v2 - v) % 12
     
      if distance > 6 then
       distance = 12 - distance
     end
     
     score = score + Stats.chromatic_penalty[distance]
    end
  end  
  
  if score == 0 then
    return 1
  else
    return 1/score
  end
end   


function consonanceRating(note,note2)
  local score    = 0 
  local distance = math.abs(note - note2) % 12  
  
  return score + Stats.consonance_rating[distance]
end 


function consonanceRating2Tb(tb,tb2)
  local score = 0 
  local len = 0
  
  for _,v in ipairs(tb) do
    for _,v2 in ipairs(tb2) do    
      local distance = math.abs(v2 - v) % 12  
      score = score + Stats.consonance_rating[distance]
      len = len + 1
    end
  end  
  
  if score == 0 then
    return 1.7
  else
    return score / len
  end
end 


local function scaleFindHlp (tb)
  local max = 0
  local ii = 0
  for i=1,#tb do
    if (tb[i] > max) then
      max = tb[i]
      ii = i
    end   
  end
  return "  " .. Scales.keys[ii] .. Scales.keys_dash[ii] .. tb[ii], tb[ii], ii 
end


local function findScaleKeyUpdatePopupExact()

  for i=1,#Scales.names do
    Scales.names[i] = Scales.names2[i]
  end 

  local scale_presets_list_tmp = {}
  local scale_notes_found = {}
  local keys_found = {}
   
  for k=2,#Scales.forms do 
    local scale_key_tb = {0,0,0,0,0,0,0,0,0,0,0,0}
    for j=0,11 do     
      local foundTb = {0,0,0,0,0,0,0,0,0,0,0,0}
      local scale = scaleKey (Scales.forms[k], j)  
      
      for _, note in ipairs(Stats.unique_notes) do
        if member (note, scale) then
          foundTb[note+1] = 1
         end 
      end
      scale_key_tb[j+1] =  sumTb(foundTb)
    end  
    
    local str, no, key = scaleFindHlp (scale_key_tb)
    scale_presets_list_tmp[k] =  Scales.names2[k] .. str
    scale_notes_found[k] = no
    keys_found[k] = key
  end

  scale_notes_found[1] = 0
  local max = (math.max(unpack(scale_notes_found)))
  

  Scales.best_scale, Scales.best_key = {}, {}
  
  for i=2,#Scales.forms do
    if (scale_notes_found[i] == max) then
      Scales.names[i] = scale_presets_list_tmp[i]

      table.insert(Scales.best_scale, i)
      table.insert(Scales.best_key,keys_found[i])      
    end 
  end

  UI.scales_pu.items = Scales.names
end


function findScaleKeyUpdatePopup() 
  local sca,key
    
  if Scales.profiles.no > 1 then 
    local profile = Scales.profiles.pu[Scales.profiles.no] 
    
    sca,key =  findScaleKeyProf(profile, Stats.pitch_duration_dst) 
    
    Scales.best_scale = {sca}
    Scales.best_key = {key} 
  else
    findScaleKeyUpdatePopupExact()    
    sca = Scales.best_scale[1]
    key = Scales.best_key[1]
  end
  
 -- sprint('Profile: ' .. profile .. '  ' .. Scales.names2[sca] .. ' ' .. Scales.keys[key])
end
