WeightedIntervals = {}

WeightedIntervals.new                = false
WeightedIntervals.lock               = 1
WeightedIntervals.first_note         = 60
WeightedIntervals.hooke_law_coef     = .5
WeightedIntervals.root_note_exists   = false
WeightedIntervals.specify_first_note = false

local int_last_best_root

local int_ui_root_w = {} 
local int_ui_pos_w = {} 
local int_ui_neg_w = {} 


local function sum0i(tb)
  local sum = 0
  for i=0,#tb do
    sum = sum + tb[i] 
  end
  return sum
end 


local function cumSum0i(tb)
  local cumsum = {}
  local sum = 0
  
  for i=0,#tb do 
    sum = sum + tb[i] 
    cumsum[i] = sum  
  end
  return cumsum  
end


local function normalize0i(tb)
  local mul = 1 / sum0i(tb) 
  
  for i=0,#tb do 
    tb[i] = tb[i] * mul    
  end
  return tb
end


local function zeroTb (to_zero, zeros_src)
  
  for k,v in ipairs(zeros_src) do
    if v == 0 then
      to_zero[k] = 0
    end
  end
end


local function notesFromIntervals(len,down,up,w2,N_tb) -- https://code.google.com/archive/p/contrapunctus/ 
  
  local notesMade = 0    

  local n = N_tb[1] 
  
  while true do
  
    local w = {}
  
    for i=0,down+up do
      w[i] = 0
    end
    
    w = copy(w2) 
  
    local n_norm, o_norm = oct0(n)  
     
    -- Avoid tritone with F-B   A D#, A# E, C F#, C# G, D G#
 --   if n_norm == 5 then w[3] = 0 end -- F
 --   if n_norm == 11 then w[3+down] = 0 end-- B  
    
  --  if n_norm == 3 then w[3] = 0 end -- D#
   -- if n_norm == 9 then w[3+down] = 0 end-- A  
    
   -- if n_norm == 4 then w[3] = 0 end -- E
   --- if n_norm == 10 then w[3+down] = 0 end-- A#  
    
  --  if n_norm == 6 then w[3] = 0 end -- F#
   -- if n_norm == 0 then w[3+down] = 0 end-- C  
    
  --  if n_norm == 7 then w[3] = 0 end -- G
    --if n_norm == 1 then w[3+down] = 0 end-- C#  
    
  --  if n_norm == 8 then w[3] = 0 end -- G#
    --if n_norm == 2 then w[3+down] = 0 end-- D
    
    -- Minor six possible when ascending
 --   if n_norm == 4 or n_norm == 9 or n_norm == 11 then w[5] = 6 end       -- E  A  B
  
    local distance = n - Rand.root_note 
    local hookelaw = math.pow(1.17604743, math.abs(distance)) 

  
    if distance > 0 then
      for i=1,up do
        w[i] = w[i] / (i * hookelaw * WeightedIntervals.hooke_law_coef)
      end
    end

    if distance < 0 then
      for i=1+up,up+down do
        w[i] = w[i] / (i * hookelaw * WeightedIntervals.hooke_law_coef)  
      end
    end

    w = normalize0i(w) 
 
    local cum = cumSum0i(w)
    
    zeroTb (cum, w2)   
     
    local nn = n 
    local found_flag = false
    
    local r = math.random(100)/100  
    
    if r < cum[0] then
      nn = n
      found_flag = true
    end

    for k=1,up do
      if r < cum[k] then
        nn = transposeFN(k,n)  
        found_flag = true
        break
      end
    end

    if not found_flag then
      for k=1,down do
        if r < cum[k+up] then
          nn = transposeFN(-k,n)
          found_flag = true
          break
        end
      end
    end 
    
    nn = clamp2(nn,11+12,108-12)
    N_tb[#N_tb+1] = nn
    n = nn
      
    notesMade = notesMade+1
    
    --Resolve to a leading tone if we have enough notes and we have a B or a D.
    n_norm, o_norm = oct0(n) 
    if notesMade > len then --if notesMade > len and (n_norm == 11 or n_norm == 2) then    -- B4  D5
      break
    end
  end
  
  N_tb[#N_tb+1] = Rand.root_note
  
  return N_tb
end
 

local function getRhythm()
  
  local start_ln
  if MW.block_editing then
   local _
    _, _, start_ln = getPatFun(MW.block_editing, renoise.song().selected_pattern_index)  
    start_ln = start_ln + 1
  else
    start_ln = 1
  end
  
  local line = 0
  
  local L_tb = {start_ln}
  local rhythms = {{2,2,2,2}, {4,4}, {4,2,2}, {2,2,4}, {1,3,4,3}, {3,1,4,4}, {3,2,2,1,4}, {2,1,1,3,1}}
   
  while line < 64  do
    local rhm = rhythms[math.random(#rhythms)]
    
    local len2 = 1 
    while len2 <= #rhm do
      L_tb[#L_tb+1] =  rhm[len2]
      
      line = line + rhm[len2]
      
      len2 = len2 + 1       
    end 
  end 
  
  return L_tb
end


local function grabNotesR()
  
  local ph  = MW.phr.no
  local col = MW.phr.col
  
  if MW.phr.editor_poly then
    
    if init then initPoly() end
    
    for i=12,1,-1 do  
      if init then initNotes() end
      
      MW.timer_poly_pos[i] = {}
      
      GrebSeq(i,MW.phr[ph].notes[ph])  
     end         
  elseif MW.phr.editor_chords then
    if init then initNotes() end
    
    GrebSeq(col, MW.phr[ph].notes[col]) 
  else 
    if init then initNotes() end

     GrebSeq(col, col,MW.phr[ph].notes[col])  
  end
end

    
function randomizeChords ()
  
  if MW.phr.event_type ~= 3 then return end     

  local key = math.random(12)
  for i=1,#UI.chrd_scale_vbs do
  
    if CHptr.rand[i] then
    
      if Rand.chord_scale == 2 then
        UI.chrd_scale_vbs[i].value = math.random(#Scales.names2)                   
      elseif Rand.chord_scale == 2 then
        UI.chrd_scale_vbs[i].value = math.random(2)
      end
      
      if Rand.chord_key == 2 then
        UI.chrd_key_vbs[i].value =  math.random(12)
      elseif Rand.chord_key == 3 then  
        UI.chrd_key_vbs[i].value =  key       
      end
        
      if Rand.chord_note then
        UI.note_vbs[i].value = math.random(-12,12)
      end  
        
      if Rand.chord_form then
        UI.chrd_form_vbs[i].value = math.random(20)
      end          
    end              
  end  
  
  
  local pat_con, pat_len, line2 = getPatFun(MW.block_editing)
  
  local line = 0 
  if MW.block_editing == false then line = line2 - 1 end 

  if MW.phr.editor_name == "Arpeggiator" then
  
    if Rand.arp_pattern == 2 then
      UI.arp_randomize(Arp.form_no)
    elseif Rand.arp_pattern == 3 then
      for i=1,#Arp.forms do
        UI.arp_randomize(i)
      end
    end
  
    for i=1,#CHptr.len do 
  
      line = line +  CHptr.len[i]

      if line > pat_len then break end  
    
      if CHptr.arp.rand[i] then
    
        if Rand.arp_oct then
          CHptr.arp.oct[i] =  math.random(3)
        end      
      
        if Rand.arp_pattern_no then
          CHptr.arp.form[i] = math.random(#Arp.forms)
        end
      
        if Rand.arp_interval then
          CHptr.arp.off[i] =  math.random(#CHptr.notes[i])
        end
      
      end
    end
  end
  
  updateSlots()        
  fill(MW.phr.col,{line = 0})
  
  if repeat_transpose_pu == 2 then
    repeatSeq()  
  end  
end


local function sumKeys(tb)
  local sum = 0
  
  for k,v in pairs(tb) do
    sum = sum + k
  end
  return sum
end
 
 
local function mulKeysValues(tb)
  local sum = 0
  
  for k,v in pairs(tb) do
    sum = sum + k*v.n
  end
  return sum
end
 
 
local function sumNOintervals(notes,root)  
  
  local pos_tb = {} 
  local neg_tb = {}  

  local zero = 0
  
  for k,note in ipairs(notes) do    
     local interval = note - root  
     local ix = math.abs(interval)
     
     if  interval == 0 then
       zero = zero + 1   
     elseif interval > 0 then  
       if pos_tb[ix] then
         pos_tb[ix].n = pos_tb[ix].n + 1
       else
         pos_tb[ix] = {}
         pos_tb[ix].n = 1
       end
     else
       if neg_tb[ix] then
         neg_tb[ix].n = neg_tb[ix].n + 1
       else
         neg_tb[ix] = {}
         neg_tb[ix].n = 1
       end   
     end       
  end  
  
  return pos_tb, neg_tb, zero
end 


WeightedIntervals.fun = function() 

  if noteLengthEmpty(DRptr.len) then
    sprint('Capture or load some notes first.')
    return
  end
  
  local no       = {} 
  local zero     = 0    
  local neg_tb   = {}
  local pos_tb   = {} 
  
  local line = 0
    
  for i,v in ipairs (DRptr.len) do 
    local dur = 0
    
    line = line + dur
    
    if DRptr.off[i] > 0 then 
      dur = DRptr.off[i] 
    elseif i < #DRptr.len then 
      dur = DRptr.len[i+1] 
    else
      dur = DRptr.pat_len - line + 1
    end
    
    if dur > 0 then
      table.insert(no, DRptr.note[i] + DRptr.oct[i]*12)
    end     
  end
  
  local notes = unique(no)  
  table.sort(notes)
   
  local best = 999
  local best_root 
 
 
  if WeightedIntervals.new then
    for _,root in ipairs(notes) do
  
      pos_tb, neg_tb, zero = sumNOintervals(no,root)
  
      local pos_sum = mulKeysValues(pos_tb)  
      local neg_sum = mulKeysValues(neg_tb)  
  
      local tmp_best = math.abs(pos_sum-neg_sum)
      if tmp_best < best then  
        best = tmp_best  
        best_root = root  
      end
    end
    
    int_last_best_root = best_root
    
    UI.notes_pu.items = map(notes, function(v) return num2note(v) end)
    UI.notes_pu.value = nearestValue(notes, best_root)
  else
    if #UI.notes_pu.items>1 and UI.notes_pu.items[UI.notes_pu.value] ~= ''  then
      best_root = notes[UI.notes_pu.value]  
    else
      best_root = int_last_best_root  
    end
  end   
  
  UI.interval_note_tx.text = num2note(best_root)
 
  pos_tb, neg_tb, zero = sumNOintervals(no,best_root)   
  
  local pos_sum = mulKeysValues(pos_tb)  
  local neg_sum = mulKeysValues(neg_tb) 
    
  local up = 0
  for k,v in pairs(pos_tb) do
    if k>up then up = k end
  end
  
  local down = 0
  for k,v in pairs(neg_tb) do
    if k>down then down = k end
  end
  
  
  local w = {}
  for i=1,up+down do
    w[i] = 0
  end  
     
  for k,v in pairs(pos_tb) do
    w[k] = v.n
  end
  
  for k,v in pairs(neg_tb) do
    w[k+up] = v.n
  end

  w[0] = zero  
  
  
  if WeightedIntervals.new or WeightedIntervals.root_note_exists then

    if #int_ui_root_w>0 then  
      UI.int_root_w:remove_child(int_ui_root_w[1])
      int_ui_root_w = {} 
    end
    
    for k,v in pairs(int_ui_pos_w)  do
      if v then
        UI.int_pos_w:remove_child(v)
        int_ui_pos_w[k] = nil
      end
    end 
    
    for k,v in pairs(int_ui_neg_w)  do
      if v then
        UI.int_neg_w:remove_child(v) 
        int_ui_neg_w[k] = nil
      end
    end
    
    local add_vb = function (val_dest,n)       
      local vb = renoise.ViewBuilder()
            
      return vb:valuebox {
        min = 0,
        max = 99,
        width = 45,
        tooltip = n,
        value = val_dest,
        notifier = function(val)
          val_dest = val
        end } 
    end 
    
    
    table.insert(int_ui_root_w, add_vb(w[0], 'get back to root'))  
    UI.int_root_w:add_child(int_ui_root_w[1])

    for i=1,#w do
      if i <= up and w[i] > 0 then
        int_ui_pos_w[i] = add_vb(w[i], '' .. i)   
        UI.int_pos_w:add_child(int_ui_pos_w[i])
      elseif w[i] > 0 then
        int_ui_neg_w[i-up] = add_vb(w[i], '' .. -1*(i-up))   
        UI.int_neg_w:add_child(int_ui_neg_w[i-up]) 
      end
    end    

    WeightedIntervals.root_note_exists = false
  else
    if #int_ui_root_w>0 then 
      w[0] = int_ui_root_w[1].value
    end    
    
    for k,v in ipairs(int_ui_pos_w) do
      w[k] = v.value
    end
    
    for k,v in ipairs(int_ui_neg_w) do
      w[k+up] = v.value
    end
  end

  WeightedIntervals.best_root = best_root
  
  local L_tb = getRhythm()
  
  local N_tb
  if WeightedIntervals.specify_first_note then  
    N_tb = {UI.int_start_N.value} 
  elseif MW.block_editing then
    local _,oct = oct0(UI.int_start_N.value)
    N_tb = {math.random(0,11) + oct*12}  
  else  
    N_tb = {WeightedIntervals.best_root}
  end
  
  N_tb = notesFromIntervals(#L_tb,down,up,w,N_tb)
   
  for k,v in ipairs(L_tb) do
    Nptr.len[k] = v
  end
  
  for k,v in ipairs(N_tb) do
    local n,o = oct0(v)      
    Nptr.oct [k] = o
    Nptr.off [k] = 0
    Nptr.vel [k] = 128    
    Nptr.note[k] = n
  end  
 
  --fill(MW.phr.col) 
  fillPtr(MW.phr.col, Nptr, 1)  
  
  
  if repeat_transpose_pu == 2 then
    repeatSeq()  
  end
  
  updateSlots()  
  
  WeightedIntervals.new = false
end

      
function Randomize ()

  local n_col = MW.phr.col
  
  local rand_N, rand_O, tmp_N, tmp_O, tmp_NO, last_CHNO, chrd
  local use_rand_note_not_chrd = true
      
  local ch_i = 1
      
  if Rand.form_lib_chord == 3 and Rand.chord_p > 0 then
    local filesCh = remExt(os.filenames(MW.path, {"*" .. ".chrd"}))
    local randfn = math.random(#filesCh)
    local fn = MW.path ..filesCh[randfn] .. ".chrd"

    initChords()
    local load_chords = {phr  = false, step = false, poly = false, chrd = true}
    Load.notes(fn,nil,load_chords)
    
    if Scales.on == 1 then
      getChordNotesDistrib()
      findScaleKeyUpdatePopup()
            
      if #Scales.best_scale > 0 and #Scales.best_key > 0 then      
        UI.scales_pu.value = Scales.best_scale[1]
        UI.keys_pu.value = Scales.best_key[1]
      end  
    end
  end
  
         
  local random_num, note          
        
--  local note,octave = oct0(Rand.root_note)      
      
  if Rand.key then
    UI.keys_pu.value = math.random(12)
  end

  if Rand.scale == 2 then
    UI.scales_pu.value = math.random(2)+1
  elseif Rand.scale == 3 then
    UI.scales_pu.value = math.random(#UI.scales_pu.items-1)+1
  end
            
  local rand_len = 0  
      
  local pat_con, pat_len, line = getPatFun(false)  

  for i=1,#Nptr.len do 

    if Rand.len_typ == 1 and Rand.note_len > 1 then
    
      random_num = math.random(Rand.note_len)      
      Nptr.len[i] = random_num
      
    elseif Rand.len_typ == 2 then
    
      random_num = 0
      while random_num == 0 do
        random_num = round(MW.fib_ratios[math.random(#MW.fib_ratios)] * Rand.note_len)
      end   
      Nptr.len[i] = random_num 
      
    elseif Rand.len_typ == 3 then   
    
      Nptr.len[i] = round(Fibonacci(math.random(99)) % Rand.note_len + 1)
                 
    elseif Rand.len_typ == 4 then
    
      if i > 1 then
        local distribution = buildProbTb(Stats.note_len_distribution,0)
        local rand_n = math.random(#distribution)          
        Nptr.len[i] = distribution[rand_n]
      else
        Nptr.len[i] = 1  
      end
    end    
         
    line = line +  Nptr.len[i]

    if line > pat_len then break end  

    if MW.phr.editor_name == "Arpeggiator" and CHptr.arp.rand[i] then
      CHptr.arp.form[i] = math.random(#Arp.forms)
        
      local oct_prob = probTb ({ {1,6},{2,4},{3,2},{4,1} })
      CHptr.arp.oct[i] = oct_prob [math.random(#oct_prob)]       
      
      if UI.arp_note_interval[i] ~= nil then
        local form_n =  #Chords.forms[UI.chord_forms[i]] 
        UI.arp_note_interval[i].max = form_n
        CHptr.arp.off[i] = math.random(form_n)
      end
    end
    
    if MW.phr.editor_chords == false or (MW.phr.editor_chords and CHptr.rand[i]) then   

      local RN = Rand.root_note          
      if Rand.b_motion and i > 1 then 
        RN = Nptr.note[i-1] + Nptr.oct[i-1]*12               
      end
      tmp_N, tmp_O = oct0(RN)
      
      if Rand.chord_p > 0 and Rand.form_lib_chord == 2 then
        local scale = UI.scales_pu.value
        local key = UI.keys_pu.value
        local form = Chords.forms[math.random(12)]  
        
        sprint('Form ' .. reduce(form, function (a,b) return a .. ' ' .. b end))
        
        tmp_N, tmp_O = oct0(RN)         
        chrd = getChrd2oct (tmp_N, scale,key,form) 
      end  
      
        
      local chprob_tb = probTb({  {true,Rand.chord_p}, {false,10-Rand.chord_p} })  
      
      
      if Rand.chord_p > 0 and Rand.form_lib_chord == 3 and tbNotZero(CHptr.len) and chprob_tb[math.random(#chprob_tb)] then

        if getChordPos(CHptr.len) then
          getChordNotes ()
         
          for j=1,#Chords.pos do
            if line >=  Chords.pos[j].st and line <=  Chords.pos[j].en then    

              local randn = #Chords.pos[j].chord_notes 
              if ch_i > randn then ch_i = 1 end
              local note = Chords.pos[j].chord_notes[ch_i]    
              
              if note then 
                use_rand_note_not_chrd = false  
              else
                use_rand_note_not_chrd = true
              end
                        
              ch_i = ch_i+1           
              
              tmp_NO = note + tmp_O*12  
              last_CHNO = tmp_NO 
            end
          end           
        end
      elseif Rand.chord_p > 0 and Rand.form_lib_chord == 2 and chprob_tb[math.random(#chprob_tb)] then 

        tmp_N = chrd[math.random(#chrd)]
        
                         
        tmp_NO = tmp_N + tmp_O*12          
        last_CHNO = tmp_NO
          
        use_rand_note_not_chrd = false      
      else
        use_rand_note_not_chrd = true  --not chrd   
      end   
        
      if use_rand_note_not_chrd then  
              
        if Rand.chord_p > 0 and last_CHNO then
          RN = last_CHNO   
          tmp_N, tmp_O = oct0(RN)  
        end
     
        if Rand.note_typ == 5 and sumTb(Stats.pitch_duration_dst) > 0 then
          local distribution = buildProbTb(Stats.pitch_duration_dst,-1)
          local rand_n = math.random(#distribution)
          
          local oct_down = 0
          if math.random(2) == 1 then
            oct_down = 12
          end          
          
          tmp_NO = distribution[rand_n] + tmp_O*12 - oct_down  
          
        elseif Rand.note_typ == 4 then 
          local pnfun = pinknoise (0,Rand.note)    
        elseif Rand.note_typ == 3 then  
          tmp_NO = RN + randomFib(13-Rand.note) 
        elseif Rand.note_typ == 2 then  
          tmp_NO = RN + round(randNeg() * Rand.note * MW.fib_ratios[math.random(#MW.fib_ratios-1)])                      
        elseif Rand.note_typ == 1 then          
          tmp_NO =  math.random(Rand.note) * randNeg() + RN  
        else 
          return
        end         
      end
      
            
      rand_N, rand_O = oct0(clamp(tmp_NO,11,108)) 
            
      
      Nptr.note[i] = noteInScale(rand_N) 
      Nptr.oct[i] = rand_O          
      
      
      if Rand.octave > 1 then
        local n 
        if math.random(2) == 1 then
          Nptr.oct[i] = clamp(rand_O + math.random(Rand.octave),1,8)
        else
          Nptr.oct[i] = clamp(rand_O - math.random(Rand.octave),1,8)
        end        
      end
  
      if Rand.note_vel == 1 then
        Nptr.vel[i] = 127
      else 
        Nptr.vel[i] = round(127 - math.random(Rand.note_vel))
      end
    end 
  end
       
  fill(MW.phr.col) 

  if repeat_transpose_pu == 2 then
    repeatSeq()  
  end
  
  updateSlots()    
end

function mutate ()   

  local n_col = MW.phr.col
  
  local random_no           
  for i=1,#Nptr.len do
    if Rand.note_len > 1 then
      if i > 1 and Rand.mutate_fib_len then
        random_no = round (MW.fib_ratios[math.random(#MW.fib_ratios)] * Nptr.len[i])
        Nptr.len[i] = clamp(random_no,1, 16)
     elseif i > 1 then
       random_no = math.random(-Rand.mutate_len, Rand.mutate_len)
       Nptr.len[i] = clamp(Nptr.len[i] + random_no,1, 16)
     end
   end
  
   if Rand.mutate_fib_note then

     random_no = round (MW.fib_ratios[math.random(#MW.fib_ratios)] * Rand.mutate_note) * randNeg()
     Nptr.note[i] = clamp(Nptr.note[i] + random_no,-12, 12)
     Nptr.note[i] = noteInScale(Nptr.note[i])
   elseif Rand.mutate_note > 1 then 
   
     random_no = math.random(Rand.mutate_note) * randNeg()
     Nptr.note[i] = clamp(Nptr.note[i] + random_no,-12, 12)
     Nptr.note[i] = noteInScale(Nptr.note[i])
   end
         
    if Rand.mutate_oct <= 2 and Rand.mutate_oct > 1 then
      random_no = math.random(2)
      Nptr.oct[i] = clamp(Nptr.oct[i] + random_no - 1, 1, 7)
    elseif Rand.mutate_oct <= 4 and Rand.mutate_oct > 2 then
      random_no = math.random(4)
      Nptr.oct[i] = clamp(Nptr.oct[i] + random_no - 2, 1, 7)
    elseif Rand.mutate_oct <= 6 and Rand.mutate_oct > 4 then
      random_no = math.random(6)
      Nptr.oct[i] = clamp(Nptr.oct[i] + random_no - 3, 1, 7)                   
    end          
         
          
    if Rand.note_vel > 1 then
      local random_no = math.random(Rand.note_vel)
      Nptr.vel[i] = clamp(Nptr.vel[i] + random_no - (random_no / 2), 1, 127)  
    end
  end  
        
  fill(n_col)
  updateSlots() 
end 

function mutateFun()    
  if Rand.mutate_markov == 1 then   
    mutate()         
  else
    MarkovProcess()    
  end
end

