

function funOverElOrTbRet(fun, ...)
  if type(arg[1]) == 'table' and type(arg[2]) == 'table' then  
    local res = {}    
   
    for k,v in pairs(arg[1]) do 
      if type(v) == 'table' then    
      
        if res[k] == nil then
          res[k] = {}
        end
        
        for j,vv in pairs(v) do  
          res[k][j] = fun(vv, arg[2][k][j])   
        end
      else  
        res[k] = fun(v, arg[2][k], unpack(arg,3))
      end
    end
    return res  
  elseif type(arg[1]) == 'table' then
    local res = {}
    for i=1,#arg[1] do      
      res[i] = fun(arg[1][i], unpack(arg,2)) 
    end
    return res
  else
    return fun(unpack(arg))
  end
end

function funOverElOrTb(fun, ...)
  if type(arg[1]) == 'table' and type(arg[2]) == 'table' then 
    for k,v in pairs(arg[1]) do 
      if type(v) == 'table' and type(arg[2][k]) == 'table' then     
        for j,vv in pairs(v) do 
          fun(vv, arg[2][k][j], unpack(arg,3))   
        end
      elseif type(v) == 'table' then
        for j,vv in pairs(v) do 
          fun(vv, arg[2][k], unpack(arg,3))   
        end
      else 
        fun(v, arg[2][k], unpack(arg,3))      
      end
    end
  elseif type(arg[1]) == 'table' then  
    for i=1,#arg[1] do      
      fun(arg[1][i], unpack(arg,2)) 
    end
  else
    fun(unpack(arg))
  end
end


function scanTags() 
  Tag.tb = {}
  local tb = splitDl (Tag.text, ',')
    
  for _,v in ipairs(tb) do
    repeat v = v:gsub('% ', '') until v == v
    table.insert(Tag.tb, v)
  end 
end

function scanSearch(tag) 
  local res = {}
  res.include = {}
  res.exclude = {}
  res.opFun = function(n) return true end
  
  local tb = splitDl (tag, ',') 
    
  for _,v in ipairs(tb) do 
    if tonumber(v) then 
        res.opFun = function(n) return n == tonumber(v) end        
    elseif string.match(v, '[%><=]') and string.match(v, '[%d]') then  
    
      local num = tonumber(string.match(v, '[%d]'))
      local op       
      
      if string.match(v, '[%>=]') then
        res.opFun = function(n) return n >= num end
      elseif string.match(v, '[%<=]') then
        res.opFun = function(n) return n <= num end
      elseif string.match(v, '[%>]') then
        res.opFun = function(n) return n > num end
      elseif string.match(v, '[%<]') then
        res.opFun = function(n) return n < num end
      end
    elseif string.match(v, '[%-]') then 
      local tag = v:gsub('%-', '')
      repeat tag = tag:gsub('% ', '') until tag == tag 
        
      table.insert(res.exclude, tag)
    elseif string.match(v, '[%+]') then 
      local tag = v:gsub('%+', '')
      repeat tag = tag:gsub('% ', '') until tag == tag
        
      table.insert(res.include, tag)
    else   
      repeat v = v:gsub('% ', '') until v == v
      table.insert(res.include, v)
    end      
  end
  return res  
end

function noteDurD(tb)
  local distributions = {}
   
  local pitch     = {0,0,0,0,0,0,0,0,0,0,0,0}
  local duration  = copy(pitch)
  local pitch_dur = copy(pitch)
    
  local pitch_c     = copy(pitch)
  local duration_c  = copy(pitch)
    
  for n,note in ipairs(tb.notes) do
    note = note %12
    pitch[note+1] = pitch[note+1] + 1
            
    local dur = tb.dur[n]      
    pitch_dur[note+1] = pitch_dur[note+1] + dur
      
    if dur[dur] == nil then dur[dur] = 0 end
    if duration_c[dur] == nil then duration_c[dur] = 0 end      
      
    dur[dur] = dur[dur] + dur
      
    pitch_c[note+1] = pitch_c[note+1] + 1 
    duration_c[dur] = duration_c[dur] + 1
  end 
    
  return {pitch      = pitch,
          pitch_c    = pitch_c,            
          duration   = duration,
          pitch_dur  = pitch_dur,
          duration_c = duration_c} 
end

function ScannerH2(o, typ) 
  local line  = 0
  local notes = {}  
  
  for i,len in ipairs(o.len) do  
    local dur = 0
    local str
    
    line = line + len      
 
    if len > 0 then
      if (o.off[i] > 0) then 
        dur = o.off[i]
      elseif i < #o.len and o.len[i+1] > 0 then  
        dur = o.len[i+1]
      else  
        dur = o.pat_len - line + 1
      end     
    end
    
    if dur > 0 then
           
      if typ == 'mono' then  
        local note = o.note[i] % 12
        
        table.insert(Stats.unique_notes, noteInScale(note%12))      
        table.insert(notes,{line = line, note =  note, dur = dur, oct = o.oct[i]})
        Stats.pitch_duration_dst[note+1] = Stats.pitch_duration_dst[note+1] + dur 
                
      elseif typ == 'chord' then
        for k,v in ipairs(o.notes[i]) do  
          local note = v % 12
          table.insert(Stats.unique_notes, noteInScale(note))
          table.insert(notes,{line = line, note =  note, dur = dur})  
          Stats.pitch_duration_dst[note+1] = Stats.pitch_duration_dst[note+1] + dur
        end        
      end
    end        
  end  
  
  return notes --, line
end


function ScannerH() 
  local line  = 0  
  local notes = {}
  
  Stats.unique_notes       = {}
  Stats.pitch_duration_dst = {0,0,0,0,0,0,0,0,0,0,0,0}
  
  if MW.phr.event_type == 1 then
  
    notes =  ScannerH2 (Nptr, 'mono')      
  
  elseif MW.phr.event_type == 2 then
   
    for i=1,12 do    
      local tb
      local ptr = MW.phr[MW.phr.no].notes[i]
            
      tb, line =  ScannerH2 (ptr, 'mono')     
      tableConcat(notes, tb)
    end  
  end
  
  table.sort(notes, function(a,b) return a.line < b.line end) 
  
  notes.fx = MW.phr[MW.phr.no].notes[1].fx
  return notes  
end


function ScannerChordH() 
  Stats.unique_notes       = {}
  Stats.pitch_duration_dst = {0,0,0,0,0,0,0,0,0,0,0,0}
 
  local notes = ScannerH2(CHptr, 'chord')    
  table.sort(notes, function(a,b) return a.line < b.line end) 
  return notes  
end


function phraseScannerMonoChrd (o) -- , Line_tb)  

  local Markov_NS = {}
  local Markov_LN = {}
  
  local line = 0
  
  local min_len 
  if MW.phr.editor_chords then
    min_len = math.min( unpack({#o.len, #o.notes}) )
  else
    min_len = math.min( unpack({#o.len, #o.note}) )
  end
  
  for i=1,min_len do  
    local len = o.len[i]
    local dur = 0
    local str
    
    line = line + len      
 
    if len > 0 then  
      if i < #o.len and o.off[i] > 0 and o.off[i] < o.len[i+1] then
        dur = o.off[i]
      elseif i < #o.len and o.len[i+1] > 0 then  
        dur = o.len[i+1]    
      elseif o.off[i] > 0 then 
        dur = o.off[i]      
      else  
        dur = o.pat_len - line + 1
      end     
    end

    if dur > 0 then    
      if Stats.note_strings_length ~= '' and Scan.scale_key_on == false then
        local len_tb = splitDl (Stats.note_strings_length, ',')
        local _, dur2 = nearestValue(len_tb, dur) 
        dur = dur2       
      end     
    
      if Rand.new_add_lock ~= 3 then 
        if Stats.note_len_distribution[dur] then
          Stats.note_len_distribution[dur] = Stats.note_len_distribution[dur] + 1
        else
          Stats.note_len_distribution[dur] = {}
          Stats.note_len_distribution[dur] = 1
        end
      end
           
      if MW.phr.editor_pattern then  
                
        local oct    = o.oct[i]
        local note   = o.note[i]
        
        if note < 0 then 
          note = note + 12
          if oct >= 1 then oct = oct - 1 end
        end
      
        note,oct = oct0(clampOct(note+oct*12, Scan.Markov_oct_root,Scan.Markov_oct_range))
            
        str = note + oct*12 .. ',' .. dur
        
        if Rand.new_add_lock ~= 3 or Scan.scale_key_on then 
          table.insert(Stats.unique_notes, noteInScale(note%12))  

          Stats.pitch_duration_dst[note+1] = Stats.pitch_duration_dst[note+1] + dur     
        end
        
      elseif MW.phr.editor_chords  then

        str  = ''
        o.notes[i] = unique(map(o.notes[i], function(v) return clampOct(v, Scan.Markov_oct_root,Scan.Markov_oct_range) end))

        for k,v in ipairs( o.notes[i]) do           
          --local nn = noteInScale(v%12) 
                    
          str = str .. v
     
          if k ~= # o.notes[i] then
            str = str .. '_'
          else
            str = str .. ','  .. dur
          end   
          
          if Rand.new_add_lock ~= 3 or Scan.scale_key_on then
            table.insert(Stats.unique_notes, v)
            Stats.pitch_duration_dst[v%12+1] = Stats.pitch_duration_dst[v%12+1] + dur
          end
        end        
      end
      table.insert(Markov_NS, str)
    end        
  end 
 
  return Markov_NS  --, Markov_LN  
end


function phraseScannerPoly()

  local fun = function(tr,dst,col) 
    local line = 0
    
    for i,len in ipairs(tr.len) do      
      local dur = 0          
      line = line + len      
 
      if len > 0 then
        if (tr.off[i] > 0) then 
          dur = tr.off[i]
        elseif i < #tr.len and tr.len[i+1] > 0 then  
          dur = tr.len[i+1]
        else  
          dur = tr.pat_len - line + 1
        end     
      end
      
      if dur > 0 then
        if dst[line] == nil then
          dst[line] = {}
        end
        
        local note = tr.note[i]
        
        if Rand.new_add_lock ~= 3 or Scan.scale_key_on then 
          table.insert(Stats.unique_notes, noteInScale(note%12))  

          Stats.pitch_duration_dst[note+1] = Stats.pitch_duration_dst[note+1] + dur     
        end
                 
        table.insert(dst[line], {note = note + tr.oct[i]*12, len = len, dur = dur, col = col})         
      end
    end   
  end
  
  local tb, NS     = {},{}
  Scan.markow_poly = {}
  
  for col=1,12 do
    fun(POLYptr[col],tb,col)
  end
  
  local order = {}
  for key,v in pairs(tb) do
    table.insert(order,key)
  end
  table.sort(order)   
  
  for n,ix in ipairs(order) do
  
    local ns = ''
    local durs,notes,cols = {},{},{}
    for _,v in ipairs(tb[ix]) do
    
      if UI.markov_L.value > 0 and v.dur > UI.markov_L.value then
        v.dur = UI.markov_L.value
      end
    
      table.insert(durs, v.dur)
      table.insert(cols, v.col)
      table.insert(notes, {note = v.note, oct = v.oct, dur = v.dur, col = v.col})  
    end
    local max_dur = math.max(unpack(durs))          
    
    local str = table.concat(map(tb[ix], function(v) return v.note end), '_') .. ',' .. max_dur
 
    if Scan.markow_poly[str] == nil then
      Scan.markow_poly[str]         = {}
      Scan.markow_poly[str].notes   = {}
      Scan.markow_poly[str].max_dur = max_dur              
    end         
    
    table.insert(Scan.markow_poly[str].notes, notes)  
        
    local nn = n+1
    while nn <= #order and order[nn] < ix + max_dur do  
     
      local key  = {}
      local subs = {}
      for _,sub in ipairs(tb[order[nn]])do
        
        if member(sub.col, cols) then              
          table.insert(key, sub.note)
          table.insert(subs, {note = sub.note, oct = sub.oct, dur = sub.dur, col = sub.col, ln = order[nn] - ix})     
        end
      end
          
      if #key > 0 then
        local n = #Scan.markow_poly[str].notes 
            
        if Scan.markow_poly[str].notes[n].sub_notes == nil then 
          Scan.markow_poly[str].notes[n].sub_notes = {}
        end
            
        key = table.concat(key,'_') .. ',' .. max_dur
        if Scan.markow_poly[str].notes[n].sub_notes[key] == nil then
          Scan.markow_poly[str].notes[n].sub_notes[key] = {}
        end
            
        table.insert(Scan.markow_poly[str].notes[n].sub_notes[key], subs)            
      end       
      nn = nn + 1
    end             
    
    table.insert(NS, str)
  end   
      
  return NS
end


function phraseScanner()      
  
  UI.markov_first_N.items = {'Random'}
  UI.markov_first_N.value = 1   
      
  if Scan.Markov_new_add_lock == 1 or Scan.scale_key_on then
    Stats.unique_notes       = {}
    Stats.pitch_duration_dst = {0,0,0,0,0,0,0,0,0,0,0,0}
  end  
           
  if Scan.Markov_new_add_lock == 1 then 
    Stats.note_strings = {}
    Scan.Markov_process = nil

    if MW.phr.editor_poly then
      Stats.note_strings[1] = phraseScannerPoly() 
    elseif MW.phr.editor_chords then 
      Stats.note_strings[1] = phraseScannerMonoChrd(CHptr)    
    elseif MW.phr.editor_poly == false then
      Stats.note_strings[1] = phraseScannerMonoChrd(Nptr)  
    end
    
  elseif Scan.Markov_new_add_lock == 2 then      
    
    local tmp_N, tmp_NS, tmp_O, tmp_L = {},{},{},{}
    
    if MW.phr.editor_poly then
      tmp_NS = phraseScannerPoly() 
    elseif MW.phr.editor_chords then 
      tmp_NS = phraseScannerMonoChrd(CHptr)  
    else
      tmp_NS = phraseScannerMonoChrd(Nptr)  
    end

    table.insert(Stats.note_strings, tmp_NS)  
  --  sprint('Added ' .. #Stats.note_strings[#Stats.note_strings] .. ' notes' )
  end   
  
  if #Stats.note_strings > 1 then
    sprint('Loaded ' ..  #Stats.note_strings .. ' phrases')
  end
  
  if Scan.Markov_process then 
    Scan.Markov_process.params = ''
    Scan.Markov_process.ns = Stats.note_strings
  end
end
      

local function phraseScaleKeyStore(tb, scale, key, transpose, file_name, pdd)

  if tb[scale] then
    if tb[scale][key] then
      
      table.insert(tb[scale][key].pdd, pdd) 
      table.insert(tb[scale][key].file_name, file_name)
      table.insert(tb[scale][key].transpose, transpose)            
    else
      tb[scale][key] = {}
      
      tb[scale][key].pdd       = {}
      tb[scale][key].file_name = {}
      tb[scale][key].transpose = {}
      
      table.insert(tb[scale][key].pdd, pdd)  
      table.insert(tb[scale][key].file_name, file_name)
      table.insert(tb[scale][key].transpose, transpose)      
    end
  else
    tb[scale] = {}
    tb[scale][key] = {}
    
    tb[scale][key].pdd = {}
    tb[scale][key].file_name = {}
    tb[scale][key].transpose = {}
    
    table.insert(tb[scale][key].pdd, pdd) 
    table.insert(tb[scale][key].file_name, file_name)
    table.insert(tb[scale][key].transpose, transpose)     
  end 
end


local function scanChordsFw(tb, dst)
  if #tb > 0 then
  
    local rlen = Scan.recomb_len
    local alen = Scan.recomb_anchor
         
    for i=1,#tb do
      local flattened_notes_i = deepFlatten(tb[i].notes)  
      
      if #flattened_notes_i >= rlen + alen then    
        
        local chord = map(tb[i].chord, function(v) return v%12 end)
        local str = table.concat(chord,'_')
        
        local notes = {}
        for j=1,#tb[i].notes do       
          notes[j] = funOverElOrTbRet(function(note,oct) return note + oct*12 end, tb[i].notes[j], tb[i].octs[j])  
        end       
         
        local k,next_chord
        local ok = false
        if i<#tb and #deepFlatten(tb[i+1].notes) >= rlen + alen then
          k  = i+1                 
          ok = true 
          next_chord = map(tb[i+1].chord, function(v) return v%12 end)          
        elseif i==#tb and #deepFlatten(tb[1].notes) >= rlen + alen then       
          k  = 1            
          ok = true
          next_chord = map(tb[1].chord, function(v) return v%12 end)
        end  
        
        if ok then 
          if dst[str] == nil then
            dst[str] = {}
          end  
          
          local flattened_notes_k = deepFlatten(tb[k].notes)
          
          local anchor    = {unpack(flattened_notes_k, #flattened_notes_k - alen + 1, #flattened_notes_k)}          
          local anchor_bw = {unpack(flattened_notes_i, #flattened_notes_i - alen + 1, #flattened_notes_i)}
          
          local tb2 = {dur        = tb[i].dur, 
                       dif        = tb[i].dif, 
                       notes      = notes,
                       anchor     = anchor,
                       anchor_bw  = anchor_bw,
                       next_chord = next_chord}
                      
          table.insert(dst[str], tb2)           
        end
      end        
    end 
  end  
end


local function scaleKeyScan2(files,ix)

  local notes_o
  
  local fun = function (notes, pitch_duration_dst, transpose, file)
    local scale,key = findScaleKeyArg(unique(notes), pitch_duration_dst) 
      
    if scale > 1 then
      phraseScaleKeyStore(Scan.db_scale_key[ix], scale, key, transpose, file, pitch_duration_dst)
    end    
        
    local tb = {} -- print('notes_o') rprint(notes_o)
    harmonizeCh2P(notes_o, tb) 
 

    if Scan.db_chords_rc[ix] == nil then
      Scan.db_chords_rc[ix] = {}
    end
      
    local sk = scale .. '_' .. key
    if Scan.db_chords_rc[ix][sk] == nil then
      Scan.db_chords_rc[ix][sk] = {}
    end
      
    scanChordsFw(tb, Scan.db_chords_rc[ix][sk])
  end
  
  for file_no,file in pairs(files) do

    sprint('File: ' .. file)
    
    local notes, octs, offs, lens, pat_len, fx = Load.notesPtr(file, ix)        
    
    if ix == 'chrd' then
      local all_notes = {}
      
      for transpose = 0,11 do 
        notes_o    = {}
        notes_o.fx = fx
        
        local line               = 0
        local pitch_duration_dst = {0,0,0,0,0,0,0,0,0,0,0,0}                    
        
        for i,chord in ipairs(notes) do
          local dur      = 0          
          local ch_notes = {}
        
          line = line + lens[i]  
          
          if lens[i] > 0 then
            for j,note in ipairs(chord) do 
                                                           
              local nn, oct = oct0(note+transpose)               
              
              if (offs[i] > 0) then 
                dur = offs[i]
              elseif i < #lens and lens[i+1] > 0 then 
                dur = lens[i+1]
              else 
                dur = pat_len - line + 1
              end
              
              if dur > 0 then                        
                pitch_duration_dst[nn+1] = pitch_duration_dst[nn+1] +  dur  
                
                table.insert(ch_notes,nn)
                table.insert(all_notes,nn)
                
                if j == #chord then  
                                                    
                  table.insert(notes_o, {line = line, note = ch_notes, dur = dur, oct = oct})
                end
              end                        
            end
          end
        end 
        
        fun(all_notes, pitch_duration_dst, transpose, file)      
      end      
            
    else       
      for transpose = 0,11 do                
        notes_o    = {}     
        notes_o.fx = fx
             
        local notes_u            = {}              
        local pitch_duration_dst = {0,0,0,0,0,0,0,0,0,0,0,0}      
        
            
        for col=1,#notes do
          local fn = map2(notes[col], octs[col], function(n,o) return n + o*12 + transpose end)
          local notes_col = map(fn, function(v) return v%12 end)
          octs[col] = map(fn, function(v) return math.floor(v/12) end)
          
          local line = 0    
         
          for i,len in ipairs(lens[col]) do
            local dur = 0
                            
            line = line + len        
        
            if len > 0 then
              if (offs[col][i] > 0) then     
                dur = offs[col][i]
              elseif i < #lens[col] and lens[col][i+1] > 0 then  
                dur = lens[col][i+1]
              else     
                dur = pat_len - line + 1
              end
            end        
         
            if dur > 0 then                      
              pitch_duration_dst[notes_col[i]+1] = pitch_duration_dst[notes_col[i]+1] +  dur  
         
              table.insert(notes_u, notes_col[i])
              
              if ix == 'mono' then              
                table.insert(notes_o, {line = line, note = notes_col[i], dur = dur, oct = octs[col][i]})
              else
                local poly_note, poly_dur,poly_oct = {},{},{}  
                
                poly_dur[col]  = dur
                poly_oct[col]  = octs[col][i]                
                poly_note[col] = notes_col[i]  
                
                table.insert(notes_o, {line = line, note = poly_note, dur = poly_dur, oct = poly_oct})                               
              end
            end   
          end                
        end
        
        if ix == 'poly' then   
          table.sort(notes_o, function (a,b) return a.line < b.line end)  
          
          local res = {}
          if #notes_o > 1 then
            table.insert(res,notes_o[1])
            
            for i=2, #notes_o do
              if notes_o[i].line == res[#res].line then
              
                for k,v in pairs(notes_o[i].note) do
                  res[#res].note[k] = v
                end
                
                for k,v in pairs(notes_o[i].dur) do
                  res[#res].dur[k] = v
                end
                
                for k,v in pairs(notes_o[i].oct) do
                  res[#res].oct[k] = v
                end                    
              else
                table.insert(res,notes_o[i])                
              end 
            end
            notes_o = res
          end
        end
        
        fun(notes_u, pitch_duration_dst, transpose, file)   
      end      
    end      
                            
    if file_no%5 == 0 then
      coroutine.yield()
    end
  end  
end


function scaleKeyScan()
  Scan.db_chords_rc = {}
  
  Scan.db_scale_key = {}
  Scan.db_scale_key.mono = {}
  Scan.db_scale_key.poly = {}
  Scan.db_scale_key.chrd = {}
    
  scaleKeyScan2(getFiles('.seq' ),'mono')    
  scaleKeyScan2(getFiles('.seqs'),'poly')  
  scaleKeyScan2(getFiles('.chrd'),'chrd')  

  sprint('Done')
end


function libScan(type)  
  
  local key   = UI.keys_pu.value  
  local scale = UI.scales_pu.value
  
  if scale == 1 then return end
    
  local ix = getEventStr()      
  
  local scan = Scan.db_scale_key[ix][scale][key] 
  if not scan then sprint('Not found')return end  
  
  local metric = tonumber(UI.scan_ec.text)
  
  local pitch_duration_dst = copy(Stats.pitch_duration_dst) 
 
  local len   = #UI.lib_pu.items  
  local co_tb = {}  
  local filename_src = UI.lib_pu.items[UI.lib_pu.value]
                  
  local found = 0
                  
  for i,pdd in ipairs (scan.pdd) do     
    if type == 'correlation' then   
      local c = pearson2tbs(pitch_duration_dst, pdd)  
      if c > metric then   -- filename_src ~= scan.file_name[i] and      
        found = found + 1
        sprint('Found ' .. found)
        
        table.insert(co_tb, {metric = c, file_name = scan.file_name[i], tr = scan.transpose[i]})
      end                
    elseif type == 'entropy' then   
      local ec = entropyCross(normalize(Stats.pitch_duration_dst),normalize(scan.pdd[i]))  
      if ec < metric then 
        found = found + 1
        sprint('Found ' .. found)
        
        table.insert(co_tb, {metric = ec, file_name = scan.file_name[i], tr = scan.transpose[i]})
      end
    end
  end   
  
  if type == 'correlation' then 
    table.sort(co_tb, function (a,b) return a.metric > b.metric end) 
  elseif type == 'entropy' then
    table.sort(co_tb, function (a,b) return a.metric < b.metric end) 
  end
    
  Lib.transpose = map(co_tb, function(v) return v.tr end)

  return remExt(map(co_tb, function(v) return v.file_name end))
end


function getChordNotesDistrib() 
  
  Stats.unique_notes       = {}
  Stats.pitch_duration_dst = {0,0,0,0,0,0,0,0,0,0,0,0}
    
  local line = 0
 
  for i=1,#CHptr.len do  
    local len = CHptr.len[i]
    local dur = 0
    local str
    
    line = line + len      
 
    if len > 0 then
      if (CHptr.off[i] > 0) then 
        dur = CHptr.off[i]
      elseif i < #CHptr.len and CHptr.len[i+1] > 0 then  
        dur = CHptr.len[i+1]
      else  
        dur = CHptr.pat_len - line + 1
      end     
    end

    if dur > 0 then    
      for k,v in ipairs(CHptr.notes[i]) do    
        table.insert(Stats.unique_notes, v)
        Stats.pitch_duration_dst[v%12+1] = Stats.pitch_duration_dst[v%12+1] + dur
      end   
    end        
  end 
end


function blockRecomb() 
  local notes = {}
  local fun = function(pat,sel)      
    local note = pat.note_value    
     
    if (note < 120) then       
      local nn, oct = oct(note)
      table.insert(notes, {notes = nn, oct = oct, line = ln})              
    end
  end

  blockFunApply(fun)
end   


function recombFun2(scale, key, ...)
  
  local distribution     = {} 
  local distribution_dur = {}
  
  local last_chord_distrib     
  local last_chord_distrib_dur  
  
  local loct = Scan.recomb_loct
  local hoct = Scan.recomb_hoct
      
  local alen  = Scan.recomb_anchor
  local rclen = Scan.recomb_len
  
  local anchor_min = Scan.recomb_anchor_min
      
  local sk   = scale .. '_' .. key   
  local scan = Scan.db_chords_rc[getEventStr()][sk]   
  
  local msg = ''
  
  if not scan then 
    sprint('Chord not found.')    
    return true
  end
    
  local i       = 1
  local i_poly  = {1,1,1,1,1,1,1,1,1,1,1,1}
  local chord_i = 1
    
  local line = 1      
   
  local line_poly_add = {0,0,0,0,0,0,0,0,0,0,0,0}
      
  local phr_len    = renoise.song().selected_pattern.number_of_lines  
  local first_note = false
  
  local chord_already_used = {}
  
  local ptr, poly = {},{}
  local fix = MW.phr.event_type
  
  local poly_cont_st 
  
  if fix == 1 then
    initNotes(phr_len) 
    ptr = Nptr
  elseif fix == 2 then
    initPoly(phr_len)
    ptr = POLYptr 
  elseif fix == 3 then
    initChords(phr_len) 
    ptr = CHptr
  end
        
  local fillUpdate = function()
        
    if chord_i > 2 then    
      ptr.len[#ptr.len] = nil   
      
      if fix ~= 2 then
        if #arg>0 then
          fill(MW.phr.col, {clr_line = arg[1].line})
        else
          fill(MW.phr.col)
        end
      else
        for k=1,12 do 
          ptr[k].len[#ptr[k].len] = nil  
          
          if #arg>0 then
            fillPtr(k, ptr[k], fix, {line_off = arg[1].line})
          else
            fillPtr(k, ptr[k], fix, {line_off = 1})   
          end
        end   
      end
      
      updateSlots(); return true
    else
      sprint('Recomb falied at chord ' .. chord_i); return false
    end
  end
  
  local getMetric = function(distribution)
    local d = {} 
    filter(distribution, function(v) if v > 0 then table.insert(d,v) end end)
    table.sort(d, function(a,b) return a>b end)  
  
    local slope, r2 =  getSlopeR2(enumFromTo(#d,1), d)  
    return math.abs(-1 - slope)/2 +  (1 - r2) 
  end
  
  local fill_prt = {} 
  fill_prt[1] = function(tb)  
        
    for j=1,#tb.notes do
      local nn,oct =  oct0(tb.notes[j])    
          
      if oct < loct then
        oct = loct
      end
          
      if oct > hoct then
        oct = hoct
      end
          
      if j == 1 then
        table.insert(ptr.fx,line)
      end
          
      ptr.oct[i]  = oct
      ptr.note[i] = nn
                
      ptr.off[i]   = 0
      ptr.vel[i]   = 128        
       
      ptr.len[i+1] = tb.dur[j]    
          
      line = line + tb.dur[j]  
      
      i=i+1
    end
  end
  
  fill_prt[2] = function(tb)  

    for j=1,#tb.notes do   
      local len
      local len_add = {0,0,0,0,0,0,0,0,0,0,0,0}
      
      if j == 1 then
        len = tb.dif[j]
        
        if i==1 then
          len = 1
        end
        
        if poly_cont_st and chord_i == 1 then
          len = len + poly_cont_st
        end
        
        table.insert(ptr[1].fx,line)
      else
        len = tb.dif[j] - tb.dif[j-1]
      end
      
      local str = ''
      for _,v in pairs(tb.notes[j]) do
        str = str .. num2note(v) .. ' '
      end
         
      for col,note in pairs(tb.notes[j]) do 
        local len2 = len
        
        if line_poly_add[col] > 0 then 
          len2 = len + line_poly_add[col] - 1          
          line_poly_add[col] = 0
        end
               
        local nn,oct = oct0(note)    
          
        if oct < loct then
          oct = loct
        end
          
        if oct > hoct then
          oct = hoct
        end
        
        ptr[col].oct[i]  = oct
        ptr[col].note[i] = nn
        
        ptr[col].vel[i] = 128 
        ptr[col].off[i] = tb.dur[j][col] 
       
        ptr[col].len[i] = len2 
      end     
      
      local line_poly = {0,0,0,0,0,0,0,0,0,0,0,0}     
      line_poly_add   = copy(line_poly)
    
      for col=1,12 do
        local last = #ptr[col].off
        if last > 0 then
          line_poly[col] = sumTb(ptr[col].len) + ptr[col].off[last]  
        end  
      end
      line = math.max(unpack(line_poly))   
    
      for col=1,12 do  
        local last = #ptr[col].off
        if last > 0 then
          line_poly_add[col] = line - sumTb(ptr[col].len) -- + ptr[col].off[last])   
        else
          line_poly_add[col] = line
        end    
      end
       
      i=i+1
    end      
  end  
     
  fill_prt[3] = function(tb)  
        
    for j=1,#tb.notes do               
      ptr.off[i]   = 0
      ptr.vel[i]   = 128    
      ptr.len[i+1] = tb.dur[j]    
      
      ptr.notes[i] = {}
      
      if j == 1 then
        table.insert(ptr.fx,line)
      end
      
      for n,ch_note in ipairs(tb.notes[j]) do
        local nn,oct = oct0(ch_note)
        
        if oct < loct then
          oct = loct
        end
          
        if oct > hoct then
          oct = hoct
        end
        
        ptr.notes[i][n] = nn + oct*12
      end

      line = line + tb.dur[j]            
      i=i+1
    end
  end   
      
  local chordNotUsed = function(chord, ix)   
    for _,v in ipairs(chord_already_used) do
      if v.chord == chord and v.ix == ix then  
        return false
      end
    end
    return true
  end           
        
  local chord, random, anchor, next_chord
  
  if #arg>0 then 
    msg        = arg[1].msg
    line       = arg[1].line
    anchor     = arg[1].anchor
    next_chord = arg[1].chord 
      
    if fix ~= 2 then
      ptr.len[1] = arg[1].line
    else
      poly_cont_st = arg[1].line 
    end
    
    if Stats.metric_typ <= 2 then
      distribution     = arg[1].distribution 
      distribution_dur = arg[1].distribution_dur 
    end
    
    if not scan[next_chord] then
      sprint('Chord not found ' .. msg)
      return false
    end
  else     
    if scan == nil then return end
    
    if fix ~= 2 then
      ptr.len[1] = 1 
    end
     
    repeat            
      chord = randomKey(scan)         
      if #scan[chord] > 0 then
        random = math.random(#scan[chord])    
      end
    until random ~= nil      
      
    fill_prt[fix](scan[chord][random])  
      
    anchor     = scan[chord][random].anchor
    next_chord = table.concat(scan[chord][random].next_chord, '_')   
    
    if Stats.metric_typ == 1 then 
      distribution = {0,0,0,0,0,0,0,0,0,0,0,0}
      
      funOverElOrTb(function(note, dur, tb) local nn = note%12; tb[nn+1] = tb[nn+1] + dur end, 
                    scan[chord][random].notes, scan[chord][random].dur, distribution)     
                    
    elseif Stats.metric_typ >= 2 or Stats.metric_typ <= 4 then      
     
      funOverElOrTb(function(note, dur, tb1, tb2) table.insert(tb1, note); table.insert(tb2, dur) end, 
                    scan[chord][random].notes, scan[chord][random].dur, distribution, distribution_dur) 
    end 
  end   
  
  last_chord_distrib     = copy(distribution)
  last_chord_distrib_dur = copy(distribution_dur) 
    
  repeat 
    if not scan[next_chord] then 
      fillUpdate(); return true
    end 
  
    local matches = {} 
      
    for k,v in ipairs(scan[next_chord]) do
    
      if chordNotUsed(table.concat(v.next_chord,'_'), k) then
        local tb = {}
        
        local anchor2 = map({unpack(deepFlatten(v.notes), 1, alen)}, function(v) return v%12 end)  
      
        if Stats.metric_typ == 1 then
          local q_distrib = copy(distribution)   

          funOverElOrTb(function(note, dur, tb) local nn = note%12; tb[nn+1] = tb[nn+1] + dur end, 
                        v.notes, scan[next_chord][k].dur, q_distrib)
          
          local metric_phrase = entropyCross(normalize(distribution),normalize(q_distrib))        
        
          local q_distrib = {0,0,0,0,0,0,0,0,0,0,0,0}

          funOverElOrTb(function(note, dur, tb) local nn = note%12; tb[nn+1] = tb[nn+1] + dur end, 
                    v.notes, scan[next_chord][k].dur, q_distrib)
        
          local metric_motif  = entropyCross(normalize(last_chord_distrib),normalize(q_distrib))  
        
          tb = {i = k, metric = (metric_phrase + metric_motif)/2}
        
        elseif Stats.metric_typ == 2 then   -------------------------------------------------------------------------------

          funOverElOrTb(function(note, dur, tb1, tb2) table.insert(tb1, note); table.insert(tb2, dur) end, 
                    v.notes, scan[next_chord][k].dur, last_chord_distrib, last_chord_distrib_dur) 
                    
      
          local metric_motif  = (getMetric(last_chord_distrib) + getMetric(last_chord_distrib_dur))/2  
          local metric_phrase = (getMetric(distribution) + getMetric(distribution_dur))/2  
             
          tb = {i = k, metric = (metric_motif + metric_phrase)/2}
        --else
            
        end       
         
        if compareTbs(anchor,anchor2) >= anchor_min and #v.notes >= anchor_min + rclen then                 
          if Stats.metric_typ <= 2 then
            table.insert(matches, tb) 
          else
            local chstr = table.concat(v.next_chord, '_') 
             
            if matches[chstr] == nil then
              matches[chstr] = {}
              matches[chstr].i = {k}
              matches[chstr].count = 1
            else
              matches[chstr].count = matches[chstr].count + 1    
              table.insert(matches[chstr].i, k)
            end
          end  
        end
      end  
    end
    
    if lenTb(matches) == 0 then 
      fillUpdate(); return true
    elseif #arg == 0 then 
      chord_i = chord_i + 1; sprint('')
    end
    
    local ix 
    
    if Stats.metric_typ <= 2 then 
      table.sort(matches, function(a,b) return a.metric < b.metric end) 
        
      if #matches > 2 then  
        local rand_tb = expInterpolate(#matches)    
        local rand_n  = rand_tb[math.random(#rand_tb)] 
        ix = matches[rand_n].i    
      else
        ix = matches[1].i
      end
    else
      local tb = {}
      for i,v in pairs(matches) do
        table.insert(tb, {count = v.count, i = v.i})
      end
      
      table.sort(tb, function(a,b) return a.count > b.count end) 
      
      if #tb > 2 then        
        local rand_tb = expInterpolate(#tb)  
        local rand_n  = rand_tb[math.random(#rand_tb)]
               
        ix =  math.random(#tb[rand_n].i) 
      else
        local rand_n = math.random(#tb[1].i)
        ix = tb[1].i[rand_n]          
      end
    end
    
    fill_prt[fix](scan[next_chord][ix])   
    
    last_chord_distrib, last_chord_distrib_dur = {},{} 
    
    if Stats.metric_typ == 1 then

      funOverElOrTb(function(note, dur, tb) local nn = note%12; tb[nn+1] = tb[nn+1] + dur end, 
                    scan[next_chord][ix].notes, scan[next_chord][ix].dur, distribution)
                    
      last_chord_distrib = {0,0,0,0,0,0,0,0,0,0,0,0}

      funOverElOrTb(function(note, dur, tb) local nn = note%12; tb[nn+1] = tb[nn+1] + dur end, 
                    scan[next_chord][ix].notes, scan[next_chord][ix].dur, last_chord_distrib)
                    
    elseif Stats.metric_typ == 2 then    

      funOverElOrTb(function(note, dur, tb1, tb2) table.insert(tb1, note); table.insert(tb2, dur) end, 
                    scan[next_chord][ix].notes, scan[next_chord][ix].dur, distribution, distribution_dur)
                    
      funOverElOrTb(function(note, dur, tb1, tb2) table.insert(tb1, note); table.insert(tb2, dur) end, 
                    scan[next_chord][ix].notes, scan[next_chord][ix].dur, last_chord_distrib, last_chord_distrib_dur) 
    end
     
    anchor     = scan[next_chord][ix].anchor
    next_chord = table.concat(scan[next_chord][ix].next_chord, '_') 
    
    table.insert(chord_already_used, {chord = next_chord, ix = ix})            
  until line > phr_len      
      
  fillUpdate()  
  return false
end


function recombFun(...)    
  
  if Scan.recomb_random_sk then  
    local i = Scan.recomb_tries
    
    repeat
      local key   = math.random(12)
      local scale = math.random(2,3)         
      
      UI.keys_pu.value   = key
      UI.scales_pu.value = scale
    
      local rc = recombFun2(scale, key, ...)
      i=i-1
    until rc or i>0   
  else
    local key   = UI.keys_pu.value
    local scale = UI.scales_pu.value    
    
    if scale == 1 then 
      sprint('Select scale')
      return
    end    
    recombFun2(scale, key, ...)  
  end
end

function grebReverse()
  local notes, last_note = {}
  
  if MW.phr.editor_chords then  
    notes     = Chords.new(0) 
    last_note = GrebSeq(col, true, notes) 
  elseif MW.phr.editor_poly then 
    last_note = {}
    for i=1,12 do        
      notes[i]  = {}
      notes[i]  = Notes.new(0)  
      
      last_note[i] = {}
      last_note[i] = GrebSeq(i, false, notes[i]) 
    end
  else 
    notes     = Notes.new(0)  
    last_note = GrebSeq(MW.phr.col, false,  notes)        
  end 
  
  
  if MW.phr.editor_poly then   
    if emptyPolyTb(notes) then
      sprint('All tracks empty')
      return nil
    end
  
    for i=1,12 do  
      if sumTb(notes[i].len) > 0 then   -- sumTb(notes[1].len) > 0 then  
        local last_l  = notes[i].len[last_note[i]]   
        local first_l = notes[i].len[1]
    
        local last_note_off = notes[i].off[last_note[i]]  
  
        Reverse(notes[i].len)
        Reverse(notes[i].oct)
        Reverse(notes[i].off)
        Reverse(notes[i].note)  
  
        wrap(notes[i].len,1)
  
        notes[i].len[1] = first_l
  
        if last_note_off > 0 then
         notes[i].off[last_note[i]] = last_note_off
        end   
      end
    end
  else  
    if #notes.note == 0 then 
      sprint('Empty track')
      return nil
    end   
    
    local last_l  = notes.len[last_note]   
    local first_l = notes.len[1]
    
    local last_note_off = notes.off[last_note]  
  
    Reverse(notes.len)
    Reverse(notes.oct)
    Reverse(notes.off)
    Reverse(notes.note)  
  
    wrap(notes.len,1)
  
    notes.len[1] = first_l
  
    if last_note_off > 0 then
      notes.off[last_note] = last_note_off
    end
  end    
   
  return notes  
end

function recombContinue()
  
  if UI.scales_pu.value == 1 then
    sprint('Select scale')
    return
  end
  
  local notes = grebReverse()
  
  local next_line
  
  if MW.phr.editor_chords then
    notes     = ScannerH2(notes, 'chord')
    next_line = notes[#notes].line + notes[#notes].dur    
  elseif MW.phr.editor_poly then     
    local notes2 = {}
        
    for col,v in pairs(notes) do  
      
      local tb = ScannerH2(v, 'mono')
      
      if tb then 
        for i=1,#tb do 
          local ln = tb[i].line
         
          if not notes2[ln] then 
            notes2[ln]      = {} 
            notes2[ln].dur  = {}
            notes2[ln].oct  = {}
            notes2[ln].note = {}

            notes2[ln].dur[col]  = tb[i].dur
            notes2[ln].oct[col]  = tb[i].oct
            notes2[ln].note[col] = tb[i].note
          else
            notes2[ln].dur[col]  = tb[i].dur
            notes2[ln].oct[col]  = tb[i].oct
            notes2[ln].note[col] = tb[i].note
          end
        end
      end
    end  
    
    notes = {}
    for ln,v in pairs(notes2) do
      v.line = ln
      table.insert(notes, v)  
    end  
    
    table.sort(notes, function(a,b) return a.line < b.line end)
    
    next_line = notes[#notes].line + math.max(unpack(deepFlatten(notes[#notes].dur))) 
  else     
    notes     = ScannerH2(notes, 'mono')    
    next_line = notes[#notes].line + notes[#notes].dur 
  end  

  local tb = {}
  harmonizePhraseAuto(notes, tb, 'exit when first chord found')
  
  local rlen = Scan.recomb_len
  local alen = Scan.recomb_anchor
  
  if #tb>0 and #tb[1].notes >= alen then   
    
    local chord  = map(tb[1].chord, function(v) return v%12 end)    
    local anchor = deepFlatten({unpack(tb[1].notes, 1, alen)}) 
    
    local chord_str = table.concat(map(chord, function(v) return num2noteOnly(v) end),' ')
    local notes_str = table.concat(map(deepFlatten(tb[1].notes), function(v) return num2noteOnly(v) end),' ')
    local msg = '  | chord in block  ' .. chord_str .. '  notes  ' .. notes_str
        
    local tmp = MW.block_editing
    MW.block_editing = false   
    
    
    local distribution, distribution_dur = {},{}     
    if Stats.metric_typ == 1 then
      distribution = {0,0,0,0,0,0,0,0,0,0,0,0}
      
      funOverElOrTb(function(note, dur, tb) local nn = note%12; tb[nn+1] = tb[nn+1] + dur end, 
                    tb[1].notes, tb[1].dur, distribution)
                    
    elseif Stats.metric_typ == 2 then
      
      funOverElOrTb(function(note, dur, tb1, tb2) table.insert(tb1, note); table.insert(tb2, dur) end, 
                    tb[1].notes, tb[1].dur, distribution, distribution_dur)    
    end
          
    local arg = {msg = msg, line = next_line, anchor = anchor, chord = table.concat(chord,'_'), distribution = distribution, distribution_dur = distribution_dur}
    recombFun2(UI.scales_pu.value, UI.keys_pu.value, arg) 
    
    MW.block_editing = tmp
  else
    sprint('Can not continue')
  end
end


function recombMotif2(scale, key, ...) 
  
  local typ,i   
  if MW.phr.event_type == 1 then
    i    = 1  
    typ  = 'mono'   
  elseif MW.phr.event_type == 2 then
    i    = {1,1,1,1,1,1,1,1,1,1,1,1}
    typ  = 'poly'          
  else
    return  
  end   
  
  if Scan.db_motifs[typ] == nil then
    sprint('Empty, scan motifs first')
    return
  end
  
  local line = 1
  
  local line_poly     = {0,0,0,0,0,0,0,0,0,0,0,0}
  local line_poly_add = {0,0,0,0,0,0,0,0,0,0,0,0}
  
  local last_dur   
  
  local distribution     = {} 
  local distribution_dur = {}
      
  local sk = scale .. '_' .. key
  
  local scan = Scan.db_motifs[typ][sk]
  if not scan then sprint('Not found') return end    
  
  local already_used = {}
            
     
  local phr_len    = renoise.song().selected_pattern.number_of_lines  
  local first_note = false
        
  local fillUpdate = function()
    if line > 8 then    
    --  Nptr.len[#Nptr.len] = nil   
      if typ == 'mono' then  
        if #arg>0 then
          fill(MW.phr.col, {clr_line = arg[1].line, pat_len  = renoise.song().selected_pattern.number_of_lines})
        else
          fill(MW.phr.col)
        end  
      else
        for col=1,12 do
          assignNoteTbs(col)
          if #arg>0 then
            fill(col, {clr_line = arg[1].line, pat_len  = renoise.song().selected_pattern.number_of_lines, line_off = arg[1].line})  
          else
            fill(col, {line_off = 1})
          end
        end
      end  
      
      updateSlots()
      
      return true    
    else
      return false
    end
  end
  
  local fill_prt = {} 
  fill_prt.mono = function(tb)  
        
    for j=1,#tb.note do          
      Nptr.oct[i]  = tb.oct[j]
      Nptr.note[i] = tb.note[j]
                
      Nptr.off[i] = tb.off[j]
      Nptr.vel[i] = tb.vel[j] 
      Nptr.len[i] = tb.len[j]              
       
      if j ==  #tb.note then
        last_dur = tb.dur[j]
      end
      
      if j == 1 then
        table.insert(Nptr.fx,line)
        
        if last_dur and i > 1 then
          Nptr.len[i] = Nptr.len[i] + last_dur - 1
        end
      end
            
      line = line + tb.dur[j]  
          
      i=i+1
    end   
  end     
  
  fill_prt.poly = function(tb)
  
    local first_len = {}
     
    for col,tr in pairs(tb) do
      assignNoteTbs(col)
      
      for j=1,#tr.note do          
        local ii = i[col]
        
        local len  = tr.len[j] 
        local len2 = len
        
        if line_poly_add[col] > 0 and  j == 1 then
          len2 = len + line_poly_add[col] - 1
          line_poly_add[col] = 0                  
        end    
              
        Nptr.oct[ii]  = tr.oct[j]
        Nptr.note[ii] = tr.note[j]
                
        Nptr.off[ii] = tr.dur[j]
        Nptr.vel[ii] = tr.vel[j] 
        Nptr.len[ii] = len2   
        
        if j == 1 then 
          local ln = sumTb(Nptr.len)
          table.insert(first_len, ln) 
        end
       
        i[col]=i[col]+1
      end  
    end 
    
    table.insert(POLYptr[1].fx, math.min(unpack(first_len)))
    
    line_poly     = {0,0,0,0,0,0,0,0,0,0,0,0} 
    line_poly_add = copy({0,0,0,0,0,0,0,0,0,0,0,0})
    
    for col=1,12 do
      assignNoteTbs(col)  
      local last = #Nptr.off 
      if last > 0 then
        line_poly[col] = sumTb(Nptr.len) + Nptr.off[last]  -- + last_dur_poly[col]
      end
    end
    line = math.max(unpack(line_poly))        
         
    for col=1,12 do  
      assignNoteTbs(col)
      local last = #Nptr.off
      if last > 0 then
        line_poly_add[col] = line - sumTb(Nptr.len) -- +  Nptr.off[last])     
      else
        line_poly_add[col] = line
      end
    end 
  end   
  
  init()                 
          
  local random = randomKey(scan) 
  if not scan[random] then sprint('Empty') end
  
  if #arg>0 then
    distribution     = arg[1].d
    distribution_dur = arg[1].dd  
  end
      
  if #arg>0 and typ == 'poly' then
    for i=1,12 do
      line_poly_add[i] = arg[1].line
    end
  end

  fill_prt[typ](scan[random])  
  
  if typ == 'mono' then
    if #arg>0 then
      line        = arg[1].line
      Nptr.len[1] = arg[1].line        
    else
      Nptr.len[1] = 1       
    end   
  end
   
  local getDistribSlopR2 = function(distribution)
    local d = {} 
    filter2(distribution, function(v) if v > 0 then table.insert(d,v) end end)
    table.sort(d, function(a,b) return a>b end)                                              
  
    local slope, r2 =  getSlopeR2(enumFromTo(#d,1), d)    
    return  math.abs(-1 - slope)/2 +  (1 - r2)  
  end
  
  local nfDistrib = {}  
  nfDistrib.mono = function(src)  
    for n,note in ipairs(src.note) do
      local fn = note + src.oct[n]*12
    
      table.insert(distribution, fn)
      table.insert(distribution_dur, src.dur[n])
    end 
  end
  
  nfDistrib.poly = function(src)
    for _,v in pairs(src) do
      nfDistrib.mono(v)
    end
  end
  
  nfDistrib[typ](scan[random])
    
  repeat 
    local matches = {}      
    
    for i,v in ipairs(scan) do
      local d  = copy(distribution)
      local dd = copy(distribution_dur)
      
      local metric_dur   = getDistribSlopR2(dd) 
      local metric_notes = getDistribSlopR2(d)  

      table.insert(matches, {i=i, metric = (metric_dur + metric_notes)/2})
    end
    
    table.sort(matches, function(a,b) return a.metric < b.metric end)
    
    local ix
    
    repeat 
      if #matches > 2 then  
        local rand_tb = expInterpolate(#matches)    
        local rand_n  = rand_tb[math.random(#rand_tb)] 
        ix = matches[rand_n].i  
      else
        ix = matches[1].i
      end     
    until not member(ix,already_used) 
    
    table.insert(already_used,ix)      
    
    fill_prt[typ](scan[ix])   
    nfDistrib[typ](scan[ix])
         
  until line > phr_len      
    
  fillUpdate()  
  return false
end

function recombMotif()    
  local i = Scan.motif_recomb_tries
  
  if Scan.motif_recomb_random_sk then 
   
    repeat
      local key   = math.random(12)
      local scale = math.random(2,3)         
      
      UI.keys_pu.value   = key
      UI.scales_pu.value = scale
    
      local rc = recombMotif2(scale, key)
      i=i-1
    until rc or i>0   
  else
    local key   = UI.keys_pu.value
    local scale = UI.scales_pu.value    
    
    if scale == 1 then 
      sprint('Select scale')
      return
    end       
    
    repeat  
      local rc = recombMotif2(scale, key)  
      i=i-1
    until rc or i>0  
  end
end


function scanMotifs()

  local ix  
  if MW.phr.event_type == 1 then 
    ix = 'mono'      
  elseif MW.phr.event_type == 2 then
    ix = 'poly'  
  else 
    sprint('Chords are not supported')
    return  
  end  
  
  if not UI.scan_motif_add.value then 
    Scan.db_motifs[ix] = {}
  end
  
  local dst = Scan.db_motifs[ix]
    
  for phr_no,phr in ipairs(MW.phr) do  
  
    if ix == 'mono' then  
      local col     = MW.phr.col
      local vel     = phr.notes[col].vel
      local lens    = phr.notes[col].len
      local octs    = phr.notes[col].oct    
      local offs    = phr.notes[col].off    
      local notes   = phr.notes[col].note
      local pat_len = phr.notes[col].pat_len  
    
      sprint('Mono phrase ' .. phr_no)
      
      for transpose = 0,11 do        
              
        local tb = {vel     = copy(vel),
                    len     = copy(lens),
                    off     = copy(offs),                     
                    oct     = {},
                    dur     = {},
                    note    = {}, 
                    pat_len = phr.notes[1].pat_len}        
    
        local pitch_duration_dst = {0,0,0,0,0,0,0,0,0,0,0,0}      
        
        local fn = map2(notes, octs, function(note, oct) return note + oct*12 + transpose end)
          
        local line = 0
      
        for i,len in ipairs(lens) do
          local dur = 0
        
          line = line + len        
        
          if len > 0 then
            if (offs[i] > 0) then     
              dur = offs[i]
            elseif i < #lens and lens[i+1] > 0 then  
              dur = lens[i+1]
            else     
              dur = pat_len - line + 1
            end
          end        
         
          if dur > 0 then           
            local note = fn[i]%12          
            pitch_duration_dst[note+1] = pitch_duration_dst[note+1] + dur  
              
            tb.dur[i]  = dur
            tb.oct[i]  = math.floor(fn[i]/12)
            tb.note[i] = note
          end   
        end     
        
        local scale,key = findScaleKeyArg(unique(tb.note), pitch_duration_dst) 
        local sk = scale .. '_' .. key
                
        if dst[sk] == nil then dst[sk] = {} end        
        table.insert(dst[sk], tb)   
      end    
      
    elseif ix == 'poly' then  

      sprint('Poly phrase ' .. phr_no)
      
      for transpose = 0,11 do        
        
        local tb,notes = {},{}            
        local pitch_duration_dst = {0,0,0,0,0,0,0,0,0,0,0,0} 
            
        for col=1,12 do        
          local vel     = phr.notes[col].vel
          local lens    = phr.notes[col].len
          local octs    = phr.notes[col].oct    
          local offs    = phr.notes[col].off    
          local notes   = phr.notes[col].note
          local pat_len = phr.notes[col].pat_len
          
          tb[col] = {}
          tb[col] = {vel     = copy(vel),
                     len     = copy(lens),
                     off     = copy(offs),                     
                     oct     = {},
                     dur     = {},
                     note    = {}}
          
          if col == 1 then            
            tb[1].pat_len = phr.notes[1].pat_len        
          end                         
      
          local fn = map2(notes, octs, function(note, oct) return note + oct*12 + transpose end)
          
          local line   = 0
          local empty = true
          
          for i,len in ipairs(lens) do
            local dur = 0
        
            line = line + len        
        
            if len > 0 then
              empty = false
              
              if (offs[i] > 0) then     
                dur = offs[i]
              elseif i < #lens and lens[i+1] > 0 then  
                dur = lens[i+1]
              else     
                dur = pat_len - line + 1
              end
            end        
         
            if dur > 0 then           
              local note = fn[i]%12          
              pitch_duration_dst[note+1] = pitch_duration_dst[note+1] + dur  
              
              tb[col].dur[i]  = dur
              tb[col].oct[i]  = math.floor(fn[i]/12)
              tb[col].note[i] = note
              
              table.insert(notes, note)
            end
          end
          
          if empty then
            tb[col] = nil
          end           
        end  
         
        local scale,key = findScaleKeyArg(unique(notes), pitch_duration_dst) 
        local sk = scale .. '_' .. key
                
        if dst[sk] == nil then dst[sk] = {} end        
        table.insert(dst[sk], tb)                   
      end  
    end 
  end
end

function motifContinue()
  local typ
  local notes = {}  
  
  if MW.phr.event_type == 1 then
    typ   = 'mono'   
    notes = Notes.new(0)   
    
    GrebSeq(MW.phr.col, false,  notes)
    
    if #notes.note == 0 then 
      sprint('Empty track')
      return nil
    end
  elseif MW.phr.event_type == 2 then
    typ  = 'poly'
    for i=1,12 do
      notes[i] = Notes.new(0)
      GrebSeq(MW.phr.col, false,  notes[i])
    end  
    
    if emptyPolyTb(notes) then
      sprint('Empty poly')
      return 
    end  
  else
    return  
  end   
  
  local dur     

  local distribution     = {} 
  local distribution_dur = {}
  
  local getDistrib = {}
  
  getDistrib.mono = function(src) 
    local line = 0
    
    for i,len in ipairs(src.len) do
      local fn = src.note[i] + src.oct[i]*12      
      table.insert(distribution, fn)
    
      line = line + len
    
      dur = 0    
      if len > 0 then          
        if (src.off[i] > 0) then 
          dur = src.off[i]
        elseif i < #src.len and src.len[i+1] > 0 then  
          dur = src.len[i+1]
        else  
          dur = src.pat_len - line + 1
        end     
      end 
    
      if dur > 0 then
        table.insert(distribution_dur, dur)
      end
    end
    return line 
  end
  
  getDistrib.poly = function(src) 
    local lines = {}
    
    for i=1,12 do
      lines[i] = getDistrib.mono(src[i])
    end 
    
    return math.max(unpack(lines))
  end

  local next_line = getDistrib[typ](notes) + dur  
  
  local tb = {line = next_line, d = distribution, dd = distribution_dur}    

  recombMotif2(UI.scales_pu.value, UI.keys_pu.value, tb)  
end

