
--------------------------------------- UI arp

function ui_arp()

  local vb = renoise.ViewBuilder()
    
  local arp_set_fn, arp_dir_sw, arp_row
     
  local arp_set_chb = vb:checkbox {
    value = false,
    tooltip = 'Set selected arp pattern to all chords',
    notifier = function(value) 
      if value then
        arp_set_fn()
      end
    end
  }    
  
  local function make_arpNote(ix)  
    Arp.ui_rows[ix] = vb:row{
      style = 'invisible',      
      vb:checkbox {
        value = Arp.skip_note_tb[Arp.form_no][ix],
        tooltip = 'Skip',
        notifier = function(value)        
          Arp.skip_note_tb[Arp.form_no][ix] = value
          fill(MW.phr.col) 
        end
      },
  
      vb:valuebox {
        tooltip = 'Interval to add or sub',
        min = -24,
        max = 24,
        value = Arp.forms[Arp.form_no][ix], 
        width = 50,
        notifier = function(value) 
          Arp.forms[Arp.form_no][ix] = value
          fill(MW.phr.col) 
        end
      }
      
    }  
    return Arp.ui_rows[ix]
  end    

  function UI.updateArp()
    for i=1,#Arp.ui_rows do
      arp_row:remove_child(Arp.ui_rows[i])
    end
         
    Arp.ui_rows = {}         
        
    for i=1,#Arp.forms[Arp.form_no] do
      arp_row:add_child(make_arpNote(i))
    end 
  
    if MW.phr.editor_name == 'Arpeggiator' then
      fill(MW.phr.col) 
    end
  end
  
  local arp_pres_vb = vb:valuebox {
    tooltip = 'Arp pattern number',
    min = 1,
    max = #Arp.forms,
    value = 1, 
    width = 50,
    notifier = function(value) 
      Arp.form_no = value
      
      arp_dir_sw.value = Arp.directions[Arp.form_no] 
        
      UI.updateArp()        
     
      if arp_set_chb.value then
        arp_set_fn()
      end
    end
  }
  
  local arp_new_form_bt = vb:button {
    text = ' + ',
    tooltip = 'Add new random arp pattern',
    notifier =  function ()
      local len = math.random(8)
      
      table.insert (Arp.forms, {})
      table.insert (Arp.skip_note_tb, {})
            
      for i=1,len do  
        Arp.forms [#Arp.forms][i] =  math.random(6) - 3   
        
        if math.random(2) == 1 then
          Arp.skip_note_tb [#Arp.skip_note_tb][i] = true
        else 
          Arp.skip_note_tb [#Arp.skip_note_tb][i] = false 
        end
        
        table.insert (Arp.directions, math.random(2))  
      end      
      arp_pres_vb.max = #Arp.forms
      arp_pres_vb.value = #Arp.forms
    end 
  }
     
  local arp_remove_form_bt = vb:button {
    text = ' - ',
    tooltip = 'Remove arp pattern',
    notifier =  function ()      

      if Arp.form_no < #Arp.forms then

        for i=Arp.form_no+1,#Arp.forms do
          Arp.forms [i-1] = Arp.forms[i]
        end        
   
        Arp.forms[#Arp.forms] = nil    
        
        arp_pres_vb.max = #Arp.forms
      elseif Arp.form_no == #Arp.forms and #Arp.forms > 1 then

        Arp.form_no = Arp.form_no - 1
        arp_pres_vb.value = Arp.form_no - 1
        
        Arp.forms[#Arp.forms] = nil   
        
        arp_pres_vb.max = #Arp.forms     
      end  
      
      UI.updateArp()   
    end 
  }
  
  local arp_randomize    
  arp_randomize = function (arp_form_no)    
    
    local len = math.random(6)
      
    local skipProb = probTb ({ {false,2},{true,1} })
            
    local noteProb = {}            
    local tb ={{0,5},{1,8},{2,4},{3,2},{4,1},{-1,8},{-2,4},{-3,2},{-4,1}}
    local noteProb = probTb (tb)
    
    Arp.forms [arp_form_no] = {}
    Arp.skip_note_tb [arp_form_no] = {}
            
    for i=1,len do
      Arp.forms [arp_form_no][i] = noteProb[math.random(#noteProb)]
      
      if len > 1 then       
        Arp.skip_note_tb [arp_form_no][i] = skipProb[math.random(3)] 
      end 
    end   
    
    Arp.directions [arp_form_no] = math.random(2)
    arp_dir_sw.value = Arp.directions [arp_form_no]    
 
    local tb = filter(Arp.skip_note_tb[arp_form_no], function(v) return v == true end)
    if #tb == #Arp.skip_note_tb[arp_form_no] then 
      arp_randomize(arp_form_no)             
    end    
    
    UI.updateArp()
  end
  
      
  local arp_random_pres_bt = vb:button {
    text = ' Rand ',
    tooltip = 'Randomize arp patten',
    notifier =  function ()
      arp_randomize(Arp.form_no)
    end 
  } 
  
  arp_set_fn = function()      
    for i=1,#CHptr.arp.form do
      CHptr.arp.form[i] = Arp.form_no
    end
      
    UI.updateArp()
    
    updateSlots()     
        
    fill(MW.phr.col)
  end
  
  local arp_set_bt = vb:button {
    text = ' Set ',
    tooltip = 'Set arp to all chords',
    notifier =  function ()
      arp_set_fn()
    end 
  } 
  
  local arplib_pu = vb:popup {
    width = 100,
    value = Arp.last_lib_selection,
    items = Arp.files,
    tooltip = 'Saved presets',
    notifier = function(new_index)
    
      getArpFiles()
      
      local filename                  
      filename = MW.path .. Arp.files_full[new_index]    

      Arp.last_lib_selection = new_index
      
      readArpForms (filename,mw)   
      
      UI.updateArp()
    end
  } 

  local lib_arp_save_but = vb:button {
      text = ' Save ',
      tooltip = 'Save all arpeggiator forms.',
      notifier =  function () 
        local name = arplib_pu.items [arplib_pu.value]

        SavePresetPrompt(name, '.arp', write_arp_lib_popup, arplib_pu.items)
      end 
    }   

  local  arp_delete_but = vb:button {
    tooltip = 'Delete preset',
    text = ' Del ',
    notifier =  function ()          
      local name = Arp.files_full[arplib_pu.value]    
      
      os.remove (MW.path ..  name)   
        
      getArpFiles()         
      arplib_pu.items = Arp.files
    end 
  } 
 
  arp_dir_sw = vb:switch {
    width = 50,
    value = Arp.directions[Arp.form_no],
    tooltip = 'Arp forward forward and back',
    items = {'>', '><'},
    notifier = function(val)    
      Arp.directions [Arp.form_no] = val
      fill(MW.phr.col) 
    end  
  }  
                 
  arp_row = vb:row{
    margin = 0,   
    spacing = 8,
    
    vb:row{
      style = 'group',  
      vb:text {
        width = 20,
        text = ' ARP'
      },  
      arp_pres_vb,
      arp_new_form_bt,
      arp_remove_form_bt,

      arp_set_bt,
      arp_set_chb,
      
      vb:text {
        width = 40,
        text = ' Step'
      },  
      vb:valuebox {
        tooltip = 'Step',
        min = 1,
        max = 8,
        value = 2, 
        width = 50,
        notifier = function(value) 
          Arp.step = value
          fill(MW.phr.col) 
        end
      },
      arp_random_pres_bt,
      
      vb:text {
        width = 30,
        tooltip = 'Pattern length',
        text = ' Length'
      },  
      
      vb:button {
        text = ' - ',
        tooltip = 'Deccrease length',
        notifier =  function ()
          local ix = #Arp.forms[Arp.form_no]
          if ix > 1 then
            Arp.forms[Arp.form_no][ix] = nil 
            Arp.skip_note_tb[Arp.form_no][ix] = nil
            UI.updateArp()
          end
        end 
      },
      vb:button {
        text = ' + ',
        tooltip = 'Increase length',
        notifier =  function ()
          local ix = #Arp.forms[Arp.form_no]
          Arp.forms[Arp.form_no][ix+1] = 0 
          Arp.skip_note_tb[Arp.form_no][ix+1] = false 
          UI.updateArp()
        end 
      },
      arp_dir_sw,    
    },
    vb:row{
      style = 'group',  
      vb:text {
        width = 20,
        text = ' Lib'
      },  
      arplib_pu,
      lib_arp_save_but,
      arp_delete_but,      
    }
  }
  
  for i=1,#Arp.forms[Arp.form_no] do
    arp_row:add_child(make_arpNote(i))
  end
  
  UI.editor_main = vb:column {style = 'group'}
  
  UI.arplib_pu = arplib_pu
  UI.arp_randomize = arp_randomize
  UI.arp_pres_vb = arp_pres_vb
  return arp_row  
end  


