
local chooser_apply
local chooser_bot_row

local vb = renoise.ViewBuilder()


local function randomizeUI() -------------------------------------------------------------------------- 
      
  local gen_note_val_tx

  local gen_random_len_knob = vb:rotary {
    tooltip = 'Amount of maximum randomizationtation of distance between notes',
    min = 1,
    max = 16,
    value = Rand.note_len,
    width = 25,
    height = 25,
    notifier = function (val)
      Rand.note_len = val
    end
    } 
    
  local gen_random_note_knob = vb:rotary {
    tooltip = 'Amount of maximum randomization of notes',
    min = 1,
    max = 12,
    value = Rand.note,
    width = 25,
    height = 25,
    notifier = function (val)
      Rand.note = val
    end
  } 
 
  local gen_random_octave_knob = vb:rotary {
    tooltip = 'Amount of maximum randomization of octaves',
    min = 1,
    max = 3,
    value = 1,
    width = 25,
    height = 25,
    notifier   = function(val)
      Rand.octave = val
    end
  } 
    
   local gen_velocity_knob = vb:rotary {
     tooltip = 'Amount of maximum randomization of note velocities',
     min = 1,
     max = 126,
     value = 1,
     width = 25,
     height = 25,
     notifier  = function(val)
       Rand.note_vel = val
     end
   }  

   local gen_chprob_knob = vb:rotary {
     tooltip = 'Increase / decrease probability of randomizing using chord notes',
     min = 0,
     max = 10,
     value = Rand.chord_p,
     width = 25,
     height = 25,
     notifier  = function(val)
       Rand.chord_p = round(val)
     end
   } 

  local gen_note_rand_root_vb = vb:valuebox {
      tooltip = 'Root note that is to be randomized up and down',
      min = 1,
      max = 119,
      width = 50,
      value = Rand.root_note,
      notifier =  function (val) 
         Rand.root_note = val
         gen_note_val_tx.text = num2note(val)
      end 
    }  

  gen_note_val_tx = 
    vb:text {
      width = 25,
      text = num2note(gen_note_rand_root_vb.value)
    }  

  local b_motion_chb = vb:checkbox {
    value = Rand.b_motion,
    tooltip = 'Brownian Motion',
    notifier = function(value) 
      Rand.b_motion = value          
    end
  } 
    
  local gen_random_root_note_col = vb:column{
    margin = 4,
    spacing = 6,
    
    vb:row{  
      spacing = 1,
      gen_note_val_tx,
      
      vb:text {
        width = 10,
        text = 'BM'
      },
      b_motion_chb,
    },   
    
    gen_note_rand_root_vb, 
  }       
     
  local n_random_pu = vb:popup {
    value = Rand.note_typ,
    width = 55,
    tooltip = 'Source of note randomization: Random, fibonacci ratios, fibonacci numbers,  pink noise, probability distribution of loaded or captured notes',
    items = {'Rand','FibsR','FibsN','PN', 'Dis'},
    notifier = function(val)
      Rand.note_typ = val
    end  
  } 
  
  local l_random_pu = vb:popup {
    value = Rand.len_typ,
    width = 55,
    tooltip = 'Source of note length randomization: Random, fibonacci ratios, fibonacci numbers, probability distribution of loaded or captured notes',
    items = {'Rand','Fibs','FibsN','Dis'},
    notifier = function(val)
      Rand.len_typ = val
    end  
  }      

  local new_add_lock_chr = vb:chooser {
    value = Rand.new_add_lock,
    tooltip = 'Reset, add to or lock note distribution',
    items = {'New','Add','Lock'},
    notifier = function(val)
      Rand.new_add_lock = val
    end  
  } 
  
  local gen_note_col = vb:column{
    vb:text {
      width = 10,
      text = 'N'
    },       
    gen_random_note_knob, 
  }
  
  local gen_octave_col = vb:column{
    vb:text {
      width = 30,
      text = '  O'
    },       
    gen_random_octave_knob 
  }  
  

  local gen_len_col = vb:column{
    vb:row{
      spacing = 1,
      vb:text {
        width = 10,
        text = 'L'
      },       
   },      
    gen_random_len_knob 
  }     
 
  local gen_velocity_col = vb:column{
    vb:text {
      width = 30,
      text = '  V'
    },       
    gen_velocity_knob 
  } 

  local gen_chprob_col = vb:column{
    vb:text {
      width = 30,
      text = 'CHP'
    },       
    gen_chprob_knob 
  } 

  local gen_lib_chb = vb:checkbox {
    tooltip = 'Load randomly selected preset',
    value = Rand.lib_preset,
    notifier = function(value) 
      Rand.lib_preset = value
    end
  }

  local form_libCh_pu = vb:popup {
    width = 70,
    tooltip = 'R Form - random chord form, CH Pres - load randomly selected chord preset to be used during note randomization when chord probability knob > 0, with scale on, find scale and key and select it',
    value =  Rand.form_lib_chord,
    items = {'NA','R Form', 'CH Pres'},      
    notifier =  function (val)   
      Rand.form_lib_chord = val
    end
  }   

  local gen_scale_pu = vb:popup {
    width = 77,
    value = Rand.scale,
    tooltip = 'Randomize scales: NA - none, MM - minor and major only, ALL - randomize all scales',
    items = {'Scales NA','Scales MM','Scales All'},
    notifier = function(new_index)
      Rand.scale = new_index
    end
  } 
   

  local gen_key_chb = vb:checkbox {
    tooltip = 'Randomize key',
    value = Rand.key,
    notifier = function(value) 
      Rand.key = value
    end
  }
  
  local n_l_random_col = vb:column{
    spacing = 2,
    margin = 4,
    
    vb:row{
      vb:text {
        width = 15,
        text = 'N'
        },
        n_random_pu,
    },
    
    vb:row{
      vb:text {
        width = 15,
        text = 'L'
        },
        l_random_pu,
    },  
  }
   
  local gen_auto_col = vb:column{
    margin = 3,
    spacing = 5,
        
    vb:row{
      spacing = 0,

      vb:text {
        width = 15,
        text = 'Lib'
      },
      gen_lib_chb, 
      form_libCh_pu,     
    },
    
    vb:row{
      spacing = 0,
       
      gen_scale_pu,
      vb:text {
        width = 15,
        text = 'Key'
      },
      gen_key_chb,    
    },
  }
        
  return vb:row{
    spacing = 4,
    
    vb:row{
      margin = 4,
      new_add_lock_chr,
    },
    gen_random_root_note_col,    
    n_l_random_col,
    gen_note_col,
    gen_octave_col,
    gen_len_col,
    gen_velocity_col,
    gen_chprob_col,
    gen_auto_col, 
  }
end


local function mutateUI () -----------------------------------------------------------------------------------------------------

  local mutate_len_knob = vb:rotary {
    tooltip = 'Amount of maximum mutation of distance between notes',
    min = 1,
    max = 16,
    value = Rand.mutate_len,
    width = 25,
    height = 25,
    notifier = function (val)
      Rand.mutate_len = val
    end
    } 
    
  local mutate_note_knob = vb:rotary {
    tooltip = 'Amount of maximum mutation of notes',
    min = 1,
    max = 24,
    value = Rand.mutate_note,
    width = 25,
    height = 25,
    notifier = function (val)
      Rand.mutate_note = val
    end
  } 
 
  local mutate_octave_knob = vb:rotary {
    tooltip = 'Amount of maximum mutation of octaves',
    min = 1,
    max = 6,
    value = Rand.mutate_oct,
    width = 25,
    height = 25,
    notifier   = function(val)
      Rand.mutate_oct = val
    end
  } 
    
   local mutate_velocity_knob = vb:rotary {
     tooltip = 'Amount of maximum mutation of note velocities',
     min = 1,
     max = 126,
     value = Rand.mutate_vel,
     width = 25,
     height = 25,
     notifier  = function(val)
       Rand.mutate_vel = val
     end
   }    
  
   local mutate_len_fib_chb = vb:checkbox {
    tooltip = 'Random Fibonacci ratios',
    value = Rand.mutate_fib_len,
    notifier = function(value) 
      Rand.mutate_fib_len = value
    end
  }
 
   local mutate_note_fib_chb = vb:checkbox {
    tooltip = 'Random Fibonacci ratios',
    value = Rand.mutate_fib_note,
    notifier = function(value) 
      Rand.mutate_fib_note = value
    end
  } 
 
 local mutate_note_col = vb:column{
    vb:row {
      vb:text {
        width = 10,
        text = 'N'
      },
      mutate_note_fib_chb,       
    },
    mutate_note_knob, 
    
  }
  
  local mutate_octave_col = vb:column{
    vb:text {
      width = 30,
      text = '  O'
    },       
    mutate_octave_knob 
  }  
  
  local mutate_len_col = vb:column{
    vb:row{
      spacing = 1,
      vb:text {
        width = 10,
        text = 'L'
      },       
      mutate_len_fib_chb,
   },      
    mutate_len_knob 
  }     
  
  local mutate_velocity_col = vb:column{
    vb:text {
      width = 30,
      text = '  V'
    },       
    mutate_velocity_knob 
  } 
  
  return vb:row{
    spacing = 4,
    
    mutate_note_col,
    mutate_octave_col,
    mutate_len_col,
    mutate_velocity_col,
  }  
end


local function MarkovUI() ---------------------------------------------------------------------------------------------------------
  
  local PST_params, IP_params
  
  local markow_sw = vb:popup {
    width = 50,
    value = 2,
    tooltip = 'Markov models: MLP - 1 node Markov note/motif with P and R threshholds, PST - prediction suffix trees, IP - Incremental parsing',
    items = {'MLP','PST', 'IP'},
    notifier = function(val)
      if val == 1 then
        Markov.typ = 'MLP'
        PST_params.visible = true
        IP_params.visible = false
      elseif val == 2 then
        Markov.typ = 'PST'
        PST_params.visible = true
        IP_params.visible = false
      elseif val == 3 then
        Markov.typ = 'IP'
        PST_params.visible = false
        IP_params.visible = true
      end      
    end  
  }  
  
  local markov_pst_len_vb = vb:valuebox {
      tooltip = 'MLP & PST motif maximum length',
      min = 1,
      max = 64,
      width = 45,
      value = Markov.PST_l,
      notifier = function(value)
        Markov.PST_l = value
      end  
  }
     
  local markov_pst_pmin_vb = vb:valuebox {
      tooltip = 'MLP & PST probability threshold not to be exceeded',
      min = 1,
      max = 100,
      width = 60,
      value = Markov.PST_minp*100,
      notifier = function(value)
        Markov.PST_minp = value/100
      end  
  }
       
  local markov_pst_rmin_vb = vb:valuebox {
      tooltip = 'MLP & PST minimum times note / motif repeats in original phrase',
      min = 1,
      max = 128,
      width = 45,
      value = Markov.PST_r,
      notifier = function(value)
        Markov.PST_r = value
      end  
  }

  local new_add_lock_chr = vb:chooser {
    value = Scan.Markov_new_add_lock,
    tooltip = 'Reset, add to or lock Markov memory.',
    items = {'New','Add','Lock'},
    notifier = function(val)
      Scan.Markov_new_add_lock = val
    end  
  } 
     
   
  local markov_first_N_pu = vb:popup {
    width = 90,
    value = 1,
    tooltip = 'Start with random or specified note / chord. The specified note returns probable follow-up notes',
    items = {'Random'},
    notifier = function(val) 
      if val > 1 then
        Markov.first_n_ix = val-1
      else 
       Markov.first_n_ix = nil
      end
    end      
  }
  
  local markov_mul_chb = vb:checkbox {
    tooltip = 'MLP, PST multiple motif lengths, 3 = 1-3, 5 = 1-5 etc.',
    value = Markov.mul,
    notifier = function(value) 
      Markov.mul = value
    end
  } 

  local markov_continue_pu = vb:popup {
    tooltip = 'Grab specifed in parameter order, number of notes from top row of block selection, to start note mutation from.',
    value = Markov.continue,
    items = {'Con NA' ,'Con Norm', 'Con Exact'},
    notifier = function(val)
      Markov.continue = val 
    end
  } 
 
  local markov_order_vb = vb:valuebox {
      tooltip = 'Markov order, simple note memory, lower more random, higher more melodic but closer to original and more input notes required.',
      min = 1,
      max = 10,
      width = 45,
      value = Markov.order,
      notifier = function(val)
        Markov.order = val
      end  
  }
  
  local IP_iterations_vb = vb:valuebox {
      tooltip = 'IP number of iterations, 0 no limit, higher number is more harmonious but closer to original, lower more unique but less harmonious',
      min = 0,
      max = 99999,
      width = 80,
      value = Markov.IP_iter,
      notifier = function(value)
        Markov.IP_iter = value
      end  
  }   
  
  local IP_depth_vb = vb:valuebox {
      tooltip = 'Max IP tree depth, 0 no limit',
      min = 0,
      max = 999,
      width = 60,
      value = Markov.IP_max_depth,
      notifier = function(value)
        Markov.IP_max_depth = value
      end  
  }   
    
  IP_params = vb:row{
    vb:text {
      width = 30,
      text = ' Iterations'
    },
    IP_iterations_vb,
    
    vb:text {
      width = 20,
      text = ' Depth'
    },
    IP_depth_vb,
  }
  IP_params.visible = false
  
  PST_params = vb:row{
    vb:text {
      width = 20,
      text = ' Mul'
    },
    markov_mul_chb,
        
    vb:text {
      width = 40,
      text = ' PST  L'
    },  
    markov_pst_len_vb,
        
    vb:text {
      width = 20,
      text = ' P'
    },
    markov_pst_pmin_vb,
        
    vb:text {
      width = 20,
      text = ' R'
    },
    markov_pst_rmin_vb,
  }
  
  UI.markov_typ     = markow_sw     
  UI.markov_first_N = markov_first_N_pu
       
  return vb:row{
    margin = 4,
    spacing = 8,
    
    new_add_lock_chr,
    
    vb:column {
      spacing = 4,
      
      vb:row{
        spacing = 3,
        
        markow_sw,      
        
        vb:text {
          width = 40,
          text = ' Order'
        },  
        markov_order_vb,
                
        vb:text {
          width = 50,
          text = '   Start N'
        },
        markov_first_N_pu,      

        markov_continue_pu,
      },
       
      vb:row{
        spacing = 3,
                       
        PST_params,
        IP_params,             
      },
    },
  }  
end


local function chordUI() -----------------------------------------------------------------------------------

  local gen_chrd_scale_pu = vb:popup {
    width = 80,
    tooltip = 'Randomize scales for all chords / Randomize only minor and major chord scales',
    value =  Rand.chord_scale,
    items = {'Scales NA','Scales All', 'Scales MM'},      
    notifier =  function (val)   
      Rand.chord_scale = val
    end
  }   
     
  local gen_chrd_key_pu = vb:popup {
    width = 68,
    tooltip = 'Randomize keys for all chord / Randomize and apply one chord key for all chords',
    value =  Rand.chord_key,
    items = {'Keys NA', 'Keys All', 'Keys 1'},      
    notifier =  function (val)   
      Rand.chord_key = val
    end
  }       
  
  local gen_chrd_form_chb = vb:checkbox {
    tooltip = 'Randomize chord forms',
    value = Rand.chord_form,
    notifier = function(value) 
      Rand.chord_form = value
    end
  }
  
  local gen_arp_form_pu = vb:popup {
    width = 60,
    value = 1,
    tooltip = 'Arp form randomization: NA - none, Sel - selected form, All - all forms',
    items = {'Arp NA','Arp Sel','Arp All'},
    notifier = function(new_index)
      Rand.arp_pattern = new_index
    end
  }  
  
  return vb:column{
    margin = 4,
    spacing = 3,
    vb:row{
      vb:text {
        width = 20,
        text = 'Chord Form'
      }, 
      gen_chrd_form_chb,  
      gen_arp_form_pu,  
    },
    vb:row {
      gen_chrd_scale_pu,    
      gen_chrd_key_pu,
    },    
  }
end


local function intervalsUI() -----------------------------------------------------------------------------
   
   local interval_note_tx = 
      vb:text {
        width = 25,
        text = '' 
      }  
   
   local interval_start_note_tx = 
      vb:text {
        width = 25,
        text = num2note(WeightedIntervals.first_note) 
      }
   
   local note_vb = vb:valuebox {
      tooltip = 'Start note',
      min = 0,
      max = 109,
      width = 45,
      value = WeightedIntervals.first_note,
      notifier = function(val)
        WeightedIntervals.first_note = val
        interval_start_note_tx.text = num2note(val)
      end  
  } 
    
   local note_chb = vb:checkbox {
    tooltip = 'Start note enable',
    value = WeightedIntervals.specify_first_note,
    notifier = function(val) 
      WeightedIntervals.specify_first_note = val
    end
  }  
      
  local interval_sw = vb:switch {
    width = 80,
    value = WeightedIntervals.lock,
    tooltip = 'NEW - new note WeightedIntervals.fun probability when capturing and loading notes, LOCK - keep existing',
    items = {'NEW','LOCK'},
    notifier = function(val)
      WeightedIntervals.lock = val    
    end  
  }  
    
  local notes_pu = vb:popup {
    width = 50,
    tooltip = 'Root note',
    value =  1,
    items = {''},      
    notifier =  function (val)  
      WeightedIntervals.root_note_exists = true     
    end
  }  

   local hookelaw_vb = vb:valuebox {
      tooltip = 'Hooke law coefficient / 10',
      min = 1,
      max = 99,
      width = 45,
      value = WeightedIntervals.hooke_law_coef*10,
      notifier = function(val)
        WeightedIntervals.hooke_law_coef = val / 10  
      end  
  }
    
  
  UI.notes_pu = notes_pu    
  UI.interval_note_tx = interval_note_tx
  UI.int_start_N = note_vb
  
  UI.int_root_w = vb:row{}   
  UI.int_pos_w = vb:row{} 
  UI.int_neg_w = vb:row{} 
  
   
  return vb:column{
    vb:row{
      margin = 4,
      spacing = 4,
    
      interval_sw,
      
      vb:text {
        width = 35,
        text = ' Root'
      },
      interval_note_tx,
      notes_pu,
      
      vb:text {
        width = 30,
        text = ' First'
      },
      interval_start_note_tx,
      note_vb,
      note_chb,
      
      vb:text {
        width = 25,
        text = ' HL'
      },
      hookelaw_vb,        
    },  
    vb:row{  
     -- margin = 4,
      spacing = 1,
      
      vb:text {
        width = 20,
        text = 'R'
      },
      UI.int_root_w, --root_weights,
      
      vb:text {
        width = 20,
        text = '+'
      },
      UI.int_pos_w, -- pos_weights,
    },
    vb:row{  
      vb:text {
        width = 20,
        text = '-'
      },
      UI.int_neg_w, -- neg_weights,
    }    
  }
end


local function ch2pUI()

  local repeat_vb = vb:valuebox {
    tooltip = 'When > 0, reapply during playback specified number of pattern rows',
    min = 0,
    max = 64,
    width = 45,
    value = CH2P.reapply,
    notifier = function(val)
      CH2P.reapply = val
    end  
  }

  local duration_vb = vb:valuebox {
    tooltip = 'Chord duration (number of rows)',
    min = 0,
    max = 64,
    width = 45,
    value = CH2P.chord_dur,
    notifier = function(val)
      CH2P.chord_dur = val
    end  
  }

  local octave_vb = vb:valuebox {
    tooltip = 'Octave',
    min = 0,
    max = 9,
    width = 45,
    value = CH2P.root_oct,
    notifier = function(val)
      CH2P.root_oct = val
    end  
  }
    
  local type_sw = vb:switch {
    width = 120,
    value = CH2P.type,
    tooltip = 'Auto - automatic harmonization, Int - place chords at specifed interval, FX - place chords with 0h00 in effect column',
    items = {'AUTO','INT', 'FX'},
    notifier = function(val)
      CH2P.type = val    
    end  
  } 
  
  local cs_slider = vb:minislider {
      min = 1.5,
      max = 3,
      width = 50,
      height = 15,
      value = Stats.consonance_score,
      tooltip = 'Consonance score for chord harmonization. Lower number more consonance and fewer chord choices, higher more dissonance but greater number of chords allowed  ',
      notifier = function(val)
        Stats.consonance_score = val
        sprint('' .. val)
      end
  }
  
  local chp_min_vb = vb:valuebox {
    tooltip = 'Minimum number of chord notes',
    min = 1,
    max = 6,
    width = 45,
    value = CH2P.chord_min,
    notifier = function(val)
      CH2P.chord_min = val
    end  
  }

  local chp_max_vb = vb:valuebox {
    tooltip = 'Maximum number of chord notes',
    min = 1,
    max = 6,
    width = 45,
    value = CH2P.chord_max,
    notifier = function(val)
      CH2P.chord_max = val
    end  
  }

  local chp_min_notes_vb = vb:valuebox {
    tooltip = 'Minimum number of notes to feed the harmonizer. Affects chord spacing',
    min = 2,
    max = 16,
    width = 45,
    value = CH2P.notes_min,
    notifier = function(val)
      CH2P.notes_min = val
    end  
  }
  
  local best_count_vb = vb:valuebox {
    tooltip = 'Lack of consonance improvement with each added note to the harmonizer is counted. If the count is greater than the specified number,' .. 
               ' accept best result and seek next chord. Lesser number will consider / capture less notes.',
    min = 1,
    max = 10,
    width = 45,
    value = CH2P.best_count,
    notifier = function(val)
      CH2P.best_count = val
    end  
  }
           
  local allow_dups_chb = vb:checkbox {
    tooltip = 'Allow duplictes; chord can be followed by the same chord',
    value = CH2P.allow_dups,
    notifier = function(val) 
      CH2P.allow_dups = val
    end
  }  
            
  return vb:column{
     vb:row{
       margin = 3,
       spacing = 4,
     
       vb:text {
         width = 35,
         text = 'Reapply'
       },
       repeat_vb,    
       
       vb:text {
         width = 40,
         align = 'right',
         text = ' Interval'
       },
       duration_vb,
       
        vb:text {
         width = 25,
         align = 'right',
         text = ' Oct'
       },
       octave_vb,
       
       vb:text {
         width = 63,
         align = 'right',
         text = ' Min Chrd N'
       },
       chp_min_notes_vb,
       
       vb:text {
         align = 'right',
         width = 23,
         text = ' Dups'
       },
       allow_dups_chb,
     },
     
     vb:row{
       type_sw,
     
       vb:text {
         width = 28,
         align = 'right',
         text = ' CS'
       },
       cs_slider,
       
       vb:text {
         width = 42,
         align = 'right',
         text = ' Min L'
       },
       chp_min_vb,
       
       vb:text {
         width = 45,
         align = 'right',
         text = ' Max L'
       },
       chp_max_vb,
       
       vb:text {
         width = 28,
         align = 'right',
         text = ' BC'
       },
       best_count_vb,
     },
   }
end


local function chrdArpUI()  

    local gen_chrd_scale_pu = vb:popup {
    width = 40,
    tooltip = 'Randomize scales for all chords / Randomize only minor and major chord scales',
    value =  Rand.chord_scale,
    items = {'NA','All', 'MM'},      
    notifier =  function (val)   
      Rand.chord_scale = val
    end
  }   
     
  local gen_chrd_key_pu = vb:popup {
    width = 40,
    tooltip = 'Randomize keys for all chord / Randomize and apply one chord key for all chords',
    value =  Rand.chord_key,
    items = {'NA', 'All', '1'},      
    notifier =  function (val)   
      Rand.chord_key = val
    end
  }  

  local note_chb = vb:checkbox {
    tooltip = 'Enable chord note randomization',
    value = Rand.chord_note,
    notifier = function(val) 
      Rand.chord_note = val
    end
  }
 
  local form_chb = vb:checkbox {
    tooltip = 'Enable chord form randomization',
    value = Rand.chord_form,
    notifier = function(val) 
      Rand.chord_form = val
    end
  }  
    
  local octave2_chb = vb:checkbox {
    tooltip = 'Enable arp octave randomization',
    value = Rand.arp_oct,
    notifier = function(val) 
      Rand.arp_oct = val
    end
  }      

  local note2_chb = vb:checkbox {
    tooltip = 'Enable randomizing arp strting note / chord interval with',
    value = Rand.arp_interval,
    notifier = function(val) 
      Rand.arp_interval = val
    end
  }  
  
  local arp_pattern_chb = vb:checkbox {
    tooltip = 'Enable arp pattern randomization',
    value = Rand.arp_interval,
    notifier = function(val) 
      Rand.arp_pattern_no = val
    end
  }  
  
   local gen_arp_form_pu = vb:popup {
    width = 40,
    value = 1,
    tooltip = 'Arp form randomization: NA - none, Sel - selected form, All - all forms',
    items = {'NA','Sel','All'},
    notifier = function(new_index)
      Rand.arp_pattern = new_index
    end
  } 
    
  return vb:column{
    vb:row{
      margin = 3,
      spacing = 1,
     
      vb:text {
        width = 30,
        text = 'Scale'
      }, 
      gen_chrd_scale_pu,
      
      vb:text {
        width = 30,
        text = '  key'
      }, 
      gen_chrd_key_pu,
     
      vb:text {
        width = 35,
        text = '  Chord Note'
      },  
      note_chb,
     
      vb:text {
        width = 30,
        text = '  Form'
      },     
      form_chb,       
    },
   
    vb:row{
      margin = 3,
      spacing = 1,     

      vb:text {
        width = 30,
        text = 'Arp Octave'
      },     
      octave2_chb,
            
    --  form2_chb,
     
      vb:text {
        width = 30,
        text = '  Interval'
      }, 
      note2_chb,     
      
      vb:text {
        width = 30,
        text = '  Pattern',
      },
      gen_arp_form_pu,      
      
      vb:text {
        width = 30,
        text = '  Pattern no'
      },
      arp_pattern_chb,
      
    },     
  }  
end

local function LibMasksUI()  

  return vb:column{
    vb:row{
      margin = 3,
      spacing = 1,
    },
    vb:row{
      margin = 3,
      spacing = 1,
    },
  }
end

local function scanUI()  
  
  local recomb_len_vb = vb:valuebox {
    tooltip = 'Minimum number of notes, not taking into account anchor notes',
    min = 0,
    max = 16,
    width = 45,
    value = Scan.recomb_len,
    notifier = function(val)
      Scan.recomb_len = val
    end  
  }
  
  local recomb_anchor_vb = vb:valuebox {
    tooltip = 'Minimum number of notes, not taking into account anchor notes',
    min = 1,
    max = 16,
    width = 45,
    value = Scan.recomb_anchor,
    notifier = function(val)
      Scan.recomb_anchor = val
    end  
  }
  
  local recomb_loct_vb = vb:valuebox {
    tooltip = 'Lowest octave, notes above will be transposed down',
    min = 0,
    max = 9,
    width = 100,
    value = Scan.recomb_loct,
    notifier = function(val)
      Scan.recomb_loct = val
    end  
  }
  
  local recomb_hoct_vb = vb:valuebox {
    tooltip = 'Highest octave, notes above will be transposed down',
    min = 0,
    max = 9,
    width = 45,
    value = Scan.recomb_hoct,
    notifier = function(val)
      Scan.recomb_hoct = val
    end  
  }
  
  local recomb_loct_vb = vb:valuebox {
    tooltip = 'Lowest octave, notes below will be transposed up',
    min = 1,
    max = 8,
    width = 45,
    value = Scan.recomb_loct,
    notifier = function(val)
      Scan.recomb_loct = val
    end  
  }
  
  local recomb_anchor_min_vb = vb:valuebox {
    tooltip = 'Minimum anchor notes required',
    min = 0,
    max = 16,
    width = 45,
    value = Scan.recomb_anchor_min,
    notifier = function(val)
      Scan.recomb_anchor_min = val
    end  
  }
  
  local recombine_bt = vb:button {
    tooltip = 'Recombine motifs',
    text = ' Recombine ',
    notifier =  function () 
      recombFun()
    end 
  }  
  
  local continue_bt = vb:button {
    tooltip = 'Resume / continue recombining block selection',
    text = ' Resume ',
    notifier =  function () 
      recombContinue()
    end 
  }  

  local scan_bt = vb:button {
    tooltip = " Scale&Key, correlation, motif scan ",
    text = ' Scan ',
    notifier =  function ()   
      local process = ProcessSlicer(scaleKeyScan, function() end)
      process:start()
    --  scaleKeyScan() 
    end
  }  
  
  local scan_random_chb = vb:checkbox {
    tooltip = 'Random scale and key',
    value = Scan.recomb_random_sk,
    notifier = function(val) 
      Scan.recomb_random_sk = val
    end
  }
  
  local save_scans_bt = vb:button {
    text = ' Save ',
    notifier =  function ()    
       
      local fn = MW.path .. 'scan_sk.lua'      
      if fileExists(fn) then
        os.remove (fn)   
      end
        
      local file = io.open(fn, "wb")   
      sprint('Dump and save scan_sk.lua')
      file:write('Scan.db_scale_key = ' .. serialize(Scan.db_scale_key))
      file:close()      
       
      ----------------------------
      
      fn = MW.path .. 'scan_recomb.lua'        
      if fileExists(fn) then
        os.remove (fn)   
      end
      
      file = io.open(fn, "wb") 
      sprint('Dump and save scan_recomb.lua')
      file:write('Scan.db_chords_rc = ' .. serialize(Scan.db_chords_rc))
      file:close()       
    end 
  }
  
  local recomb_tries_vb = vb:valuebox {
    tooltip = 'If fails, retry this number of times',
    min = 1,
    max = 99999,
    width = 60,
    value = Scan.recomb_tries,
    notifier = function(val)
      Scan.recomb_tries = val
    end  
  }
  
  local scan_metric_typ_pu = vb:popup {
    width = 70,
    tooltip = 'Metric: cross-entropy, Zipfer\'s slope+R2, Chord Count',
    value =  Stats.metric_typ,
    items = {'Entropy','Zipfer','Count'},      
    notifier =  function (val)   
      Stats.metric_typ = val
    end
  }   
  
  local cs_slider = vb:minislider {
      min = 1.5,
      max = 6,
      width = 50,
      height = 15,
      value = Scan.recomb_cs ,
      tooltip = 'Consonance score for anchors first and last notes. Lower number more consonance and fewer choices, higher more dissonance and more choices ',
      notifier = function(val)
        Scan.recomb_cs  = val
        sprint('' .. val)
      end
  }
  
  return vb:column{
    vb:row{
      margin = 3,
      spacing = 1,
      
      vb:text {
        width = 30,
        text = '  Motif Min L'
      }, 
      recomb_len_vb,
      
      vb:text {
        width = 30,
        text = '  Anchor L'
      }, 
      recomb_anchor_vb,
      
      vb:text {
        width = 30,
        text = ' min'
      }, 
      recomb_anchor_min_vb,
      
      vb:text {
        width = 45,
        align = 'right',
        text = '  Oct< '
      }, 
      recomb_loct_vb,
      
      vb:text {
        width = 30,
        text = ' Oct>'
      }, 
      recomb_hoct_vb,
      
    },
    vb:row{
      margin = 3,
      spacing = 1,      
      
      scan_bt,
      scan_metric_typ_pu,
      save_scans_bt,
      
      vb:text {
        width = 8,
        text = '  '
      }, 
      
      recombine_bt,

      vb:text {
        width = 45,
        text = 'Random SK'
      }, 
      scan_random_chb,
      
      vb:text {
        align = 'right',
        width = 15,
        text = '#'
      }, 
      recomb_tries_vb,
      
      continue_bt,
    },
  }
end


function libUI()
  local search_txf
  
  local function loadAll() 
    local len = #UI.lib_pu.items
    
    if len > 0 then
      UI.suspend_update = true
      local tmp_ms, tmp_rs = Scan.Markov_new_add_lock, Rand.new_add_lock              
            
      if len > 1 then                    
        for i=1,len do
          sprint('Files left ' .. len - i)
          if i == 1 then 
            Scan.Markov_new_add_lock, Rand.new_add_lock = 1,1
          else
            Scan.Markov_new_add_lock, Rand.new_add_lock = 2,2   
          end             
             
          local  filename = MW.path .. UI.lib_pu.items[i] .. Lib.file_ext
            
        --  if MW.phr.editor_poly and MW.phr.editor_chords == false then
        --    readSeq(filename)
        --  elseif MW.phr.editor_chords then
        --    readSeqChord(filename) 
        --  else
        --    Load.notes((filename) 
       --   end
          
          Load.notes(filename) 
          phraseScanner()  
          
          coroutine.yield()  
        end
      end
        
      Scan.Markov_new_add_lock, Rand.new_add_lock = tmp_ms, tmp_rs        
      UI.suspend_update = false
      
      sprint('Done')
    end
  end  

  local load_all_bt = vb:button {
    tooltip = "Load all phrases to note generator / mutator",
    text = "Load All",
    width = 50,
    visible = false,
    notifier =  function ()   
      Load.pattern_len_adj = false
      local process = ProcessSlicer(loadAll, function() end)
      process:start()
      Load.pattern_len_adj = true     
    end 
  }  

  local funTag = function()
    local items = {}
    local files = os.filenames(MW.path, {'*.tag'})
       
    for _,file in ipairs(files) do 
      local tag = Load.tagsTb(file) 
          
      local s_tag = scanSearch(search_txf.text)   
      if search_txf.text ~= '' then     
        if compareTbs(tag.tb, s_tag.include) >= #s_tag.include and compareTbs(tag.tb, s_tag.exclude) == 0 and s_tag.opFun(tag.score-1) then
          table.insert(items, remExt2End(remExt2End(file)))  
        end
      end
    end       
    
    if #items > 0 then
      load_all_bt.visible = true
    else
      load_all_bt.visible = false   
    end  
        
    UI.lib_pu.items = items
  end
  
  local scan_last_val = 0   
  
  local scan_sw
  scan_sw = vb:switch {
    width = 400,
    value = 1,
    tooltip = "Limit phrase selection based on: CO - correlation, SK - same scale and key, EN - cross entropy, NA - all files",
    items = {'All', 'Tags', 'Correlation', 'Entropy', 'Scale & Key'},
    notifier = function(val)          
      
      Scan.typ = val
      
      if val == 1 then
        UI.lib_pu.items = remExt(os.filenames(MW.path, {"*" .. Lib.file_ext}))  
        load_all_bt.visible = true
      elseif val == 2 then
        funTag()       
      elseif val > 2 and val < 5 then    
        if UI.scales_pu.value == 1 then 
          sprint('Select scale.')  
          return
        end      
        local type = {}
        type[3] = 'correlation'
        type[4] = 'entropy'
        
        local fns = libScan(type[val])
        
        if fns and #fns>0 then
          UI.lib_pu.items = fns
          
          load_all_bt.visible = true
          Scan.correlation_entropy_on = true           
        else
          load_all_bt.visible = false 
          sprint('Not found. Make sure phrases were scanned (Setup, Scan)')
        end        
      elseif scan_last_val == 3 or scan_last_val == 4 then
      
        UI.lib_pu.items = remExt(os.filenames(MW.path, {"*" .. Lib.file_ext})) 
        Scan.correlation_entropy_on = false    
      end      
      
      if val == 5 then       
        local scale = UI.scales_pu.value 
        local key = UI.keys_pu.value  
        
        local ix 
        if MW.phr.editor_poly == false and MW.phr.editor_chords == false then
          ix = 'mono'
        elseif MW.phr.editor_poly then
          ix = 'poly'  
        elseif MW.phr.editor_chords then
          ix='chrd'
        end 
        
      ---------------- scale key  
      if val == 5 and UI.scales_pu.value == 1 then 
        sprint('Select scale.')            
                       
      elseif val == 5 and scale > 1 and Scan.db_scale_key[ix][scale] then 
          if Scan.db_scale_key[ix][scale][key] then
            UI.lib_pu.items = remExt(Scan.db_scale_key[ix][scale][key].file_name)

            Scan.scale_key_on = true 
            load_all_bt.visible = true
            --UI.load_scale_key.value = false
            return
          end
        end  
        
        sprint('Not found. Make sure phrases were scanned (Setup, Scan)')
        
        scan_sw.value = 1
        Scan.scale_key_on = false
        load_all_bt.visible = false    
        --UI.load_scale_key.value = false         
      elseif val == 5 or scan_last_val == 5 then 
        UI.lib_pu.items = remExt(os.filenames(MW.path, {"*" .. Lib.file_ext})) 
        Scan.scale_key_on = false
        load_all_bt.visible = false    
        --UI.load_scale_key.value = true 
      end   
      scan_last_val = val
    end  
  }
  
  local tags_txf =  vb:textfield {
    width = 100,
    align = "left",
    tooltip = 'Comma separated tags',
    text = Tag.text,
    notifier = function(text)       
      Tag.text = text
      scanTags()
    end
  }
  
  local score_sw = vb:switch {
    width = 100,
    value = Tag.score,    
    items = {'NA','1', '2', '3'},
    notifier = function(val)    
      Tag.score = val 
    end  
  }
  
  search_txf = vb:textfield {
    width = 100,
    align = "left",
    tooltip = 'Comma separated tags, score to search for. - excludes, For ex: bass, -fast. Equal, less and greater signs are supprted, for ex: >1, >=2, <3',
    text = '',
    notifier = function(text) 
      if scan_sw.value == 2 then
        funTag()
      end
    end
  }
  
  local ec_txf = vb:textfield {
    tooltip = 'Minimum correlation, maximum entropy',
    width = 25,
    align = "center",
    text = '0.5',
  }
  
  UI.tegs    = tags_txf
  UI.score   = score_sw
  UI.scan_sw = scan_sw
  UI.scan_ec = ec_txf
  
  return vb:column{    
    vb:row{
      margin = 3,
      spacing = 1,      
      
      scan_sw,
      load_all_bt,
    },
    
    vb:row{
      margin = 3,
      spacing = 1,   
      
      vb:text {
        text = 'EC',
        width = 20,
      },
      ec_txf,
      
      vb:text {
        text = 'Tags',
        width = 20,
      },
      tags_txf,
      vb:text {
        text = 'Score',
        width = 20,
      },
      score_sw,
      
      vb:text {
        text = 'Search',
        align = 'right',
        width = 40,
      }, 
      search_txf
    }
  }  
end

function stats_UI()

  local entropy_tx = vb:text {
    text = 'Entropy',
    width = 50,
  }
 
  local zipf_slope_tx = vb:text {
    text = 'Zipf\'s Slope',
    width = 40,
  }
 
  local zipf_r2_tx = vb:text {
    text = 'Zipf\'s R2',
    width = 40,
  }
  
  UI.stats_entropy     = entropy_tx
  UI.stats_zipfs_r2    = zipf_r2_tx 
  UI.stats_zipfs_slope = zipf_slope_tx
 
  statsFun()
 
  return vb:column{
    vb:row{
      margin = 3,
      spacing = 1,      

      vb:row{     
        spacing = 15, 
         entropy_tx,
        zipf_slope_tx,
        zipf_r2_tx,
      }
    } 
  }
end

function statsFun()
  
  if not MW.started then return end

  UI.stats_entropy.text = string.format("Entropy  %.2f", entropy(normalize(Stats.pitch_duration_dst)))  
  
  local pdd = {} 
  filter(Stats.pitch_duration_dst, function(v) if v > 0 then table.insert(pdd,v) end end)
  table.sort(pdd, function(a,b) return a>b end)
  
  local slope, r2 = getSlopeR2(enumFromTo(#pdd,1), pdd) 
  UI.stats_zipfs_r2.text    = string.format("Zipf\'s R2  %.2f", r2)
  UI.stats_zipfs_slope.text = string.format("Zipf\'s Slope  %.2f", slope)
end


function motifUI()

  local scan_bt = vb:button {
    text = " Scan ",
    notifier =  function ()
      scanMotifs()
    end 
  }
  
  local save_bt = vb:button {
    text = " Save ",
    notifier =  function ()
      local fn = MW.path .. 'scan_motifs.lua'      
      if fileExists(fn) then
        os.remove (fn)   
      end
        
      local file = io.open(fn, "wb")   
      sprint('Dump and save scan_sk.lua')
      file:write('Scan.db_motifs = ' .. serialize(Scan.db_motifs))
      file:close()  
    end 
  }
  
  local recomb_bt = vb:button {
    text = " Recombine ",
    notifier =  function ()
      recombMotif()
    end 
  }
 
  local resume_bt = vb:button {
    text = " Resume  ",
    notifier =  function ()
      motifContinue()
    end 
  }

  local scan_random_chb = vb:checkbox {
    tooltip = 'Random scale and key',
    value = Scan.motif_recomb_random_sk,
    notifier = function(val) 
      Scan.motif_recomb_random_sk = val
    end
  }

  local recomb_tries_vb = vb:valuebox {
    tooltip = 'If fails, retry this number of times',
    min = 1,
    max = 99999,
    width = 60,
    value = Scan.motif_recomb_tries,
    notifier = function(val)
      Scan.motif_recomb_tries = val
    end  
  }      
         
  local scan_add_chb = vb:checkbox {
    tooltip = 'Add to already scanned motifs',
    value = false,
  }
         
  UI.scan_motif_add = scan_add_chb
         
  return vb:column{
    vb:row{
      margin = 3,
      spacing = 1,   
            
      scan_bt,
      
      vb:text{
        width = 25,
        text = 'Add'
      },      
      scan_add_chb,
      
      save_bt,      
      recomb_bt,
      
      vb:text {
        width = 45,
        text = 'Random SK'
      }, 
      scan_random_chb,
      
      
     vb:text {
        align = 'right',
        width = 15,
        text = '#'
      }, 
      recomb_tries_vb,
      
      resume_bt,
    }, 
    
    vb:row{
      margin = 3,
      spacing = 1,      
      
    }
  }
end


function setupUI()

  local editor_btsz_vb = vb:valuebox {
    min = 15,
    max = 25,
    value = UI.bt_size,
    notifier = function(val)
      UI.bt_size = val
      updateSlots()
    end  
  }
  
  local row_spacing_vb = vb:valuebox {
    min = 0,
    max = 5,
    value = UI.slot_spacing,
    notifier = function(val)
      UI.slot_spacing = val
      updateSlots()
    end  
  }
  
  local bt_spacing_vb = vb:valuebox {
    min = 0,
    max = 6,
    value = UI.bt_spacing,
    notifier = function(val)
      UI.bt_spacing = val
      updateSlots()
    end  
  }
  
  local algnment_sw = vb:switch {
    width = 125,
    value = UI.alignment,
    items = {'Vertical','Horizontal'},
    notifier = function(val)    
      UI.alignment = val
      
      if MW.phr.editor_pattern_N then  
        MW.timer_pos2 = {}
      end
      updateSlots()
    end  
  }
  
  local detachChb = vb:checkbox {
    value = false,
    notifier = function(value)                
      if value then        
        if (UI.dialog2 and UI.dialog2.visible) then
          UI.dialog2:show()  
          return 
        end
 
        UI.dialog_content:remove_child(UI.editor_main)
        UI.dialog_content2:add_child(UI.editor_main) 
        
        UI.dialog2 = renoise.app():show_custom_dialog("Notes Editor", UI.dialog_content2, function(dialog,key) return key end)
      else
        UI.dialog_content2:remove_child(UI.editor_main)
        UI.dialog_content:add_child(UI.editor_main) 
        if (UI.dialog2 and UI.dialog2.visible) then UI.dialog2:close() end
      end
    end
  } 
  
  UI.DT = detachChb
     
  return vb:column{
    vb:row{
      margin = 3,
      spacing = 1,   

      vb:text {
        width = 105,
        text = 'Editor buttons size ',
        align = 'right',
      }, 
      editor_btsz_vb,
      
      vb:text {
        width = 95,
        text = 'Button spacing ',
        align = 'right',
      }, 
      bt_spacing_vb,
      
      vb:text {
        width = 120,
        text = 'Column / row spacing ',
        align = 'right',
      }, 
      row_spacing_vb,
      
    }, 
    
    vb:row{
      margin = 3,
      spacing = 1,      
      
      algnment_sw,
      
      vb:text {
        width = 50,
        text = 'Detach ',
        align = 'right',
      }, 
      detachChb,
    }
  }
end


function setupIO()
  
  local load_max_len_vb = vb:valuebox {
    min = 0,
    max = 64,
    value = Load.max_note_len,
    notifier = function(val)
      Load.max_note_len = val
    end  
  }
  
  local load_max_pat_len_vb = vb:valuebox {
    min = 32,
    max = 512,
    value = Load.max_pat_len,
    notifier = function(val)
      Load.max_pat_len = val
    end  
  }
    
  local markov_notes_len_txf =  vb:textfield {
    width = 100,
    align = "left",
    tooltip = 'Coma separated note lengths to be used with Markov chains (to reduce number of combinations). If the field is empty, all lengths are used.',
    text = Stats.note_strings_length,
    notifier = function(text)       
      Stats.note_strings_length = text
    end
  }
  
  local markov_clamp_root_vb = vb:valuebox {
    min = 0,
    max = 64,
    width = 45,
    value = Scan.Markov_oct_root,
    tooltip = "Root osctave",
    notifier = function(val)     
      Scan.Markov_oct_root = val
    end
  }  
    
  local markov_clamp_oct_vb = vb:valuebox {
    min = 0,
    max = 64,
    width = 45,
    value = Scan.Markov_oct_range,
    tooltip = "Clamp above and belove specified octave, 0 dont't clamp",
    notifier = function(val)     
      Scan.Markov_oct_range = val
    end
  } 
  
  local tx = vb:text {
    text = "Folder:  " .. MW.path
  }   
  
  local but_new_folder = vb:button {
    text = " New ",
    notifier =  function ()
      Lib.temp_path = renoise.app():prompt_for_path("Select Library Folder")
      tx.text =  "Folder:  " .. Lib.temp_path
    end 
  } 
  
  local len_one_chb = vb:checkbox {
    tooltip = 'Start grabbed notes from first row',
    value = false,
  } 
  
  local pat_resize_chb = vb:checkbox {
    tooltip = 'Resize pattern',
    value = false,
  } 
  
  UI.len1    = len_one_chb  
  UI.pat_rsz = pat_resize_chb 
    
  return vb:column{
    vb:row{
      margin = 3,
      spacing = 1,   
      
      tx,
      but_new_folder,  
      
      vb:text {
        width = 85,
        text = 'Max Note C&L',
        align = 'right',
      }, 
      load_max_len_vb,
      
      vb:text {
        width = 57,
        text = 'First Row',
        align = 'right',
      },
      len_one_chb,
      
      vb:text {
        width = 62,
        text = 'Pattern L ',
        align = 'right',
      },
      load_max_pat_len_vb,
      
    }, 
    
    vb:row{
      margin = 3,
      spacing = 1,  
      
      vb:text {
        width = 70,
        text = 'Pattern Resize'
      },
      pat_resize_chb,
      
      vb:text {
        width = 65,
        text = 'Markov NL',
        align = 'right',
      },       
      markov_notes_len_txf,  
      
      vb:text {
        width = 60,
        text = 'Root Oct ',
        align = 'right',
      },
      markov_clamp_root_vb,
    
      vb:text {
        width = 45,
        text = 'Range ',
        align = 'right',
      },
      markov_clamp_oct_vb,        
    },
  }
end


function utilitiesUI()
  
  local combine_but = vb:button {
    tooltip = 'Combine notes from first column of unmuted tracks to selected track',
    text = ' Comb ',
    notifier =  function ()
      combine(MW.phr.editor_poly)
    end 
  }

  local combineR_but = vb:button {
    tooltip = 'Combine notes from first column of unmuted tracks to selected track',
    text = ' CombR ',
    notifier =  function ()
      combine(MW.phr.editor_poly, "rep")
    end 
  }
  
  local combine_poly_but = vb:button {
    text = ' CombPoly ',
    tooltip = 'Combine notes from unmuted tracks to selscted track',
    notifier =  function ()
      combinePoly()
    end 
  }
  
  local remap_instruments_but = vb:button {
    tooltip = 'Remap all instruments in a pattern or sequence to selected instrument',
    text = ' Remap Instruments',
    notifier =  function ()
      remapInst()
    end 
  }
  
  local newLen_vb = vb:valuebox {
    min = 1,
    max = 128,
    width = 50,
    value = 64,
  }

  local newLen_but = vb:button {
    text = " Block Resize ",
    notifier =  function ()
      resizeBlock (newLen_vb.value)
    end 
  }   
    
  return vb:column{
    vb:row{
      margin = 3,
      spacing = 1,         

      combine_but,
      combineR_but,
      combine_poly_but,
      
      remap_instruments_but,     
    }, 
    
    vb:row{
      margin = 3,
      spacing = 1,   

      newLen_but,
      newLen_vb,
    }
  }
end

function skcsUI()

  local profiles_pu = vb:popup {
    width = 120,
    value = Scales.profiles.no,
    items = Scales.profiles.pu,
    tooltip = "Scale and key Scales.profiles",
    notifier = function(new_index)
      Scales.profiles.no = new_index
    end
  }
  
  local cs_slider = vb:minislider {
      min = 1.5,
      max = 4,
      width = 75,
      height = 15,
      value = Stats.forms_cs,
      tooltip = 'Consonance score used for chord forms. Lower number more consonance and fewer chord forms choices, higher more dissonance but greater number of forms allowed  ',
      notifier = function(val)
        Stats.forms_cs = val
        sprint('' .. val)
      end
  }
  
  local scan_bt = vb:button {
    text = " Scan ",
    tooltip = 'Filter out chord forms according to consonance score',
    notifier =  function ()
      buildFormCS()
    end 
  }
  
  return vb:column{
    vb:row{
      margin = 3,
      spacing = 1,         
      
      vb:text {
        width = 100,
        text = 'Scale and key profile '
      }, 
      profiles_pu,
      
      vb:text {
        width = 70,
        text = 'Forms CS ',
        align = 'right',        
      }, 
      cs_slider,
      scan_bt,
    }, 
    
    vb:row{
      margin = 3,
      spacing = 1,   
    }
  }
end

function MarkovUI2()
  
  local Max_len_vb = vb:valuebox {
    min = 0,
    max = 16,
    width = 50,
    value = 0,
  }
  
  local reset_bt = vb:button {
    text = " Reset ",
    tooltip = 'Reset',
    notifier =  function ()
      Stats.unique_notes       = {}
      Stats.pitch_duration_dst = {0,0,0,0,0,0,0,0,0,0,0,0}
      
      Stats.note_strings = {}
      Scan.Markov_process = nil
    end 
  }
  
  local scan_bt = vb:button {
    text = " Scan ",
    tooltip = 'Scan',
    notifier =  function ()
      phraseScanner()   
    end 
  }
  
  UI.markov_L = Max_len_vb
  
  return vb:column{
    vb:row{
      margin = 3,
      spacing = 1,         
      
      vb:text {
        width = 100,
        text = 'Max poly length'
      }, 
      Max_len_vb,
      
      reset_bt,
      scan_bt,
    }, 
    
    vb:row{
      margin = 3,
      spacing = 1,   
    }
  }
end

function ui_randomize() -----------------------------------------------------------------------

  local markovFun = function()
    if #Stats.note_strings == 0 then
      phraseScanner()
    end
     
    if Markov.continue == 1 then
      return MarkovProcess()   
    elseif Markov.continue == 2 then   
      return continueMarkov(blockNotesNrow(Markov.order)) 
    elseif Markov.continue == 3 then     
      return grabBlockNLcontinueExact() 
    end
  end  

  
  UI.menu = {}
  table.insert(UI.menu, {name = 'Library',     fun = function()end,         ui = libUI(),       apply = false, bot_row = false, chb = false})  
  table.insert(UI.menu, {name = 'Randomize',   fun = Randomize,             ui = randomizeUI(), apply = true,  bot_row = true,  chb = true})
  table.insert(UI.menu, {name = 'R Mutation',  fun = mutate,                ui = mutateUI(),    apply = true,  bot_row = true,  chb = true})
  table.insert(UI.menu, {name = 'P Muation',   fun = markovFun,             ui = MarkovUI(),    apply = true,  bot_row = true,  chb = true})
  table.insert(UI.menu, {name = 'P Muation2',  fun = function()end,         ui = MarkovUI2(),   apply = false, bot_row = false, chb = false})
  table.insert(UI.menu, {name = 'P Intervals', fun = WeightedIntervals.fun, ui = intervalsUI(), apply = true,  bot_row = true,  chb = true})
 -- table.insert(UI.menu, {name = 'Chords R',    fun = randomizeChords,       ui = chordUI(),     apply = true,  bot_row = true,  chb= true})
  
  table.insert(UI.menu, {name = 'Chord Arp R', fun = randomizeChords,       ui = chrdArpUI(),   apply = true,  bot_row = true,  chb = true})
  table.insert(UI.menu, {name = 'CH > P',      fun = harmonizePhraseApply,  ui = ch2pUI(),      apply = true,  bot_row = true,  chb = true})
  table.insert(UI.menu, {name = 'Scan Recomb', fun = recombFun,             ui = scanUI(),      apply = true,  bot_row = true,  chb = true})
  table.insert(UI.menu, {name = 'Motifs',      fun = recombMotif,           ui = motifUI(),     apply = true,  bot_row = true,  chb = true})
  table.insert(UI.menu, {name = 'SK CS',       fun = function()end,         ui = skcsUI(),      apply = false, bot_row = false, chb = false})
  table.insert(UI.menu, {name = 'Setup UI',    fun = function()end,         ui = setupUI(),     apply = false, bot_row = false, chb = false})
  table.insert(UI.menu, {name = 'Setup IO',    fun = function()end,         ui = setupIO(),     apply = false, bot_row = false, chb = false})
  table.insert(UI.menu, {name = 'Utilities',   fun = function()end,         ui = utilitiesUI(), apply = false, bot_row = false, chb = false})
  
  
  --table.insert(UI.menu, {name = 'Stats',       fun = statsFun,              ui = stats_UI(),    apply = true,  bot_row = false, chb = false})  
  --table.insert(UI.menu, {name = 'Lib Masks',   fun = function() end,        ui = LibMasksUI(),  apply = false, bot_row = false, chb = false})        
  
  Rand.auto_tb = createFillTb(#UI.menu,false)
  Rand.apply_tb = createFillTb(#UI.menu,false)
  
  UI.rand_menu_n = 1 
        
  local chooser_auto_line = vb:valuebox {
      tooltip = 'Pattern row that triggers auto randomization',
      min = 1,
      max = 128,
      width = 45,
      value = Rand.auto_line,
      notifier = function(value)
        Rand.auto_line = value
      end  
  }   
  
  local function GetAutoChb (i)
    return vb:checkbox {
      tooltip = 'Randomize everytime playback reaches specified pattern line',
      value = Rand.auto_tb[i],
      notifier = function(value) 
        Rand.auto_tb[i] = value
      end
    }  
  end
  
  local auto_chb_tb = {}
  
  for i=1,#UI.menu do
    auto_chb_tb[i] = GetAutoChb(i)
  end 
  
  local chooser_auto = vb:row{}
  
  chooser_auto:add_child(auto_chb_tb[UI.rand_menu_n])

    
  local function GetApplyChb (i)
    return vb:checkbox {
      tooltip = 'Apply',
      value = Rand.apply_tb[i],
      notifier = function(value) 
        Rand.apply_tb[i] = value
      end
    }  
  end
  
  local apply_chb_tb = {}  
   
  for i=1,#UI.menu do
    apply_chb_tb[i] = GetApplyChb(i)
  end  
    
  local chooser_apply_chb = vb:row{}
  
  chooser_apply_chb:add_child(apply_chb_tb[UI.rand_menu_n])     
  
   
  chooser_apply = vb:button {
    tooltip = 'Apply randomizaton',
    text = 'Apply',
    notifier =  function () 
    
      UI.menu[UI.rand_menu_n].fun()
      
      for i,v in ipairs(Rand.apply_tb) do 
        if v then  --and i ~= UI.rand_menu_n
            UI.menu[i].fun()
        end  
      end
    end 
  }  
  
  local chooser_op = vb:column {}  
  
  local chooser_menu = vb:popup {
    width = 100,
    value = UI.rand_menu_n,
    items = map(UI.menu, function(v) return v.name end),
    notifier = function(nix)
      chooser_op:remove_child(UI.menu[UI.rand_menu_n].ui)       
      chooser_auto:remove_child(auto_chb_tb[UI.rand_menu_n])
      chooser_apply_chb:remove_child(apply_chb_tb[UI.rand_menu_n])      
      
      chooser_op:add_child(UI.menu[nix].ui)  
      chooser_auto:add_child(auto_chb_tb[nix])
      chooser_apply_chb:add_child(apply_chb_tb[nix])
      
      -- table.insert(UI.menu, {fun = function() end,   ui = LibMasksUI(),  apply = false, bot_row = false, chb = false})
      
      if UI.menu[nix].apply then
        chooser_apply.visible = true
      else 
        chooser_apply.visible = false 
      end
            
      if UI.menu[nix].bot_row then 
        chooser_bot_row.visible  = true
      else
        chooser_bot_row.visible  = false 
      end
      
      if UI.menu[nix].chb then 
        apply_chb_tb[nix].visible = true
      else
        apply_chb_tb[nix].visible = false 
      end
      
      UI.rand_menu_n = nix 
    end
  } 

  chooser_op:add_child(UI.menu[UI.rand_menu_n].ui)         

  local chooser_top_row = vb:row{
    spacing = 3,
    
    chooser_apply,
    chooser_apply_chb,
    chooser_menu,  
  }  
  
  chooser_bot_row = vb:row{
    spacing = 3,

    vb:text {
      width = 30,
      text = 'Auto'
    },
    chooser_auto,
    
    vb:text {
      width = 30,
      text = ' Line'
    },    
    chooser_auto_line,   
  }  
              
  local chooser_col = vb:column{
    margin = 4,
    spacing = 2,
    
    chooser_top_row,
    chooser_bot_row, 
  }            
  
  local gen_row = vb:row {
    style = 'group',
    spacing = 3,
    
    chooser_col,    
    chooser_op,    
  }  
   
  return gen_row
end

