
local vb = renoise.ViewBuilder()

local rename_dialog

local phrase_tb = {}
local phrase_pu      

local phr_copy  = {}
        
function renamePrompt()
  
  if rename_dialog  and rename_dialog.visible then
     rename_dialog:show()  
     return 
  end  
   
  local name_txf =  vb:textfield {
    width = 150,
    align = "left",
    tooltip = 'Phrase name',
    text = MW.phr[MW.phr.no].name,
  }
    
  local but_ok = vb:button {
    text = "    OK    ",
    notifier =  function ()
      local text = name_txf.text 
      
      phrase_tb[MW.phr.no]   = text   
      MW.phr[MW.phr.no].name = text   
      
      phrase_pu.items = phrase_tb
      
      rename_dialog:close()
    end 
  }
       
  local but_cancel = vb:button {
    text = "  Cancel  ",
    notifier =  function ()
      rename_dialog:close() 
    end 
  }       

  local dialog_content = vb:column { 
    margin = 10,
    spacing = 6,
    
    name_txf,
         
    vb:row {
      but_ok,
      but_cancel,
    },
  }    

  rename_dialog = renoise.app():show_custom_dialog('Enter new name', dialog_content, function(dialog,key) return key end)      
end

function ui_seq_row()
  
  local phrase_no_vb
    
  local phraseTbBuild = function()
    phrase_tb = {}
    for _,phrase in ipairs (MW.phr) do
      table.insert(phrase_tb, phrase.name)
    end  
    
    phrase_pu.items = phrase_tb
  end
  
  local update = function()        
    local col    = MW.phr.col   
    local update = false
    
    UI.suspend_update = true
           
    if MW.phr.editor_poly then
      assignPolyTbs()
          
      if emptyPoly() then 
        sprint('Empty Poly')
        return 
      end
           
      local oct_lowest, oct_highest = 8,2     
           
      for i=1,12 do
        assignNoteTbs(i)
        
        if Nptr.oct_lowest < oct_lowest then
          oct_lowest = Nptr.oct_lowest
        end
      
        if Nptr.oct_highest > oct_highest then
          oct_highest = Nptr.oct_highest
        end
                    
        fill(i)
      end  
      
      UI.oct_st.value  = Nptr.oct_lowest
      UI.oct_end.value = Nptr.oct_highest+1
      
      update = true 

    elseif MW.phr.editor_chords then    
      assignChordTbs()
          
      if noteLengthEmpty(CHptr.len) then return end
          
      UI.oct_st.value  = CHptr.oct_lowest
      UI.oct_end.value = CHptr.oct_highest+1
          
      fill(col)
      update = true
    else
      assignNoteTbs(col)
          
      if noteLengthEmpty(Nptr.len) then return end
          
      UI.oct_st.value  = Nptr.oct_lowest
      UI.oct_end.value = Nptr.oct_highest+1
                             
      fill(col)   
      update = true    
    end 
    
     UI.suspend_update = false
     if update then updateSlots() end 
  end     
  
  local insertPhrase = function()
    local new_phr = MW.phr.no + 1    
    
    local name = 'Phrase ' .. new_phr
    while member(name, phrase_tb) do
      name = name .. '~'
    end
      
    local tb = {}
    tb.name  = name
    tb.notes  = {}
    tb.chords = {}
    table.insert(MW.phr,new_phr,tb)
      
    MW.phr.no = new_phr
    init()      
      
    MW.phr[new_phr].name = name            
      
    phrase_no_vb.max = #MW.phr 
  end  
  
  local phrase_no = 
    vb:text {
      width = 15,
      align = 'center',
      text = tostring(MW.phr.no)
    }

  phrase_no_vb = vb:valuebox {
      min = 1,
      max = #MW.phr,
      width = 55,
      value = 1,
      tooltip = "Internal phrase / sequence number", 
      notifier = function(ph)         
        phrase_pu.value = ph
      end
    }
       
  local phrase_new_bt = vb:button {
    text = ' + ',
    tooltip = 'Add new phrase',
    notifier =  function ()
      
      insertPhrase()
    
      phraseTbBuild() 
    
      phrase_pu.value    = MW.phr.no
      phrase_no_vb.value = MW.phr.no      
    end 
  }

  local phrase_del_bt = vb:button {
    text = ' - ',
    tooltip = 'Delete phrase',
    notifier =  function ()
      if #MW.phr > 1 then
        local phr = MW.phr.no
        
        if MW.phr.no == #MW.phr then
          MW.phr[phr] = nil
          MW.phr.no = MW.phr.no - 1          
          
          phrase_no_vb.value = MW.phr.no
        else
          table.remove(MW.phr,MW.phr.no)            
        end
        
        phrase_no_vb.max = #MW.phr
        phraseTbBuild() 
      end
    end 
  }     

  local phrase_duplicate_bt = vb:button {
    text = ' D ',
    tooltip = 'Duplicate phrase',
    notifier =  function ()
      phr_copy = deepcopy(MW.phr[MW.phr.no]) 
      insertPhrase() 
      MW.phr[MW.phr.no] = deepcopy(phr_copy) 
      MW.phr[MW.phr.no].name = '#' .. MW.phr[MW.phr.no].name 
      
      phraseTbBuild() 
    
      phrase_pu.value    = MW.phr.no
      phrase_no_vb.value = MW.phr.no
    end 
  }  

  local phrase_copy_bt = vb:button {
    text = ' C ',
    tooltip = 'Copy phrase',
    notifier =  function ()
      phr_copy = deepcopy(MW.phr[MW.phr.no]) 
    end 
  } 
  
  local phrase_paste_bt = vb:button {
    text = ' P ',
    tooltip = 'Paste phrase',
    notifier =  function ()
      MW.phr[MW.phr.no] = deepcopy(phr_copy) 
      
      update()
    end 
  } 
  
  local phrase_rename_bt = vb:button {
    text = ' R ',
    tooltip = 'Rename phrase',
    notifier =  function()
      renamePrompt()
    end 
  }  
  
  phrase_pu = vb:popup {
    width = 100,
    value = MW.phr.no,
    items = phrase_tb,
    tooltip = "Phrases",
    notifier = function(new_index)  
    
      if MW.phr.no == new_index then return end
      
      MW.phr.no        = new_index  
      phrase_no_vb.value = new_index
      
      local col = MW.phr.col
      
      local empty = true
      if MW.phr.editor_poly then
        assignPolyTbs()          
        if emptyPoly() then
           sprint('Poly of phrase ' .. MW.phr.no .. ' is empty') 
        else 
          sprint('')
          empty = false
        end                        
      elseif MW.phr.editor_chords then    
        assignChordTbs()  
        
        if lenTb(CHptr) == 0 then
          CHptr = initChords()      
          empty = true 
        elseif noteLengthEmpty(CHptr.len) then 
          empty = true           
        else
          empty = false
        end     
        
        if empty then
          sprint('Chords of phrase ' .. MW.phr.no .. ' are empty')
        else
          sprint('')
        end
      else  
        Nptr = MW.phr[new_index].notes[col]      
        
        if lenTb(Nptr) == 0 then 
          Nptr = initNotes() 
          empty = true
        elseif noteLengthEmpty(Nptr.len) then  
          empty = true           
        else          
          empty = false  
        end
        
        if empty then
          sprint('Column ' .. col .. ' of phrase ' .. new_index .. ' is empty')
        else
          sprint('')
        end
      end 
      
      if not empty then 
        initFillPhr(true)
        update() 
      end   
    end
  }
       
  phraseTbBuild()     
         
  local note_row = vb:row{
    style = 'group', 
    spacing = 1, 
    
    phrase_pu,

    phrase_no_vb,    
    phrase_new_bt,
    phrase_del_bt,
    phrase_rename_bt,
    phrase_copy_bt,
    phrase_paste_bt,
    phrase_duplicate_bt,
  }
  
  UI.phrase_no = phrase_no_vb
  UI.phraseTbBuild = phraseTbBuild
  
  return note_row
end
