

function getPattern(pattern_ix)
  local patterns = renoise.song().patterns
  
  return patterns[pattern_ix] , pattern_ix
end


function curPattern()
  local pattern_ix = renoise.song().selected_pattern_index
  return getPattern(pattern_ix)
end


function getPatTrFun(tr)  
  local pattern =  curPattern ()   
  return function (line, note_col)
    return  pattern:track(tr):line(line):note_column(note_col)
  end
end   


function fillNote2(tr,note, note_col, ln) 
  local pattern =  curPattern ()

  pattern:track(tr):line(ln):note_column(note_col).note_value = note
    
  if note == 121 then
    pattern:track(tr):line(ln):note_column(note_col).instrument_string  = ""
    pattern:track(tr):line(ln):note_column(note_col).volume_string  = ""
  end
end


function fillNoteIns(tr, note, note_col, ln, inst) 
  local pattern = curPattern ()
  
  pattern:track(tr):line(ln):note_column(note_col).note_value = note
  pattern:track(tr):line(ln):note_column(note_col).instrument_value = inst  
end


function fillNotes(notes, tr, ln, inst)
  for note_col,note in ipairs(notes) do
    fillNoteIns(tr, note, note_col, ln, inst)   
  end
end


function patternMW(note_col,type,block,pattern) 

  local sel, pat_len, len
  local inst_no = renoise.song().selected_instrument_index  

  local tr = renoise.song().selected_track_index  
  local tr_en = tr    
  
  if (type == 'pattern' or type == 'pat_fx') then
    pat_len = renoise.song().selected_pattern.number_of_lines  
  else
    pattern = renoise.song().instruments[inst_no].phrases[seq.ph_no]
    pat_len = renoise.song().instruments[inst_no].phrases[seq.ph_no].number_of_lines 
  end
  
  len = pat_len
     
  local st = 0  
    
  if block then 
    if (type == 'pattern' or type == 'pat_fx') then  
      sel = renoise.song().selection_in_pattern 
      
      if (sel == nil) then return function() end end 
      
      tr = sel.start_track  
      tr_en = sel.end_track 
    else
      sel = renoise.song().selection_in_phrase   
      tr = 1
      tr_en = 1
      
      note_col = {}
      for i=sel.start_column, sel.end_column do
       table.insert(note_col,i)     
      end      
    end
      
    if (sel ~= nil) then    

      len = sel.end_line - sel.start_line
    
      st = sel.start_line  - 1  
      pat_len = sel.end_line 
      len = sel.end_line - st      
    end
  end

  local line = st
  local j = 0
  local en = len * #note_col * (tr_en-tr+1)
  local col_ix = 1
  return function ()
    j = j+1
    line = line+1
      
    local ncol = note_col[col_ix]

    if (j <= en) then
      if (line > pat_len and j ~= en) then 
        col_ix = col_ix + 1
          
        if (col_ix>#note_col) then
          col_ix = 1
        end     

        if (line * ncol > #note_col * len) then
          tr = tr + 1
          col_ix = 1
        end
          
        ncol = note_col[col_ix]     
        line = st + 1                   
      end      
      
      if not pattern.tracks[tr].is_empty then
      
        if (type == 'pattern' or type == 'pat_fx') then        
          local is_selected = pattern:track(tr):line(line):note_column(ncol).is_selected
        
          local fun        
          if (type == 'pattern') then
            fun = pattern:track(tr):line(line):note_column(ncol)
          elseif (type == 'pat_fx') then
            fun = pattern:track(tr):line(line):effect_column(ncol)
          end
        
          return fun, line, ncol, tr, is_selected
        else
          return pattern:line(line):note_column(ncol),line,ncol,1,true
        end
      end  
    end    
  end    
end


function resizeBlock (newLen) 

  local selection = {}
  local tr_col = {}
  tr_col.tr = {}
  tr_col.col = {}
   
  
  local all_cols = {1,2,3,4,5,6,7,8,9,10,11,12}
    
  local sel = renoise.song().selection_in_pattern 
  local len = sel.end_line - sel.start_line + 1
  
  local pattern  = curPattern ()

  local ln2   
    
  if sel ~= nil then
    
    local rt = newLen / len

    for pat,ln,col,tr,sel in patternMW(all_cols,'pattern',true,pattern) do
      if sel then      
        if (pat.note_value < 121) then
     
          if selection[tr] == nil then
            selection[tr] = {}
          end
          if selection[tr][ln] == nil then
            selection[tr][ln] = {}
          end
      
          selection[tr][ln][col] = { 
            line = ln, 
            column = col,
            track = tr,
            note_value = pat.note_value, 
            instrument_value = pat.instrument_value , 
            volume_value = pat.volume_value,   
            panning_value = pat.panning_value , 
            delay_value = pat.delay_value }    
        end
           
        pat.volume_value = 255 
        pat.panning_value = 255       
        pat.delay_value = 0            
        pat.note_value = 121
        pat.instrument_value = 255        
        
        if not(member(tr, tr_col.tr)) then table.insert (tr_col.tr, tr) end
        if not(member(col, tr_col.col)) then table.insert (tr_col.col, col) end
      end
    end 
      
    for i, tr in pairs(tr_col.tr) do
      for j, cl in pairs(tr_col.col) do
        for ln=sel.start_line, sel.end_line * rt do
            
          pattern:track(tr):line(ln):note_column(cl).note_value = 121 
          pattern:track(tr):line(ln):note_column(cl).instrument_value = 255  
          pattern:track(tr):line(ln):note_column(cl).volume_value = 255    
          pattern:track(tr):line(ln):note_column(cl).panning_value = 255  
          pattern:track(tr):line(ln):note_column(cl).delay_value = 0
        end
      end
    end    
              
    for i, tr in pairs(selection) do
      for j, ln in pairs(tr) do
        for k, col in pairs(ln) do
          ln2 = round((col.line - sel.start_line + 1) * rt + sel.start_line - 1 )
          
          if (rt < .5) then
            ln2 = ln2 + 1
          end

          if (ln2 == 0 or (ln == sel.start_line and ln2 > ln)) then
            ln2 = sel.start_line
          end

          local tr = col.track
          local cl = col.column
              
          pattern:track(tr):line(ln2):note_column(cl).note_value = col.note_value  
          pattern:track(tr):line(ln2):note_column(cl).instrument_value = col.instrument_value   
          pattern:track(tr):line(ln2):note_column(cl).volume_value = col.volume_value        
          pattern:track(tr):line(ln2):note_column(cl).panning_value = col.panning_value   
          pattern:track(tr):line(ln2):note_column(cl).delay_value = col.delay_value
        end
      end
    end    
    
    
    all_cols = {1,2,3,4,5,6,7,8}
    selection = {}
    
    tr_col.tr = {}
    tr_col.col = {}
  
    for fx,ln,col,tr,sel in patternMW(all_cols,'pat_fx',true,curPattern ()) do
      if sel then      
     
      if (fx.amount_value > 0 or fx.number_value > 0) then 
          if selection[tr] == nil then
            selection[tr] = {}
          end
          if selection[tr][ln] == nil then
            selection[tr][ln] = {}
          end
      
          selection[tr][ln][col] = { 
            line = ln, 
            column = col,
            track = tr,
            fx_amt = fx.amount_value, 
            fx_num = fx.number_value }    
        end
           
        fx.number_value = 0
        fx.amount_value = 0   
        
        if not(member(tr, tr_col.tr)) then table.insert (tr_col.tr, tr) end
        if not(member(col, tr_col.col)) then table.insert (tr_col.col, col) end
      end
    end 
    
    for i, tr in pairs(tr_col.tr) do
      for j, cl in pairs(tr_col.col) do
        for ln=sel.start_line, sel.end_line * rt do
            
          pattern:track(tr):line(ln):effect_column(cl).number_value = 0
          pattern:track(tr):line(ln):effect_column(cl).amount_value = 0  
        end
      end
    end  

    for i, tr in pairs(selection) do
      for j, ln in pairs(tr) do
        for k, col in pairs(ln) do
          ln2 = round((col.line - sel.start_line + 1) * rt + sel.start_line - 1 )
          
          if (rt < .5) then
            ln2 = ln2 + 1
          end

          if (ln2 == 0 or (ln == sel.start_line and ln2 > ln)) then
            ln2 = sel.start_line
          end

          local tr = col.track
          local cl = col.column
          
          pattern:track(tr):line(ln2):effect_column(cl).number_value = col.fx_num
          pattern:track(tr):line(ln2):effect_column(cl).amount_value = col.fx_amt   
        end
      end
    end   
  end
end
   
  
function combine2 (poly,n)
  
    local rn = renoise.song()
    local pat_len = rn.selected_pattern.number_of_lines  
    local pattern =  curPattern ()
    local tr = rn.selected_track_index 
    
    pat_len = renoise.song().selected_pattern.number_of_lines
    
    local col = 1
    if poly then
      col = 2
    end
        
    local i = 1
    while (rn.tracks[i].min_note_columns > 0) do
      if (rn.tracks[i].mute_state == 1 and i ~= tr and col < 13) then
      
        for j=1,pat_len do
          local note = pattern:track(i):line(j):note_column(1).note_value
          if (note < 121 and pattern:track(tr):line(j):note_column(1).note_value >= n) then 
            pattern:track(tr):line(j):note_column(col).note_value = note  
            local inst = pattern:track(i):line(j):note_column(1).instrument_value 
            pattern:track(tr):line(j):note_column(col).instrument_value = inst 
          end      
        end  
        
        if poly then
          col = col + 1
        end       
      end     
    i = i+1  
    end 
end


function combine(poly,...)
  if (arg.n == 0) then
    combine2(poly, 120)
  elseif (arg[1] == "rep") then
    combine2(poly, 0)
  end
end


function combinePolyHlp(pattern)
    
  local get_col = function(pattern, note_in)
    local notes = {}
    
    for col=1,12 do
      local note = pattern:note_column(col).note_value
      
      if note > 119 then 
      
        local exist = member(note_in, notes)  
        if not exist then  
          return col
        end
      else  
        table.insert(notes, note)        
      end
    end
  end
  
  local rn      = renoise.song()
  local tr      = rn.selected_track_index     
  local pat_len = pattern.number_of_lines     
         
  local i = 1
  while (rn.tracks[i].min_note_columns > 0) do
    if (rn.tracks[i].mute_state == 1 and i ~= tr) then
        
      for col=1,12 do
        for j=1,pat_len do
          local note = pattern:track(i):line(j):note_column(col).note_value
          if (note < 120) then 
              
            local dst_col = get_col(pattern:track(tr):line(j), note)
             
            if dst_col then
              pattern:track(tr):line(j):note_column(dst_col).note_value = note  
              local inst = renoise.song().selected_instrument_index - 1 
              pattern:track(tr):line(j):note_column(dst_col).instrument_value = inst 
            end
          end      
        end  
      end        
               
    end     
  i = i+1  
  end 
end


function combinePoly()
  local seq, pos_tb = {},{}
  
  if MW.grab_seq then   
    local song = renoise.song()

    local song_len = #song.sequencer.pattern_sequence 
    local patterns = song.sequencer.pattern_sequence 

    local range = renoise.song().sequencer.selection_range 
     
    if range[1] > 0 then          
      for i=range[1],range[2] do
      
        local pattern =  getPattern(patterns[i])                
        combinePolyHlp(pattern)
      end  
    else
      sprint('Range empty')
    end    
  else
    local pattern_ix = renoise.song().selected_pattern_index  
    local pattern =  getPattern (pattern_ix)
 
    combinePolyHlp(pattern)
  end
end


function moveTo1stCol(pat)

  local tr = renoise.song().selected_track_index  
  local pat_len = renoise.song().selected_pattern.number_of_lines  
  
  for ln = 1, pat_len do
    local n_col_dst = 1
    local pat_n_col_dest = pat:track(tr):line(ln):note_column(n_col_dst)
       
    for n_col = 2,12 do      
      local pat_n_col = pat:track(tr):line(ln):note_column(n_col)
         
      if (pat_n_col.note_value < 120 and pat_n_col_dest.note_value >= 120) then
        pat_n_col_dest.note_value = pat_n_col.note_value
        pat_n_col_dest.instrument_value = pat_n_col.instrument_value    
               
        pat_n_col.note_value = 121
        pat_n_col.instrument_value = 255
       
        n_col_dst = n_col_dst+1
        pat_n_col_dest = pat:track(tr):line(ln):note_column(n_col_dst)                       
      end
    end
  end   
end


function trEmpty()
  local col_tb = {MW.phr.col}
  
  local type = 'pattern'
  
  if MW.edit_RN_phrase then
    type = 'phrase'
  end
  
  for pat,ln,col,tr,sel in patternMW(col_tb,type,false,curPattern ()) do
    if (pat.note_value < 121) then
      return false
    end
  end
  return true  
end


function datiter (mw,tb,in_pattern)

  local pat_con, pat_len, line2 = getPatFun(mw, MW.block_editing)

  if (MW.block_editing == false) then line2 = line2 - 1 end

  local line = 0
  
  local i = 0
  
  return function ()
    i=i+1
     
    local dur = 0
      
    if  (i <= #Nptr.len) then
      line = line +  Nptr.len[i]
        
      if (Nptr.off[i] > 0) then
        dur = Nptr.off[i]  
      elseif (i+1 <= #Nptr.len) then
        dur = Nptr.len[i+1]
      elseif (pat_len >= line) then 
        dur = pat_len - line + 1   
      end
    end    
      
    if (in_pattern and line > pat_len) then return end
      
    if (line > line2 and i <= #Nptr.len and Nptr.len[i] > 0) then
      local ret_tb = {}
        
      for j,key in ipairs(tb) do
        if (key == "dur") then
          ret_tb[j] = dur
        elseif (key == "ln") then
          ret_tb[j] = line
        elseif (key == "i") then
          ret_tb[j] = i  
        else          
          ret_tb[j] = mw[key][i]           
        end
      end        
      return ret_tb[1],ret_tb[2],ret_tb[3],ret_tb[4],ret_tb[5],ret_tb[6]       
    end   
  end
end  


function getLen(len_tb)
  local _, pat_len, line2 = getPatFun(MW.block_editing)
    
  local line = 0  
  local last_ln = 0
  local pos = 0 
  
  if MW.block_editing then pos = line2 end
 
  for k,v in ipairs(len_tb) do
    last_ln = line
    line = line + v
    
    if line > pos then  
      return k, last_ln,pat_len
    end    
  end
   
  return 1,last_ln,pat_len 
end


function emptyTr()
  local pat    
  local pat_con, pat_len, line2 = getPatFun(MW.block_editing)

  local line = 0 

  if MW.block_editing then line = line2  end 
  
  for j=1,12 do    
    for i=line+1,pat_len do 
      pat = pat_con(i, j)
      pat.note_value = 121
      pat.instrument_string  = ""
      pat.volume_string  = ""
    end   
  end 
end


function emptyTrN(tr)
  local pat = curPattern ()
  local pat_len = renoise.song().selected_pattern.number_of_lines  
  
  for note_col=1,12 do    
    for ln=1,pat_len do 
      pat:track(tr):line(ln):note_column(note_col).note_value = 121
      pat:track(tr):line(ln):note_column(note_col).instrument_string  = ""
      pat:track(tr):line(ln):note_column(note_col).volume_string  = ""
    end   
  end 
end


function nodeAdd(pitches,chrd,i,chrd_tb)

  if i <= #chrd then  
    local key = chrd[i]

    if type(pitches) == "table" then
      if pitches[key] == nil then
        pitches[key] = {}
        
        if i == #chrd then
          pitches[key].chrd = chrd_tb     
        else                    
          nodeAdd(pitches[key],chrd,i+1,chrd_tb)        
        end               
      elseif i == #chrd and pitches[key].chrd == nil then
        pitches[key].chrd = chrd_tb     
      elseif i == #chrd and pitches[key].chrd then
        if pitches[key].chrd.scale > 3 then
          pitches[key].chrd = chrd_tb   
       end 
      else     
        nodeAdd(pitches[key],chrd,i+1,chrd_tb)
      end      
    end   
  end
end


function genChrdTb()
  Scan.chords_tb = {}
  
  for note=0,12 do    
    for scale=2,#Scales.forms do
      for key=1,12 do 
        for form=1,#Chords.forms do      

          local chrd = getChrd2oct (note, scale, key, Chords.forms[form])  
       
          local chrd_tb   = {}
          chrd_tb.key     = key
          chrd_tb.scale   = scale
          chrd_tb.note    = note
          chrd_tb.form    = form     
          chrd_tb.chrd_tb = chrd  
          
          nodeAdd(Scan.chords_tb,chrd,1,chrd_tb)  
        end
      end        
    end
  end  
end


function genChrdTbOct0()
  Scan.chords_tb_oct0 = {}
  
  for note=0,12 do    
    for scale=2,#Scales.forms do
      for key=1,12 do 
        for form=1,#Chords.forms-1 do      

          local chrd = getChrd (note, scale, key, Chords.forms[form]) 
       
          local chrd_tb   = {}      
          chrd_tb.key     = key
          chrd_tb.scale   = scale
          chrd_tb.note    = note
          chrd_tb.form    = form     
          chrd_tb.chrd_tb = chrd  
          
          nodeAdd(Scan.chords_tb_oct0,chrd,1,chrd_tb)  
        end
      end        
    end
  end  
end


function genChrdTbOct0SK(dst, scale, key)
  for note=0,12 do  
    for form=1,#Chords.forms-1 do      
    
      local chrd = getChrd (note, scale, key, Chords.forms[form])
       
      local chrd_tb   = {}      
      chrd_tb.key     = key
      chrd_tb.scale   = scale
      chrd_tb.note    = note
      chrd_tb.form    = form     
      chrd_tb.chrd_tb = chrd  
          
      nodeAdd(dst,chrd,1,chrd_tb)  
    end
  end            
end


function getChFromNodes(ch_tb,tb,i) 
  if i <= #tb and ch_tb then
    local k = tonumber(tb[i])
    if i == #tb and ch_tb[k] == nil then 
      return nil
    elseif i == #tb and ch_tb[k].chrd then
      return ch_tb[k].chrd
    else
      return getChFromNodes(ch_tb[k],tb,i+1)  
    end
  end 
end


function blockNotesNrow(n)   
  local ln2
  local sel = renoise.song().selection_in_pattern 
  local note_tb = {}  
  
  if sel ~= nil then      
    local pat_con = getPatFun(MW.block_editing)      
    local k = 1
    
    for line=sel.start_line, sel.end_line do 
      local note_row = {}
      
      for n_col=sel.start_column,sel.end_column do
        if n_col<=12 then   
          local pat = pat_con(line, n_col)      
       
          if pat.note_value < 120  then 
            table.insert(note_row, pat.note_value)
          end
        end
      end   
      
      if #note_row > 0 then   
        table.insert(note_tb, note_row)
        k=k+1 
      end
      
      if k > n then
        return note_tb
      end              
    end
  end  
  return note_tb
end    


function randomExceptHelper (exept_tb, rand)
  local res = false
  
  for _,v in ipairs(exept_tb) do
    if v == rand then 
      return true 
    end
  end
  return res  
end


function randomExcept(n,except_tb)
  local rand = math.random(n)
  
  while randomExceptHelper (except_tb, rand) do
    rand = math.random(n)
  end
   
  return rand 
end


function remapInstHlp(pattern)
  
  local rn      = renoise.song()
  local tr      = rn.selected_track_index     
  local inst    = renoise.song().selected_instrument_index - 1
  local pat_len = pattern.number_of_lines     
         
  local i = 1
  while (rn.tracks[i].min_note_columns > 0) do
    if (rn.tracks[i].mute_state == 1) then
        
      for col=1,12 do
        for j=1,pat_len do
          local note = pattern:track(i):line(j):note_column(col).note_value
          if (note < 120) then 
            pattern:track(i):line(j):note_column(col).instrument_value = inst             
          end      
        end  
      end        
               
    end     
  i = i+1  
  end 
end


function remapInst()
  local seq, pos_tb = {},{}
  
  if MW.grab_seq then   
    local song = renoise.song()

    local song_len = #song.sequencer.pattern_sequence 
    local patterns = song.sequencer.pattern_sequence 

    local range = renoise.song().sequencer.selection_range 
     
    if range[1] > 0 then          
      for i=range[1],range[2] do
      
        local pattern =  getPattern(patterns[i])                
        remapInstHlp(pattern)
      end  
    else
      sprint('Range empty')
    end    
  else
    local pattern_ix = renoise.song().selected_pattern_index  
    local pattern =  getPattern (pattern_ix)
 
    remapInstHlp(pattern)
  end
end


function clearTrack()   
  local pat_con, pat_len = getPatFun(false)

  for j=1,12 do 
    for i=1,pat_len do  
      local pat = pat_con(i,j)
      
      pat.note_value         = 121
      pat.volume_string      = ""
      pat.instrument_string  = ""        
    end   
  end 
end    


function buildChordTable()
  genChrdTb() 
     
  local fn = MW.path .. 'chord_tb2.lua'   
  if fileExists(fn) then
    os.remove (fn)   
  end 
        
  local file = io.open(fn, "wb") 
  sprint('Saving ' .. fn)
  
  file:write('Scan.chords_tb = ' .. serialize(Scan.chords_tb))    
  file:close()   
end


function buildChordTable0()
  genChrdTbOct0() 
     
  local fn = MW.path .. 'chord_tb.lua'   
  if fileExists(fn) then
    os.remove (fn)   
  end 
        
  local file = io.open(fn, "wb") 
  sprint('Saving ' .. fn)
  
  file:write('Scan.chords_tb_oct0 = ' .. serialize(Scan.chords_tb_oct0))   
  file:close()   
end


function buildChordTable0SK()
  Scan.chords_tb_oct0_sk = {}
  
  for scale=2,#Scales.forms do
    for key=1,12 do 
      local sk = scale .. '_' .. key
      Scan.chords_tb_oct0_sk[sk] = {}
      
      genChrdTbOct0SK(Scan.chords_tb_oct0_sk[sk], scale, key)
    end
  end  
     
  local fn = MW.path .. 'chord_tb_sk.lua'   
  if fileExists(fn) then
    os.remove (fn)   
  end 
        
  local file = io.open(fn, "wb") 
  sprint('Saving ' .. fn)
  
  file:write('Scan.chords_tb_oct0_sk = ' .. serialize(Scan.chords_tb_oct0_sk)) 
  file:close()   
end


function buildFormCS2()
  Scan.forms_tb_cs = {}
  
  Scan.forms_tb_cs.cs = Stats.forms_cs
  
  for scale=2,#Scales.forms do
    sprint("CS scan scale " .. Scales.names[scale])
    
    coroutine.yield()
    
    for key=1,12 do 
      local sk = scale .. '_' .. key
      Scan.forms_tb_cs[sk] = {}
        
      for note = 0,11 do        
        local forms_ix, forms_str = {},{}        
                
        Scan.forms_tb_cs[sk][note] = {}
          
        for f = 1, #Chords.forms-1 do           
          local form = Chords.forms[f] 
          
          local chord = {}
          for _,f in ipairs(form) do  
            local noteChrd = transposeCh(f, note ,0 , scale, key)
            table.insert(chord, noteChrd)
          end        
          
          chord = unique(chord)
          table.sort(chord)
          
          local cs = consonanceRating2Tb(chord,chord)
          
          if cs < Stats.forms_cs then
            table.insert(forms_ix, f)  
            table.insert(forms_str, Chords.forms_to_str[f])
          end
        end
        
        Scan.forms_tb_cs[sk][note].ix  = forms_ix
        Scan.forms_tb_cs[sk][note].str = forms_str  
      end 
    end
  end 
  
  local fn = MW.path .. 'chord_tb_cs.lua'   
  if fileExists(fn) then
    os.remove (fn)   
  end 
        
  local file = io.open(fn, "wb") 
  sprint('Saving ' .. fn)
  
  file:write('Scan.forms_tb_cs = ' .. serialize(Scan.forms_tb_cs)) 
  file:close() 
end


function buildFormCS()
  local process = ProcessSlicer(buildFormCS2, function() end)
  process:start()
end

